
from collections import defaultdict
import os
from matplotlib import pyplot as plt
import pandas as pd
from tqdm import tqdm


import re

import logging

from features_aux import add_lags, get_date_exog, normalize_cols, add_all_lags

logger = logging.getLogger("WATTSON_EV_IDS.FeaturesCPO")

ROUND_TO=6

def get_cpo_files(DIR):
    ret={}
    sub_folders = [name for name in os.listdir(DIR) if os.path.isdir(os.path.join(DIR, name))]
    for x in [f for f in sub_folders if f.startswith("CPO_")]:
        x_path = os.path.join(DIR, x)
        onlyfiles = [f for f in os.listdir(x_path) if os.path.isfile(os.path.join(x_path, f))]
        onlyfiles = [f for f in onlyfiles if f.startswith("cpo_ocpp_data_") and f.endswith(".csv.gz")]
        ret[x] = onlyfiles
    return ret

def read_cpo_files(dir, files):
    # print(dir, files)
    regex = re.compile(r"^cpo_ocpp_data_(CP_\d+)_([\w\d]+).csv.gz$")
    cpo_files=defaultdict(lambda: dict())
    for f in tqdm(files, desc="loading cpo_files"):
        #cpo_ocpp_data_CP_1_0bc01518df38656accbde55bfb0e38591.csv.gz
        result = regex.search(f)
        if result:
            f_df = pd.read_csv(os.path.join(dir, f), index_col="meter_value_ts", parse_dates=["meter_value_ts"])
            if result.group(1) != f_df.iloc[0]["cp_group"]:
                logger.error(f"wrong cp_group {result.group(1)} != {f_df.iloc[0]['cp_group']}")
            # f_df = f_df[[c for c in f_df.columns if not c.startswith("custom_data") and c != "cp_c_id" and c != "cp_group"]]
            cpo_files[result.group(1)][result.group(2)] = f_df
            # break
        else:
            logger.error(f"unk file {f}")
            continue
    return cpo_files



def get_features_cpo(df, per_cp_normalize, per_cp_date_exog):
    keep_cols = ["charge_speed", "meter_value_sampled_value", "meter_diff", "time_diff", "cp_c_id"]
    f_df = df[[c for c in df.columns if c != "cp_group"]] #  c != "cp_c_id" and not c.startswith("custom_data") and
    f_df = pd.concat([f_df, pd.get_dummies(f_df["event"], prefix="event")], axis=1)
    # f_df["meter_value_p"] = (f_df["meter_value_sampled_value"] - f_df["meter_value_sampled_value"].min()) / (f_df["meter_value_sampled_value"].max() - f_df["meter_value_sampled_value"].min())
    # f_df["meter_value_p_diff"] = f_df["meter_value_p"] - f_df["meter_value_p"].shift(1)
    # f_df["meter_value_p_diff"] = f_df["meter_value_p_diff"].fillna(0)
    f_df["time_diff"] = f_df["time_diff"].fillna(0)

    f_df["is_attack"] = 0
    f_df.loc[f_df[f_df["custom_data_atk_type"] != "None"].index, "is_attack"] = 1
    f_df = pd.concat([f_df, pd.get_dummies(f_df["custom_data_atk_type"], prefix="custom_data_atk_type")], axis=1)

    if per_cp_date_exog:
        c1 = f_df.columns
        get_date_exog(f_df, prefix="date_exog_cp_")
        c_del = [c for c in f_df.columns if c not in c1]
        for c in f_df.columns:
            if c.startswith("date_exog_cp_"):
                f_df = pd.concat([f_df, pd.get_dummies(f_df[c], prefix=c)], axis=1)
        f_df = f_df[[c for c in f_df.columns if c not in c_del]] #del non dummy date exog
        keep_cols+= [c for c in f_df.columns if c.startswith("date_exog_cp_")]

    keep_cols+= [c for c in f_df.columns if c.startswith("custom_data") and "custom_data_vendor_id" not in c and "custom_data_atk_type" not in c]
    keep_cols+= ["is_attack"]
    f_df = f_df[[c for c in f_df.columns if c.startswith("event_") or c in keep_cols]]

    # if len(f_df["is_attack"].drop_duplicates()) > 0:
    #     print(f_df)
    #     exit()
    # if len(f_df["is_attack"].drop_duplicates()) > 0:
    #     print(f_df)
    #     exit()

    if per_cp_normalize:
        normalize_cols(f_df, "_cp_norm")
    return f_df.copy()

def resample_feats3(df, t_min, t_max, freq='15Min'):
    oidx = df.index
    t_min2=df.index.min()
    t_max2=df.index.max()
    #nidx = pd.date_range(oidx.min(), oidx.max(), freq='5Min')
    nidx = pd.date_range(t_min2, t_max2, freq=freq) #, normalize=True
    res = df.reindex(oidx.union(nidx)).interpolate('time', limit_direction="both").reindex(nidx)
    res[[c for c in df.columns if "event_Ended" in c]] = df[[c for c in df.columns if "event_Ended" in c]].reindex(oidx.union(nidx)).interpolate('time', limit_direction="backward").reindex(nidx)
    res[[c for c in df.columns if "event_Started" in c]] = df[[c for c in df.columns if "event_Started" in c]].reindex(oidx.union(nidx)).interpolate('time', limit_direction="forward").reindex(nidx)
    res[[c for c in df.columns if "event_Ended" in c]] = res[[c for c in df.columns if "event_Ended" in c]].fillna(0)
    res[[c for c in df.columns if "event_Started" in c]] = res[[c for c in df.columns if "event_Started" in c]].fillna(0)
    res = resample_feats(res, t_min, t_max, freq)
    return res

def resample_feats2(df, t_min, t_max, freq='15Min'):
    oidx = df.index
    #nidx = pd.date_range(oidx.min(), oidx.max(), freq='5Min')
    nidx = pd.date_range(t_min, t_max, freq=freq)
    res = df.reindex(oidx.union(nidx)).interpolate('time', limit_direction="both").reindex(nidx)
    res[[c for c in df.columns if "event_Ended" in c]] = df[[c for c in df.columns if "event_Ended" in c]].reindex(oidx.union(nidx)).interpolate('time', limit_direction="backward").reindex(nidx)
    res[[c for c in df.columns if "event_Started" in c]] = df[[c for c in df.columns if "event_Started" in c]].reindex(oidx.union(nidx)).interpolate('time', limit_direction="forward").reindex(nidx)
    res[[c for c in df.columns if "event_Ended" in c]] = res[[c for c in df.columns if "event_Ended" in c]].fillna(0)
    res[[c for c in df.columns if "event_Started" in c]] = res[[c for c in df.columns if "event_Started" in c]].fillna(0)

    for fward_c in  ["meter_value_sampled_value", "meter_diff", "charge_speed", "meter_no_atk"]: 
        fward_cs = [c for c in df.columns if fward_c in c]
        res[fward_cs] = df[fward_cs].reindex(oidx.union(nidx)).ffill().reindex(nidx)
    for fward_c in  ["meter_value_sampled_value", "meter_no_atk"]:
        fward_cs = [c for c in df.columns if fward_c in c]
        res[fward_cs] = res[fward_cs].fillna(res[fward_cs].min())
    for fward_c in  ["meter_diff", "charge_speed"]:
        fward_cs = [c for c in df.columns if fward_c in c]
        res[fward_cs] = res[fward_cs].fillna(0)
    return res

def resample_feats(df, t_min, t_max, freq='15Min'):
    oidx = df.index
    #nidx = pd.date_range(oidx.min(), oidx.max(), freq='5Min')
    nidx = pd.date_range(t_min, t_max, freq=freq)
    res = df.reindex(oidx.union(nidx)).interpolate('time', limit_direction="both").reindex(nidx)
    res[[c for c in df.columns if "event_Ended" in c]] = df[[c for c in df.columns if "event_Ended" in c]].reindex(oidx.union(nidx)).interpolate('time', limit_direction="backward").reindex(nidx)
    res[[c for c in df.columns if "event_Started" in c]] = df[[c for c in df.columns if "event_Started" in c]].reindex(oidx.union(nidx)).interpolate('time', limit_direction="forward").reindex(nidx)
    res[[c for c in df.columns if "event_Ended" in c]] = res[[c for c in df.columns if "event_Ended" in c]].fillna(0)
    res[[c for c in df.columns if "event_Started" in c]] = res[[c for c in df.columns if "event_Started" in c]].fillna(0)
    return res


def get_cpo_dfs(cpo_files, per_cp_normalize, per_cp_date_exog):
    cpo_dfs=dict()
    t_min=dict()
    t_max=dict()
    for group, cps in tqdm(cpo_files.items(), desc="loading cpo_dfs"):
        cp_dfs=[]
        for cp, cp_df in cps.items():
            cp_dff = get_features_cpo(cp_df, per_cp_normalize=per_cp_normalize, per_cp_date_exog=per_cp_date_exog)
            if group not in t_min:
                t_min[group] = cp_dff.index.min()
                t_max[group] = cp_dff.index.max()
            else:
                t_min[group] = min(t_min[group], cp_dff.index.min())
                t_max[group] = max(t_max[group], cp_dff.index.max())
            
            cp_dfs.append(cp_dff)
            # print(group, cp, cp_dff)
            # exit()
        cpo_dfs[group] = cp_dfs
    return cpo_dfs, t_min, t_max


    # ['meter_value_sampled_value', 'custom_data_meter_no_atk',
    #    'custom_data_scaling', 'custom_data_energy_interval',
    #    'custom_data_original_energy_interval', 'custom_data_average_power',
    #    'custom_data_original_average_power', 'meter_diff', 'time_diff',
    #    'charge_speed', 'custom_data_meter_diff', 'custom_data_charge_speed',
    #    'event_Ended', 'event_Started', 'event_Updated', 'is_attack',
    #    'date_exog_cp_dayofweek_1', 'date_exog_cp_dayofweek_2',
    #    'date_exog_cp_dayofweek_6', 'date_exog_cp_hour_7',
    #    'date_exog_cp_hour_8', 'date_exog_cp_hour_9', 'date_exog_cp_hour_10',
    #    'date_exog_cp_hour_11', 'date_exog_cp_hour_12', 'date_exog_cp_hour_13',
    #    'date_exog_cp_hour_14', 'date_exog_cp_hour_15', 'date_exog_cp_hour_16',
    #    'date_exog_cp_discretize_hour_only_Play',
    #    'date_exog_cp_discretize_hour_only_Work',
    #    'date_exog_cp_discretize_hour_day_High-Home',
    #    'date_exog_cp_discretize_hour_day_High-Work',
    #    'date_exog_cp_discretize_day_is_work_False',
    #    'date_exog_cp_discretize_day_is_work_True',
    #    'date_exog_cp_discretize_hour_balancing_low',
    #    'date_exog_cp_discretize_hour_balancing_peak'],
def _add_individual_cp_features(cp_dfs):
    for cp_df in cp_dfs:
        do_cols = [c for c in cp_df.columns if "date_exog_" not in c and "custom_data_" not in c  and c != "cp_c_id"]
        cp_id = cp_df["cp_c_id"].drop_duplicates().iloc[0]
        cp_df2 = cp_df[[c for c in do_cols]].copy()
        cp_df2.rename(columns={c:"CP_"+str(cp_id)+"_"+c for c in do_cols}, inplace=True)
        cp_df = pd.concat([cp_df, cp_df2], axis=1)
        cp_df.drop(columns="cp_c_id", inplace=True)
        yield cp_df


def insert_sess_ends(cp_dff):
    #print([c for c in cp_dff.columns])
    #print(cp_dff[['time_diff', 'charge_speed','event_Ended', 'event_Started', 'event_Updated', "is_attack",]])
    #print(cp_dff[cp_dff['event_Ended'] == 1])
    if "event_Ended" not in cp_dff.columns:
        cp_dff["event_Ended"] = 0
        cp_dff.iloc[-1,cp_dff.columns.get_loc("event_Ended")] = 1
        cp_dff.iloc[-1,cp_dff.columns.get_loc("event_Updated")] = 0
        cp_dff.iloc[-1,cp_dff.columns.get_loc("event_Started")] = 0
    if cp_dff.iloc[-1]["event_Ended"] != 1:
        logger.debug('cp_dff.iloc[-1]["event_Ended"] != 1')
        #logger.error(cp_dff[['time_diff', 'charge_speed','event_Started', 'event_Updated', "event_Ended", "is_attack",]])
        cp_dff.iloc[-1,cp_dff.columns.get_loc("event_Ended")] = 1
        cp_dff.iloc[-1,cp_dff.columns.get_loc("event_Updated")] = 0
        cp_dff.iloc[-1,cp_dff.columns.get_loc("event_Started")] = 0
        #exit()
    for idx in cp_dff[cp_dff['event_Ended'] == 1].index:
        idx_l = cp_dff.index.get_loc(idx)
        idx_c = cp_dff.index[idx_l]
        idx_max=len(cp_dff)-1
        idx_n = cp_dff.index[min(idx_max,idx_l+1)]
        new_idx=None
        time_diff=None
        for x in [10,1,0.1]:
            #print(idx_c+pd.Timedelta(seconds=x),idx_n)
            if idx_c+pd.Timedelta(seconds=x) < idx_n or idx_n == cp_dff.index[idx_max]:
                new_idx = idx_c+pd.Timedelta(seconds=x)
                time_diff=(x/60)/60 # s to h
                break
        if new_idx is None:
            raise Exception(f"no new_idx found for {idx_c=}, {idx_n=} in:\n {cp_dff}")
        #print(new_idx)
        cp_dff.loc[new_idx] = cp_dff.loc[idx_c]
        cp_dff.loc[new_idx, "time_diff"] = time_diff
        zero_list = ['energy_interval', 'average_power', 'meter_diff', "charge_speed", "is_attack"]
        for z in zero_list:
            for c in cp_dff.columns:
                if z in c:
                    cp_dff.loc[new_idx, c] = 0
        cp_dff = cp_dff.sort_index()
    #print(cp_dff[['time_diff', 'charge_speed','event_Ended', 'event_Started', 'event_Updated', "is_attack",]])
    return cp_dff


def get_cp_df_interp2(cp_dfs, group, t_min, t_max, add_individual_cp_features=True):
    base_cols=sorted(set([c for cp_df in cp_dfs for c in cp_df.columns if c != "cp_c_id"]))
    if add_individual_cp_features:
        cp_dfs = _add_individual_cp_features(cp_dfs)

    cp_df_interp2=None
    for cp_dff in cp_dfs:
        cp_dff = insert_sess_ends(cp_dff)
        cp_dff = resample_feats(cp_dff, t_min[group], t_max[group])
        # fig, ax = plt.subplots(figsize=(16,9))
        # ax.plot(cp_dff.index, cp_dff["charge_speed"], label="charge_speed")
        # plt.legend()
        # plt.show()
        # plt.close()
        # print(row)
        #exit()
        if  cp_df_interp2 is None:
            cp_df_interp2=cp_dff
        else:
            add_cols=[c for c in cp_dff.columns if c in base_cols and c in cp_df_interp2.columns]
            concat_cols=[c for c in cp_dff.columns if c not in add_cols]
            cp_df_interp2[add_cols]+=cp_dff[add_cols]
            # cp_df_interp2 = pd.concat([cp_df_interp2, cp_dff[add_cols]]) 
            # cp_df_interp2 = cp_df_interp2.groupby(cp_df_interp2.index).sum()
            cp_df_interp2 = pd.concat([cp_df_interp2, cp_dff[concat_cols]], axis=1)
    return cp_df_interp2.round(ROUND_TO)

def get_cp_df_interp(cp_dfs, group, t_min, t_max, add_individual_cp_features=True):
    cp_dfs_interp=[]

    if add_individual_cp_features:
        cp_dfs = _add_individual_cp_features(cp_dfs)

    for cp_dff in cp_dfs:
        cp_dff = resample_feats(cp_dff, t_min[group], t_max[group])
        cp_dfs_interp.append(cp_dff)
        # print(cp_dff["meter_value_sampled_value"].min())
        # print(group, cp_dff)

    cp_df_interp = pd.concat(cp_dfs_interp)
    cp_df_interp = cp_df_interp.groupby(cp_df_interp.index).sum()
    return cp_df_interp.round(ROUND_TO)

def resample_cpo_dfs(cpo_dfs, t_min, t_max, per_cp_group_normalize, per_cp_group_date_exog, add_individual_cp_features=True):
    for group, cp_dfs in tqdm(cpo_dfs.items(), desc="loading cp_df_interp"):

        cp_df_interp = get_cp_df_interp2(cp_dfs, group, t_min, t_max, add_individual_cp_features)
        # cp_df_interp = get_cp_df_interp(cp_dfs, group, t_min, t_max, add_individual_cp_features)
        #print(cp_df_interp["charge_speed"])
        # print(cp_df_interp.columns)
        # print(cp_df_interp2)
        # print(cp_df_interp2.columns)
        # print(cp_df_interp.equals(cp_df_interp2))
        #exit()
        #cp_df_interp.loc[:, "num_current_sessions"] = cp_df_interp[[c for c in cp_df_interp if "event_Updated" in c]].sum(axis=1) # not needed since event_Updated already summed up...

        if per_cp_group_date_exog:
            c1 = cp_df_interp.columns
            get_date_exog(cp_df_interp, prefix="date_exog_group_")
            c_del = [c for c in cp_df_interp.columns if c not in c1]
            for c in cp_df_interp.columns:
                if c.startswith("date_exog_group_"):
                    cp_df_interp = pd.concat([cp_df_interp, pd.get_dummies(cp_df_interp[c], prefix=c)], axis=1)
            cp_df_interp = cp_df_interp[[c for c in cp_df_interp.columns if c not in c_del]]

        if per_cp_group_normalize:
            normalize_cols(cp_df_interp, "_group_norm")
        
        # if len(cp_df_interp.loc[cp_df_interp.isna().any(axis=1), cp_df_interp.isna().any()]) > 1:
        #     print("1",cp_df_interp.loc[cp_df_interp.isna().any(axis=1), cp_df_interp.isna().any()])
        #     exit()
        # print(group, cp_df_interp)
        # break
        yield group, cp_df_interp


def _get_cp_feat_cols(cp_df_interp, group, cpo_feat_cols, keep_only_normalized=False, keep_all_normalized=False, pred_col = "charge_speed_lag_0"):
    keep_lag_0_CPs = ["charge_speed"]
    keep_lag_0_n = ["time_diff", 'event_Ended', 'event_Started', 'event_Updated', "date_exog_cp_", "date_exog_group_"]
    if keep_all_normalized:
        keep_lag_0_n+=["_norm"]
    keep_lag_1_n = ["meter_value_sampled_value", "meter_diff", "charge_speed"]
    
    
    for col in cp_df_interp.columns:

        for s in keep_lag_0_CPs:
            if (s in col and "_lag_" in col and "CP_" in col) and ("custom_data_" not in col and "is_attack" not in col):
                if keep_only_normalized and "_norm" not in col:
                    continue
                cpo_feat_cols[group]["keep_lag_0_n_cols"].add(col)

        for s in keep_lag_0_n:
            if (s in col and "_lag_" in col) and ("custom_data_" not in col and "is_attack" not in col):
                if keep_only_normalized and "_norm" not in col:
                    continue
                cpo_feat_cols[group]["keep_lag_0_n_cols"].add(col)
        for s in keep_lag_1_n:
            if (s in col and "_lag_0" not in col) and ("custom_data_" not in col and "is_attack" not in col):
                if keep_only_normalized and "_norm" not in col:
                    continue
                if col in cpo_feat_cols[group]["keep_lag_0_n_cols"]:
                    continue
                cpo_feat_cols[group]["keep_lag_1_n_cols"].add(col)

    cpo_feat_cols[group]["pred_col"].add(pred_col)

def _get_cpo_feat_cols(cpo_df_interp_l, keep_only_normalized=False):
    cpo_feat_cols = defaultdict(lambda: defaultdict(lambda: set()))
    for group, cp_df_interp in cpo_df_interp_l.items():
        _get_cp_feat_cols(cp_df_interp, group, cpo_feat_cols, keep_only_normalized)
    return cpo_feat_cols

def get_cp_feat_cols(g,t,cpo_group_feat_cols):
    for col_t,cols in t.items():
        if col_t in ["keep_lag_0_n_cols", "keep_lag_1_n_cols"]:
            cpo_group_feat_cols[g]["feat_cols"].extend(sorted(cols))
        elif col_t in ["pred_col"]:
            cpo_group_feat_cols[g]["pred_col"].extend(sorted(cols))

def get_cpo_feat_cols(cpo_df_interp_l):
    cpo_feat_cols = _get_cpo_feat_cols(cpo_df_interp_l)
    cpo_group_feat_cols = defaultdict(lambda: defaultdict(lambda: list()))
    for g,t in cpo_feat_cols.items():
        get_cp_feat_cols(g,t,cpo_group_feat_cols)
    return cpo_group_feat_cols

def get_cpo_group_feat_dfs(cpo_files, per_cp_normalize=False, per_cp_date_exog=True, per_cp_group_normalize=True, per_cp_group_date_exog=True, num_lags=4):
    cpo_dfs, t_min, t_max = get_cpo_dfs(cpo_files, per_cp_normalize=per_cp_normalize, per_cp_date_exog=per_cp_date_exog)
    # cpo_df_interp = {group: cp_df_interp for group, cp_df_interp in resample_cpo_dfs(cpo_dfs, t_min, t_max, per_cp_group_normalize=per_cp_group_normalize, per_cp_group_date_exog=per_cp_group_date_exog)}
    # cpo_df_interp_l = {group: cp_df_interp_l for group, cp_df_interp_l in add_all_lags(cpo_df_interp, num_lags=num_lags)}
    # cpo_feat_cols = get_cpo_feat_cols(cpo_df_interp_l)
    pred_col_init = "charge_speed"
    pred_col = pred_col_init+"_lag_0"


    for group, cp_df_interp in resample_cpo_dfs(cpo_dfs, t_min, t_max, per_cp_group_normalize=per_cp_group_normalize, per_cp_group_date_exog=per_cp_group_date_exog):
        cp_df_interp_l = add_lags(cp_df_interp, num_lags=num_lags, only_do_col=pred_col_init)
        _cpo_feat_cols = defaultdict(lambda: defaultdict(lambda: set()))
        _get_cp_feat_cols(cp_df_interp_l, group, _cpo_feat_cols, pred_col=pred_col)
        cpo_feat_cols = defaultdict(lambda: defaultdict(lambda: list()))
        for g,t in _cpo_feat_cols.items():
            get_cp_feat_cols(g,t,cpo_feat_cols)
        yield {group: {"cp_g_df": cp_df_interp_l, 
                        "feat_cols": cpo_feat_cols[group]["feat_cols"], 
                        "pred_col": cpo_feat_cols[group]["pred_col"][0]}}



    # cpo_df_interp_l_feat = {group: {"cp_g_df": cp_df_interp_l, 
    #                                 "feat_cols": cpo_feat_cols[group]["feat_cols"], 
    #                                 "pred_col": cpo_feat_cols[group]["pred_col"][0]} for group, cp_df_interp_l in cpo_df_interp_l.items()}
    # return cpo_df_interp_l_feat
    # for group, cp_df_interp_l in cpo_df_interp_l.items():
    #     print(group, cp_df_interp_l)
    #     print([c for c in cp_df_interp_l.columns if "_norm" not in c and "_lag_0" in c])
    #     """['meter_value_sampled_value', 'custom_data_meter_no_atk', 'custom_data_scaling', 'custom_data_energy_interval', 'custom_data_original_energy_interval', 
    #         'custom_data_average_power', 'custom_data_original_average_power', 'meter_diff', 'time_diff', 'charge_speed', 'custom_data_meter_diff', 'custom_data_charge_speed', 
    #         'event_Ended', 'event_Started', 'event_Updated', 'is_attack', 
    #         'date_exog_cp_dayofweek', 'date_exog_cp_hour', 'date_exog_cp_dayofweek_4', 'date_exog_cp_hour_8', 
    #         'date_exog_cp_hour_9', 'date_exog_cp_hour_10', 'date_exog_cp_hour_11', 'date_exog_cp_hour_12', 'date_exog_cp_hour_13', 'date_exog_cp_hour_14', 
    #         'date_exog_cp_discretize_hour_only_Play', 'date_exog_cp_discretize_hour_only_Work', 'date_exog_cp_discretize_hour_day_High-Work', 'date_exog_cp_discretize_day_is_work_True', 
    #         'date_exog_cp_discretize_hour_balancing_low', 'date_exog_cp_discretize_hour_balancing_peak', 'date_exog_cp_dayofweek_3', 'date_exog_cp_hour_15', 'date_exog_cp_dayofweek_0', 
    #         'date_exog_cp_hour_16', 'date_exog_cp_dayofweek_5', 'date_exog_cp_hour_0', 'date_exog_cp_hour_1', 'date_exog_cp_hour_2', 'date_exog_cp_hour_3', 'date_exog_cp_hour_4',
    #         'date_exog_cp_hour_5', 'date_exog_cp_hour_6', 'date_exog_cp_hour_18', 'date_exog_cp_hour_19', 'date_exog_cp_hour_20', 'date_exog_cp_hour_21', 'date_exog_cp_hour_22', 
    #         'date_exog_cp_hour_23', 'date_exog_cp_discretize_hour_only_Sleep', 'date_exog_cp_discretize_hour_day_High-Home', 'date_exog_cp_discretize_day_is_work_False', 'date_exog_cp_hour_7', 
    #         'date_exog_cp_hour_17', 'date_exog_cp_dayofweek_1', 
    #         'date_exog_group_dayofweek_0', 'date_exog_group_dayofweek_1', 'date_exog_group_dayofweek_3', 'date_exog_group_dayofweek_4', 
    #         'date_exog_group_dayofweek_5', 'date_exog_group_dayofweek_6', 'date_exog_group_hour_0', 'date_exog_group_hour_1', 'date_exog_group_hour_2', 'date_exog_group_hour_3',
    #         'date_exog_group_hour_4', 'date_exog_group_hour_5', 'date_exog_group_hour_6', 'date_exog_group_hour_7', 'date_exog_group_hour_8', 'date_exog_group_hour_9', 'date_exog_group_hour_10', 
    #         'date_exog_group_hour_11', 'date_exog_group_hour_12', 'date_exog_group_hour_13', 'date_exog_group_hour_14', 'date_exog_group_hour_15', 'date_exog_group_hour_16', 'date_exog_group_hour_17', 
    #         'date_exog_group_hour_18', 'date_exog_group_hour_19', 'date_exog_group_hour_20', 'date_exog_group_hour_21', 'date_exog_group_hour_22', 'date_exog_group_hour_23', 
    #         'date_exog_group_discretize_hour_only_Play', 'date_exog_group_discretize_hour_only_Sleep', 'date_exog_group_discretize_hour_only_Work', 'date_exog_group_discretize_hour_day_High-Home', 
    #         'date_exog_group_discretize_hour_day_High-Work', 'date_exog_group_discretize_day_is_work_False', 'date_exog_group_discretize_day_is_work_True', 'date_exog_group_discretize_hour_balancing_low', 
    #         'date_exog_group_discretize_hour_balancing_peak']"""
    #     break #TODO no
    # exit()
    #TODO dont cheat on features, eg only prev speed but not current...
        