
from collections import defaultdict
from datetime import datetime
import os
import warnings
from matplotlib import pyplot as plt
import pandas as pd
from tqdm import tqdm

import numpy as np

import re

import logging

from features_aux import get_date_exog, get_date_exog_col, get_elems_attatched_to_bus, get_grid_pp, load_single_feat, normalize_cols, add_all_lags
from features_clf import _interp


logger = logging.getLogger("WATTSON_EV_IDS.FeaturesDSO")

ROUND_TO=6

def get_dso_files(x_path):
    onlyfiles = [f for f in os.listdir(x_path) if os.path.isfile(os.path.join(x_path, f))]
    onlyfiles = [f for f in onlyfiles if f.startswith("dso_oscp_data_") and f.endswith(".csv.gz")]
    return onlyfiles

def read_dso_files(dir, files):
    # print(dir, files)
    regex = re.compile(r"^dso_oscp_data_(CP_\d+).csv.gz$")
    dso_files=dict()
    for f in tqdm(files, desc="loading read_dso_files"):
        #dso_oscp_data_CP_1.csv.gz
        #sim_time	oscp_interval_h	group_id	measurements_value	measurements_phase	measurements_unit	measurements_energy_type	measurements_direction	measurements_measure_time	measurements_initial_measure_time	meter_diff	time_diff	charge_speed
        result = regex.search(f)
        if result:
            f_df = pd.read_csv(os.path.join(dir, f), index_col="sim_time", parse_dates=["sim_time", "measurements_measure_time", "measurements_initial_measure_time"])
            if result.group(1) != f_df.iloc[0]["group_id"]:
                logger.error(f"wrong cp_group {result.group(1)} != {f_df.iloc[0]['group_id']}")
            # f_df = f_df[[c for c in f_df.columns if not c.startswith("custom_data") and c != "cp_c_id" and c != "cp_group"]]
            dso_files[result.group(1)] = f_df
            # print(f_df)
            # break
        else:
            logger.error(f"unk file {f}")
            continue
    return dso_files

def read_dso_estimations(file):
    m_df = pd.read_csv(file, parse_dates=["Unnamed: 0"])
    m_df.rename(columns={"Unnamed: 0":"sim_time"}, inplace=True)
    m_df["sim_time"] = m_df["sim_time"].dt.tz_localize(None)
    m_df = m_df.set_index("sim_time", drop=True)
    return m_df

def read_dso_measurements(file):
    try:
        m_df = pd.read_csv(file, parse_dates=["sim_time"])
    except ValueError as e:
        logger.info(f"{e} in read_dso_measurements... trying Unnamed: 0")
        m_df = pd.read_csv(file, parse_dates=["Unnamed: 0"])
        m_df.rename(columns={"Unnamed: 0":"sim_time"}, inplace=True)
        logger.info(f"Unnamed: 0 worked.")
    m_df["sim_time"] = m_df["sim_time"].dt.tz_localize(None)
    m_df = m_df.set_index("sim_time", drop=True)
    return m_df


def get_features_dso(df, per_cp_group_normalize, per_cp_group_date_exog):
    # ['oscp_interval_h', 'group_id', 'measurements_value',
    #    'measurements_phase', 'measurements_unit', 'measurements_energy_type',
    #    'measurements_direction', 'measurements_measure_time',
    #    'measurements_initial_measure_time', 'meter_diff', 'time_diff',
    #    'charge_speed']
    #print(df)
    # exit()
    keep_cols = ["charge_speed", "measurements_value", "meter_diff", "time_diff"]
    f_df = df[[c for c in df.columns if c != "oscp_interval_h" and c != "group_id" and c != "measurements_phase" and c != "measurements_unit" and c != "measurements_energy_type" and c != "measurements_direction"]].copy()
    #f_df = pd.concat([f_df, pd.get_dummies(f_df["event"], prefix="event")], axis=1)
    f_df["time_diff"] = f_df["time_diff"].fillna(0)

    if per_cp_group_date_exog:
        c1 = f_df.columns
        get_date_exog(f_df, prefix="date_exog_cp_g_")
        get_date_exog_col(f_df, col="measurements_measure_time", prefix="date_exog_cp_g_")
        get_date_exog_col(f_df, col="measurements_initial_measure_time", prefix="date_exog_cp_g_")
        c_del = [c for c in f_df.columns if c not in c1]
        for c in f_df.columns:
            if c.startswith("date_exog_cp_g_"):
                f_df = pd.concat([f_df, pd.get_dummies(f_df[c], prefix=c)], axis=1)
        f_df = f_df[[c for c in f_df.columns if c not in c_del]] #del non dummy date exog
        keep_cols+= [c for c in f_df.columns if c.startswith("date_exog_cp_g_")]

    f_df = f_df[[c for c in f_df.columns if c in keep_cols]]

    if per_cp_group_normalize:
        normalize_cols(f_df, "_cp_g_norm")
    return f_df.copy()

def resample_feats(df, t_min, t_max, freq='15Min'):
    oidx = df.index
    #nidx = pd.date_range(oidx.min(), oidx.max(), freq='5Min')
    nidx = pd.date_range(t_min, t_max, freq=freq)
    res = df.reindex(oidx.union(nidx)).interpolate('time', limit_direction="both").reindex(nidx)
    return res


def get_dso_dfs(dso_files, per_cp_group_normalize, per_cp_group_date_exog):
    dso_dfs=dict()
    t_min=dict()
    t_max=dict()
    for group, cpg_df in tqdm(dso_files.items(), desc="loading dso_dfs"):
        cp_dff = get_features_dso(cpg_df, per_cp_group_normalize=per_cp_group_normalize, per_cp_group_date_exog=per_cp_group_date_exog)
        if group not in t_min:
            t_min[group] = cp_dff.index.min()
            t_max[group] = cp_dff.index.max()
        else:
            t_min[group] = min(t_min[group], cp_dff.index.min())
            t_max[group] = max(t_max[group], cp_dff.index.max())
        
        dso_dfs[group] = cp_dff
    return dso_dfs, t_min, t_max


def get_group_meas_est_cols(pp, group, cols):
    g_regex = re.compile(r"^CP_(\d+)$")
    res = g_regex.search(group)
    col_n = "bus."+res.group(1)+"."
    group_meas_cols = [c for c in cols if col_n in c]

    if pp is not None:
        conn_cols = get_elems_attatched_to_bus(pp, int(res.group(1)))
        logger.debug(f"get_group_meas_est_cols adding: {conn_cols} for {col_n}")
        for col_c in conn_cols:
            col_c=col_c+"."
            group_meas_cols += [c for c in cols if col_c in c and (("line" in col_c and not "from_" in c) or ("line" not in col_c))]
            group_meas_cols = sorted(set(group_meas_cols))
        logger.debug(f"get_group_meas_est_cols new: {group_meas_cols}")
    else:
        logger.warning(f"empty pp grid in get_group_meas_est_cols")
    return group_meas_cols

def resample_dso_dfs(pp, dso_dfs, dso_meas_df, dso_est_df, t_min, t_max, post_interp_normalize, post_interp_date_exog, add_meas_data=True, add_est_data=False, further_meas_export_data=None, pred_col="charge_speed"):
    g_regex = re.compile(r"^CP_(\d+)$")
    for group, cp_dff in tqdm(dso_dfs.items(), desc="loading cp_df_interp"):
        cp_df_interp = resample_feats(cp_dff, t_min[group], t_max[group])

        if add_meas_data:
            group_meas_cols = get_group_meas_est_cols(pp, group, dso_meas_df.columns)
            dso_meas_df_g = dso_meas_df[group_meas_cols].copy()
            dso_meas_df_g.rename(columns={c:"grid_meas_"+c for c in group_meas_cols}, inplace=True)
            dso_meas_df_interp = resample_feats(dso_meas_df_g, t_min[group], t_max[group])
            # print([c for c in dso_meas_df_interp.columns])
            # print(dso_meas_df_interp["grid_meas_849.10030.bus.8.active_power"])
            # print(cp_df_interp[pred_col])
            cp_df_interp = pd.concat([cp_df_interp, dso_meas_df_interp], axis=1)
            for c in dso_meas_df_interp.columns:
                if ".bus." in c and ".active_power" in c: #should be only one
                    cp_df_interp.loc[:,c+"_relation_to_"+pred_col] = cp_df_interp[pred_col]/cp_df_interp[c]

                    cp_df_interp.loc[:, c.replace("bus.", "bus_and_sgen.")] = cp_df_interp[c]
                    for c2 in dso_meas_df_interp.columns:
                        if ".sgen." in c2 and ".active_power" in c2:
                            cp_df_interp.loc[:, c.replace("bus.", "bus_and_sgen.")] += cp_df_interp[c2]
                    cp_df_interp.loc[:, c.replace("grid_meas_", "grid_expo_").replace("bus.", "no_ev_load.")] = cp_df_interp.loc[:, c.replace("bus.", "bus_and_sgen.")] - cp_df_interp[pred_col]
                    cp_df_interp.loc[:, c.replace("grid_meas_", "grid_expo_").replace("bus.", "no_ev_load.") +"_relation_to_"+ pred_col] = cp_df_interp[pred_col] / cp_df_interp.loc[:, c.replace("grid_meas_", "grid_expo_").replace("bus.", "no_ev_load.")]
                    cp_df_interp.loc[:, c.replace("grid_meas_", "grid_expo_").replace("bus.", "no_ev_load.") +"_relation_to_bus"] =  cp_df_interp.loc[:, c.replace("grid_meas_", "grid_expo_").replace("bus.", "no_ev_load.")] / cp_df_interp[c]
        

        if further_meas_export_data is not None:
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', pd.errors.PerformanceWarning)
                group_int = int(g_regex.search(group).group(1))
                for exp_data in further_meas_export_data[group_int]:
                    exp_data_interp = resample_feats(exp_data, t_min[group], t_max[group])
                    rng = np.random.default_rng(seed=group_int)
                    for x in range(10,110,10):
                        rnd_fac = rng.integers(low=x, high=min(x+11, 101), size=None)
                        rnd_fac_s = rnd_fac * 0.01
                        #print(rnd_fac_s)
                        exp_data_interp_fac=exp_data_interp*rnd_fac_s
                        exp_data_interp_fac.name = exp_data_interp.name+"_static_fac."+str(x) #same random fac over all measurements of a load (diff for diff loads)
                        cp_df_interp = pd.concat([cp_df_interp, exp_data_interp_fac], axis=1)
                    for x in range(10,110,10):
                        rnd_fac = rng.integers(low=x, high=min(x+11, 101), size=exp_data_interp.size)
                        rnd_fac_s = pd.Series(rnd_fac, index = exp_data_interp.index)
                        rnd_fac_s = rnd_fac_s * 0.01
                        exp_data_interp_fac=exp_data_interp*rnd_fac_s
                        exp_data_interp_fac.name = exp_data_interp.name+"_rnd_fac."+str(x) #diff random fac over all measurements of same load
                        cp_df_interp = pd.concat([cp_df_interp, exp_data_interp_fac], axis=1)
                    for c in [col for col in cp_df_interp.columns]: #grid_expo_load.5.MEASUREMENT.active_power_static_fac.80
                        if ".load." in c and ".active_power" in c and "_relation_to_" not in c: #should be only one
                            cp_df_interp.loc[:,c+"_relation_to_"+pred_col] = cp_df_interp[pred_col]/cp_df_interp[c]
                            for c2 in dso_meas_df_interp.columns:
                                if ".bus." in c2 and ".active_power" in c2 and "_relation_to_" not in c2: #should be only one
                                    cp_df_interp.loc[:,c+"_relation_to_"+c2] = cp_df_interp[c2]/cp_df_interp[c]
                    cp_df_interp = cp_df_interp.copy()
        else:
            group_int = int(g_regex.search(group).group(1))
            rng = np.random.default_rng(seed=group_int)
            load_cols=[]
            for c in [col for col in cp_df_interp.columns]: #grid_meas_668.10340.load.11.active_power
                if ".load." in c and ".active_power" in c and "_relation_to_" not in c: #should be only one or two
                    load_cols.append(c)
            load_cols_new = {l:l.replace("grid_meas_", "grid_expo_") for l in load_cols}
            cp_df_interp = cp_df_interp.rename(columns=load_cols_new)
            load_cols = [v for v in load_cols_new.values()]
            for lc in load_cols:
                exp_data_interp = cp_df_interp[lc]
                #for x in range(10,110,10):
                for x in [100]:
                    rnd_fac = rng.integers(low=x, high=min(x+11, 101), size=None)
                    rnd_fac_s = rnd_fac * 0.01
                    #print(rnd_fac_s)
                    exp_data_interp_fac=exp_data_interp*rnd_fac_s
                    exp_data_interp_fac.name = exp_data_interp.name+"_static_fac."+str(x) #same random fac over all measurements of a load (diff for diff loads)
                    cp_df_interp = pd.concat([cp_df_interp, exp_data_interp_fac], axis=1)
                for x in range(2,21,2):
                    index_4 = exp_data_interp.index.values[::4]
                    rnd_fac4 = rng.uniform(low=-1*x, high=x, size=len(index_4))
                    rnd_fac_s4 = pd.Series(rnd_fac4, index = index_4)
                    rnd_fac_s1 = _interp(rnd_fac_s4, exp_data_interp.index, col=None)
                    rnd_fac_s = rnd_fac_s1 * 0.01
                    exp_data_interp_fac=exp_data_interp*(1+rnd_fac_s)
                    exp_data_interp_fac.name = exp_data_interp.name+"_rnd_fac."+str(x) #diff random fac over all measurements of same load
                    cp_df_interp = pd.concat([cp_df_interp, exp_data_interp_fac], axis=1)
                    # print(cp_df_interp)
                    # exit()

                for x in [90]: #TODO: deprecated... delete
                    rnd_fac = rng.integers(low=x, high=min(x+11, 101), size=exp_data_interp.size)
                    rnd_fac_s = pd.Series(rnd_fac, index = exp_data_interp.index)
                    rnd_fac_s = rnd_fac_s * 0.01
                    exp_data_interp_fac=exp_data_interp*rnd_fac_s
                    exp_data_interp_fac.name = exp_data_interp.name+"_rnd_fac."+str(x) #diff random fac over all measurements of same load
                    cp_df_interp = pd.concat([cp_df_interp, exp_data_interp_fac], axis=1)

                for c in [col for col in cp_df_interp.columns]: #grid_expo_load.5.MEASUREMENT.active_power_static_fac.80
                    if ".load." in c and ".active_power" in c and "_relation_to_" not in c and ("_static_fac." in c or "_rnd_fac." in c): #should be only one
                        cp_df_interp.loc[:,c+"_relation_to_"+pred_col] = cp_df_interp[pred_col]/cp_df_interp[c]
                        for c2 in dso_meas_df_interp.columns:
                            if ".bus." in c2 and ".active_power" in c2 and "_relation_to_" not in c2: #should be only one
                                cp_df_interp.loc[:,c+"_relation_to_"+c2] = cp_df_interp[c2]/cp_df_interp[c]
                cp_df_interp = cp_df_interp.copy()

        if add_est_data and dso_est_df is not None:
            group_est_cols = get_group_meas_est_cols(pp, group, dso_est_df.columns)
            dso_est_df_g = dso_est_df[group_est_cols].copy()
            dso_est_df_g.rename(columns={c:"grid_est_"+c for c in group_est_cols}, inplace=True)
            dso_est_df_interp = resample_feats(dso_est_df_g, t_min[group], t_max[group])
            cp_df_interp = pd.concat([cp_df_interp, dso_est_df_interp], axis=1)


        # print(group, [c for c in cp_df_interp.columns if "bus" in c])
        cp_df_interp = cp_df_interp.groupby(cp_df_interp.index).sum()
        cp_df_interp = cp_df_interp.round(ROUND_TO)
        if post_interp_date_exog:
            c1 = cp_df_interp.columns
            get_date_exog(cp_df_interp, prefix="date_exog_group_")
            c_del = [c for c in cp_df_interp.columns if c not in c1]
            for c in cp_df_interp.columns:
                if c.startswith("date_exog_group_"):
                    cp_df_interp = pd.concat([cp_df_interp, pd.get_dummies(cp_df_interp[c], prefix=c)], axis=1)
            cp_df_interp = cp_df_interp[[c for c in cp_df_interp.columns if c not in c_del]]

        if post_interp_normalize:
            normalize_cols(cp_df_interp, "_group_norm")
        
        # if len(cp_df_interp.loc[cp_df_interp.isna().any(axis=1), cp_df_interp.isna().any()]) > 1:
        #     print("1",cp_df_interp.loc[cp_df_interp.isna().any(axis=1), cp_df_interp.isna().any()])
        #     exit()
        # print(group, [c for c in cp_df_interp.columns if "bus" in c])
        # exit()
        # break
        yield group, cp_df_interp

def oscp_speed_workaround(cpo_group_speed, oscp_group_speed,  pred_col = "charge_speed_lag_0"):
    oidx = cpo_group_speed.index
    #nidx = pd.date_range(oidx.min(), oidx.max(), freq='5Min')
    nidx = oscp_group_speed.index
    res = cpo_group_speed[pred_col].reindex(oidx.union(nidx)).interpolate('time', limit_direction="both").reindex(nidx)
    return res

# ['measurements_value_lag_0', 'meter_diff_lag_0', 'time_diff_lag_0', 'charge_speed_lag_0', 
#  'date_exog_cp_g_...', 
#  'date_exog_group_...']
def _get_dso_feat_cols(pp,dso_df_interp_l, keep_only_normalized=False, pred_col = "charge_speed_lag_0"):
    # print(dso_df_interp_l["CP_5"])
    # print([c for c in dso_df_interp_l["CP_5"].columns if "_norm" not in c and "_lag_0" in c])
    # exit()
    keep_lag_0_n = ["time_diff", "date_exog_cp_g_", "date_exog_group_"]
    keep_lag_1_n = ["measurements_value", "meter_diff", "charge_speed"]
    
    
    dso_feat_cols = defaultdict(lambda: defaultdict(lambda: set()))
    for group, cp_df_interp in dso_df_interp_l.items():
        keep_lag_0_n += [c.replace("_lag_0","") for c in get_group_meas_est_cols(pp, group, cp_df_interp.columns) if ("grid_meas_" in c or "grid_est_" in c) and "grid_expo_" not in c and "load." not in c] #exclude grid_expo_ (for load active_power) by default
        # print([c.replace("_lag_0","") for c in get_group_meas_est_cols(pp, group, cp_df_interp.columns) if "grid_meas_" in c or "grid_est_" in c])
        # print(get_group_meas_est_cols(pp, group, cp_df_interp.columns))
        # exit()
        for col in cp_df_interp.columns:
            for s in keep_lag_0_n:
                if (s in col and "_lag_" in col) and ("custom_data_" not in col and "is_attack" not in col and "grid_expo_" not in col):
                    if keep_only_normalized and "_norm" not in col:
                        continue
                    dso_feat_cols[group]["keep_lag_0_n_cols"].add(col)
            for s in keep_lag_1_n:
                if (s in col and "_lag_0" not in col) and ("custom_data_" not in col and "is_attack" not in col and "grid_expo_" not in col):
                    if keep_only_normalized and "_norm" not in col:
                        continue
                    dso_feat_cols[group]["keep_lag_1_n_cols"].add(col)

        dso_feat_cols[group]["pred_col"].add(pred_col)
    return dso_feat_cols

def get_dso_feat_cols(pp,dso_df_interp_l, pred_col = "charge_speed_lag_0"):
    dso_feat_cols = _get_dso_feat_cols(pp,dso_df_interp_l, pred_col = pred_col)
    dso_group_feat_cols = defaultdict(lambda: defaultdict(lambda: list()))
    for g,t in dso_feat_cols.items():
        for col_t,cols in t.items():
            if col_t in ["keep_lag_0_n_cols", "keep_lag_1_n_cols"]:
                dso_group_feat_cols[g]["feat_cols"].extend(sorted(cols))
            elif col_t in ["pred_col"]:
                dso_group_feat_cols[g]["pred_col"].extend(sorted(cols))
    return dso_group_feat_cols

def get_dso_group_feat_dfs(DIR, dso_files, dso_meas_df, dso_est_df, per_cp_group_normalize=False, per_cp_group_date_exog=True, post_interp_normalize=True, post_interp_date_exog=True, num_lags=4):
    pp = get_grid_pp(DIR)
    pred_col_init = "charge_speed"
    pred_col = pred_col_init+"_lag_0"

    if False:
        further_meas_export_data = get_grid_measurements_from_export(DIR, pp, bus_ns=[], target_elem="load", target_var=".MEASUREMENT.active_power" )
    else:
        further_meas_export_data=None

    dso_dfs, t_min, t_max = get_dso_dfs(dso_files, per_cp_group_normalize=per_cp_group_normalize, per_cp_group_date_exog=per_cp_group_date_exog)
    dso_df_interp = {group: cp_df_interp for group, cp_df_interp in resample_dso_dfs(pp, dso_dfs, dso_meas_df, dso_est_df, t_min, t_max, post_interp_normalize=post_interp_normalize, further_meas_export_data=further_meas_export_data, post_interp_date_exog=post_interp_date_exog)}
    dso_df_interp_l = {group: cp_df_interp_l for group, cp_df_interp_l in add_all_lags(dso_df_interp, num_lags=num_lags, only_do_col=pred_col_init)}
    dso_feat_cols = get_dso_feat_cols(pp,dso_df_interp_l, pred_col=pred_col)

    dso_df_interp_l_feat = {group: {"cp_g_df": cp_df_interp_l, 
                                    "feat_cols": dso_feat_cols[group]["feat_cols"], 
                                    "pred_col": dso_feat_cols[group]["pred_col"][0]} for group, cp_df_interp_l in dso_df_interp_l.items()}
    return dso_df_interp_l_feat

       