
from collections import defaultdict
from copy import deepcopy
import gzip
import json
import pickle
import os
from typing import Dict, List, Tuple
import pandas as pd
from tqdm import tqdm
import yaml
import matplotlib.pyplot as plt

import ast
import re

import logging

from features_aux import get_grid_measurements_from_export, get_grid_pp

logger = logging.getLogger("WATTSON_EV_IDS.Load_Data")

# os.environ["QT_QPA_PLATFORM"] = "wayland"
#os.environ["QT_QPA_PLATFORM"] = "xcb"
# py load_data.py -c=print_data -v test



START_DATE="2023-11-01 22:13:00"


def get_dataset_l(configs):
    #atk_1_0.2_powerowl_example
    atk_regex = re.compile(r"^.*?\/(atk_[\d\.]+_[\d\.]+)_[\w\d_]+$")

    clean_dsets=[]
    for dataset in list(configs.keys()):
        if "atk" in dataset:
            clean_dsets.append(dataset)
            for subdir in configs[dataset]["DIR"]:
                result = atk_regex.search(subdir)
                if result:
                    new_dset=dataset.replace("_atk","")+"_"+result.group(1)
                    clean_dsets.append(new_dset)
                    configs[new_dset] = deepcopy(configs[dataset])
                    configs[new_dset]["DIR"]=subdir
                    configs[new_dset]["OutDataDIR"] = "data/"+new_dset
                else:
                    logger.error(f"get_dataset_l unk atk dir {subdir}")
        else:
            clean_dsets.append(dataset)
    return clean_dsets


def clean_dataset_l(datasets, configs, update_config=False):
    #atk_1_0.2_powerowl_example
    atk_regex = re.compile(r"^.*?\/(atk_[\d\.]+_[\d\.]+)_[\w\d_]+$")

    clean_dsets=[]
    for dataset in datasets:
        if dataset.endswith("_atk"):
            for subdir in configs[dataset]["DIR"]:
                result = atk_regex.search(subdir)
                if result:
                    new_dset=dataset.replace("_atk","")+"_"+result.group(1)
                    clean_dsets.append(new_dset)
                    if update_config:
                        configs[new_dset] = deepcopy(configs[dataset])
                        configs[new_dset]["DIR"]=subdir
                        configs[new_dset]["OutDataDIR"] = "data/"+new_dset
                else:
                    logger.error(f"clean_dataset_l unk atk dir {subdir}")
        else:
            clean_dsets.append(dataset)
    return clean_dsets

def print_data_dict(file_dict):
    for d_type, x_path_dict in file_dict.items():
        logger.info(d_type)
        for x_path, f_list in x_path_dict.items():
            logger.info("\t"+x_path)
            logger.info(f"\t\t{len(f_list)}")

def print_data_df(file_df):
    logger.info(file_df)


def get_file_dfs(datasets, configs) -> Dict[str, pd.DataFrame]:
    file_dict = {}
    file_dfs = {}
    for dataset in datasets:
        DIR = configs[dataset]["DIR"]
        file_dict[dataset] = []
        # file_dict[dataset]["controller-export"] = {}
        # file_dict[dataset]["logs"] = {}

        for x in ["estimation", "measurements", "power-grid"]:
            x_path = os.path.join(DIR, "controller-export", x)
            onlyfiles = [f for f in os.listdir(x_path) if os.path.isfile(os.path.join(x_path, f))]
            #file_dict[dataset]["controller-export"][x] = onlyfiles
            for f in onlyfiles:
                file_dict[dataset].append({"type":"controller-export", "sub_type":x, "x_path":x_path, "file":f})

        node_regex = re.compile(r"^n\d+$")
        sub_folders = [name for name in os.listdir(DIR) if os.path.isdir(os.path.join(DIR, name))]
        for x in [f for f in sub_folders if f.startswith("CPO_") or f.startswith("NodeCP_") or node_regex.match(f)]:
            x_path = os.path.join(DIR, x)
            #/home/dk/git/wattson-artifacts/test_powerowl_example_2023-12-22-18-22-51/CPO_0/CPO_0-service-34.log
            onlyfiles = [f for f in os.listdir(x_path) if os.path.isfile(os.path.join(x_path, f))]
            #file_dict[dataset]["logs"][x] = onlyfiles
            for f in onlyfiles:
                file_dict[dataset].append({"type":"logs", "sub_type":x, "x_path":x_path, "file":f})
        
        file_dfs[dataset] = pd.DataFrame().from_dict(file_dict[dataset])
    return file_dfs


def get_cpo_ocpp_data(file_df) ->  Dict[str, Dict[str, List[Tuple[str, pd.DataFrame]]]]:
    regex = re.compile(r"^\d{1,4}-\d{1,2}-\d{1,2} \d{1,2}:\d{1,2}:\d{1,2},\d{1,3} - CPOService - INFO - [\d\.]+ -  on_tx (.*?)$")
    cpo_tx_df_cps={}
    for cpo_id in [s for s in file_df["sub_type"].drop_duplicates() if s.startswith("CPO_")]:
        cpo_log = file_df[(file_df["type"] == "logs") & (file_df["sub_type"] == cpo_id) & (file_df["file"] == "CPO_0-service-34.log")][["x_path", "file"]]
        cpo_log_file = os.path.join(*cpo_log.iloc[0].values)
        logger.info(cpo_log_file)

        num_lines = sum(1 for line in open(cpo_log_file,'r'))
        tx_data = []
        with open(cpo_log_file) as f:
            for line in tqdm(f, total=num_lines, desc=f"Loading {cpo_id}"):
                if "ERROR" in line or "Traceback" in line:
                    logger.error(f"found error in {cpo_log_file}:\t{line}")
                result = regex.search(line)
                #2023-12-23 01:39:29,777 - CPOService - INFO - 140672877031424.35739 -  on_tx ('8ada5dedf167ddcb685bcc4e679570651', 'CP_14', '2023-01-01T22:57:07.401781', '2023-01-01T23:56:54.334702', [{'timestamp': '2023-01-01T23:45:00', 'sampled_value': [{'value': 659575}]}], {'vendor_id': 'wattson_v2g', 'atk_type': 'None'})
                if result:
                    try:
                        #('8ada5dedf167ddcb685bcc4e679570651', 'CP_14', '2023-01-01T22:57:07.401781', '2023-01-01T23:56:54.334702', [{'timestamp': '2023-01-01T23:45:00', 'sampled_value': [{'value': 659575}]}], {'vendor_id': 'wattson_v2g', 'atk_type': 'None'})
                        #('1e9ce1f6770ed662e8810c827bf8f2611', 'CP_11', '2023-11-09T18:01:58.949633', '2023-11-09T18:00:02.158313', 'Updated', [{'timestamp': '2023-11-09T18:00:00', 'sampled_value': [{'value': 8905865}]}], {'vendor_id': 'wattson_v2g', 'atk_type': 'None'})
                        g = result.group(1)
                        g_lit = ast.literal_eval(g)
                    except ValueError as e:
                        if "nan" in g or "inf" in g:
                        #     g = g.replace("nan", "0")
                        #     g_lit = ast.literal_eval(g)
                        # else:
                            logger.warning(f"nan or inf in {result.group(0)}")
                            continue
                        else: raise e

                    tx_msg = {
                        "cp_c_id": g_lit[0],
                        "cp_group": g_lit[1],
                        "sim_time": g_lit[2],
                        "timestamp": g_lit[3], #tx msg timestamp
                        "event": g_lit[4], #
                        "meter_value_ts": g_lit[5][0]["timestamp"], #meter read timestamp
                        "meter_value_sampled_value": g_lit[5][0]["sampled_value"][0]["value"],
                        # "custom_data_vendor_id": g_lit[6]['vendor_id'],
                        # "custom_data_atk_type": g_lit[6]['atk_type'],
                    }
                    for k,v in g_lit[6].items():
                        #{'vendor_id': 'wattson_v2g', 'atk_type': 'MAD', 'meter_no_atk': 7285447, 'energy_interval': 2.539881115023535, 'original_energy_interval': 2.43, 'average_power': 10.137, 'original_average_power': 9.69845}
                        tx_msg["custom_data_"+k] = v
                    tx_data.append(tx_msg)
                        
        tx_df = pd.DataFrame(tx_data)
        logger.info(tx_df.iloc[0:])
        tx_df["sim_time"] = pd.to_datetime(tx_df["sim_time"])
        tx_df["timestamp"] = pd.to_datetime(tx_df["timestamp"])
        tx_df["meter_value_ts"] = pd.to_datetime(tx_df["meter_value_ts"])
        #with pd.option_context('display.max_rows', 100, 'display.max_columns', 9):
        logger.info(tx_df.iloc[0:])
        logger.debug(tx_df.columns)

        tx_df_cps=defaultdict(lambda: [])
        for cp in tqdm(tx_df["cp_c_id"].drop_duplicates(), desc="Loading cp_c_id"):
            # logger.debug(cp)
            tx_df_cp = tx_df[tx_df["cp_c_id"] == cp].copy()

            tx_df_cp = tx_df_cp.sort_values(by="meter_value_ts", ascending=True)
            tx_df_cp["meter_diff"] = tx_df_cp["meter_value_sampled_value"] - tx_df_cp["meter_value_sampled_value"].shift(1)
            tx_df_cp["meter_diff"].fillna(0, inplace=True)
            tx_df_cp["time_diff"] = (tx_df_cp["meter_value_ts"] - tx_df_cp["meter_value_ts"].shift(1)).apply(lambda x: x.total_seconds())
            tx_df_cp["time_diff"] = tx_df_cp["time_diff"] / (60*60) #s to h
            tx_df_cp["charge_speed"] = tx_df_cp["meter_diff"] / tx_df_cp["time_diff"] #wh in w
            tx_df_cp["charge_speed"].fillna(0, inplace=True)

            tx_df_cp["custom_data_meter_diff"] = tx_df_cp["custom_data_meter_no_atk"] - tx_df_cp["custom_data_meter_no_atk"].shift(1)
            tx_df_cp["custom_data_meter_diff"].fillna(0, inplace=True)
            tx_df_cp["custom_data_charge_speed"] = tx_df_cp["custom_data_meter_diff"] / tx_df_cp["time_diff"] #wh in w
            tx_df_cp["custom_data_charge_speed"].fillna(0, inplace=True)
            for c in ["custom_data_energy_interval",  "custom_data_original_energy_interval",  "custom_data_average_power",  "custom_data_original_average_power"]:
                tx_df_cp[c].fillna(0, inplace=True)

            # logger.debug(tx_df_cp)
            # logger.debug(tx_df_cp[tx_df_cp["charge_speed"] < 0])
            # exit()

            tx_df_cp = tx_df_cp.set_index("meter_value_ts")
            tx_df_cp = tx_df_cp.sort_index(ascending=True)
            if len(tx_df_cp[tx_df_cp["charge_speed"] < 0]) > 0:
                logger.warning(f"tx_df_cp {cp} w/ <0 speed")
                logger.warning(tx_df_cp)
                logger.warning(tx_df_cp[tx_df_cp["charge_speed"] < 0])
                logger.warning(tx_df_cp.index[0])
                exit()
                
            if False: #TODO: START_DATE?
                tx_df_cp = tx_df_cp[["cp_group", "charge_speed"]]
                tx_df_cp.loc[tx_df_cp.index[0] - pd.Timedelta(seconds=1)] = {"cp_group":tx_df_cp["cp_group"].iloc[0],"charge_speed":0}
                tx_df_cp.loc[pd.to_datetime(START_DATE)] = {"cp_group":tx_df_cp["cp_group"].iloc[0],"charge_speed":0}

            tx_df_cp = tx_df_cp.sort_index(ascending=True)

            #tx_df_cp = tx_df_cp.asfreq(freq='5Min', method='bfill')
            if False: #TODO: resample?
                tx_df_cp = tx_df_cp.resample('5Min', offset="0s").bfill()
            # logger.debug(tx_df_cp.iloc[0:])
            tx_df_cps[tx_df_cp["cp_group"].iloc[0]].append((cp, tx_df_cp))
        cpo_tx_df_cps[cpo_id] = tx_df_cps
    return cpo_tx_df_cps

def plot_cpo_ocpp_data(tx_df_cps):
    fig0, axes0 = plt.subplots()
    for k,v in sorted(tx_df_cps.items()):
        logger.debug(k,len(v))
        tx_df_group = pd.concat([_v[1] for _v in v])
        tx_df_group = tx_df_group.reset_index(drop=False)
        # logger.debug(tx_df_group.iloc[0:])
        tx_df_group = tx_df_group[["meter_value_ts", "charge_speed"]]

        tx_df_group_sum = tx_df_group.groupby("meter_value_ts").sum()
        # logger.debug(tx_df_group_sum.iloc[0:])
        # logger.debug(tx_df_group_sum.columns)
        # break
        tx_df_group_sum.plot(ax=axes0, y='charge_speed', label=k)
        # for _v in v:
        #     #tx_df[tx_df["cp_c_id"] == _v[0]].plot(ax=axes0, x="meter_value_ts", y='charge_speed')
        #     _v[1].plot(ax=axes0, y='charge_speed')
        #break
    # plt.show()
    # plt.close()


def get_dso_oscp_data(file_df, ts_string = "sim_time", meas_string = "charge_speed", time_offset_in_h=0) ->  Dict[str, pd.DataFrame]:
    #/home/dk/git/wattson-artifacts/test_powerowl_example_2023-12-22-18-22-51/n375/n375-service-35.log
    dso_oscp = file_df[(file_df["type"] == "logs") & (file_df["sub_type"] == "n375")][["x_path", "file"]]
    dso_oscp_log = dso_oscp[(dso_oscp["file"] == "n375-service-35.log")]
    dso_oscp_log_file = os.path.join(*dso_oscp_log.iloc[0].values)
    
    regex_dso = re.compile("^\d{1,4}-\d{1,2}-\d{1,2} \d{1,2}:\d{1,2}:\d{1,2},\d{1,3} - DSO_OSCP - INFO - [\d\.]+ -  handleUpdateGroupMeasurements (.*?)$")
    num_lines = sum(1 for line in open(dso_oscp_log_file,'r'))
    dso_tx_data = []
    with open(dso_oscp_log_file) as f:
        for line in tqdm(f, total=num_lines, desc=f"Loading {dso_oscp_log['file'].iloc[0]}"):
            if "ERROR" in line or "Traceback" in line:
                logger.error(f"found error in {dso_oscp_log_file}:\t{line}")
            result = regex_dso.search(line)
            if result:
                try:
                    #({'group_id': 'CP_4', 'measurements': [{'value': 9, 'phase': 'ALL', 'unit': 'WH', 'energy_type': 'FLEXIBLE', 'direction': 'NET', 'measure_time': '2023-01-02T14:05:00', 'initial_measure_time': '2023-01-01T17:34:05'}]}, 'wjm1neFbwSEVkns5IYstFsheEWR3tsueIWJbJvVFCtE')
                    g = result.group(1)
                    #('2023-11-03T09:24:52.617660', 0.25, {'group_id': 'CP_9', 'measurements': [{'value': 2533, 'phase': 'ALL', 'unit': 'WH', 'energy_type': 'FLEXIBLE', 'direction': 'NET', 'measure_time': '2023-11-03T10:12:14', 'initial_measure_time': '2023-11-02T21:42:00'}]}, 'Xng5QTKs7KgjYyJL-ygXl259BVYjUdYROq96ojOj9eI')
                    g_lit = ast.literal_eval(g)
                except ValueError as e:
                    if "nan" in g or "inf" in g:
                    #     g = g.replace("nan", "0")
                    #     g_lit = ast.literal_eval(g)
                    # else:
                        logger.error(f"ValueError {e=} in {result.group(0)}")
                        continue
                    else: raise e
                sim_time = g_lit[0]
                oscp_interval_h = g_lit[1]
                g_lit_d = g_lit[2]
                tx_msg = {
                    "sim_time": sim_time,
                    "oscp_interval_h": oscp_interval_h,
                    "group_id": g_lit_d["group_id"],
                    # "token": g_lit[3],
                }
                for k,v in g_lit_d["measurements"][0].items():
                    tx_msg["measurements_"+k] = v
                dso_tx_data.append(tx_msg)

    dso_tx_df = pd.DataFrame(dso_tx_data)
    dso_tx_df["sim_time"] = pd.to_datetime(dso_tx_df["sim_time"])
    dso_tx_df["measurements_measure_time"] = pd.to_datetime(dso_tx_df["measurements_measure_time"]) - pd.Timedelta(hours=time_offset_in_h)
    dso_tx_df["measurements_initial_measure_time"] = pd.to_datetime(dso_tx_df["measurements_initial_measure_time"])
    

    # dso_tx_df["measurements_value"] = dso_tx_df["measurements_value"] / dso_tx_df["oscp_interval_h"]
    oscp_interval = dso_tx_df["oscp_interval_h"].mean() * 60
    logger.info(dso_tx_df.iloc[0:])
    logger.debug(dso_tx_df.columns)

    dso_tx_df_cps = dso_tx_df["group_id"].drop_duplicates()
    #fig, axes = plt.subplots(nrows=len(dso_tx_df_cps), ncols=1)

    # ts_strings = ["measurements_measure_time", "sim_time"]
    # ts_string = "measurements_measure_time"
    # ts_string = "sim_time"
    # meas_string = "charge_speed"
    # meas_string = "measurements_value"
    # fig, axes = plt.subplots()
    # for i,cp_g in enumerate(dso_tx_df_cps):
    #     dso_tx_df_cp = dso_tx_df[dso_tx_df["group_id"] == cp_g]
    #     dso_tx_df_cp = dso_tx_df_cp.sort_values(by=ts_string, ascending=True)
    #     dso_tx_df_cp.plot(ax=axes, x='measurements_measure_time', y='measurements_value', label=cp_g)

    # plt.show()
    # fig.close()

    # for ts_string in ts_strings:
    # fig2, axes2 = plt.subplots()

    dso_tx_df_cp_dict = {}
    for i,cp_g in enumerate(sorted(dso_tx_df_cps)):
        try:
            dso_tx_df_cp = dso_tx_df[dso_tx_df["group_id"] == cp_g]
            # dso_tx_df_cp = dso_tx_df_cp[::4]

            dso_tx_df_cp = dso_tx_df_cp.sort_values(by=ts_string, ascending=True)
            if len(dso_tx_df_cp.loc[dso_tx_df_cp[ts_string].duplicated(keep="first")]) > 0:
                logger.info(dso_tx_df_cp[dso_tx_df_cp[ts_string].duplicated(keep=False)])
                logger.info(dso_tx_df_cp[dso_tx_df_cp[ts_string].duplicated(keep="last")].index[0])
                logger.info(dso_tx_df_cp[dso_tx_df_cp[ts_string].duplicated(keep="first")].index[0])
                dso_tx_df_cp_t_i1 = dso_tx_df_cp[dso_tx_df_cp[ts_string].duplicated(keep="first")].index
                dso_tx_df_cp_t_i2 = dso_tx_df_cp[dso_tx_df_cp[ts_string] > max(dso_tx_df_cp.loc[dso_tx_df_cp_t_i1][ts_string])].index
                dso_tx_df_cp.loc[dso_tx_df_cp_t_i1 , ts_string] += pd.Timedelta(hours=1)
                dso_tx_df_cp.loc[dso_tx_df_cp_t_i2 , ts_string] += pd.Timedelta(hours=1)

            #dso_tx_df_cp = dso_tx_df_cp.iloc[0::8]

            # print(dso_tx_df_cp)
            if True:
                OSCP_INTERVAL_M=15
                dso_tx_df_cp["meter_diff"] = dso_tx_df_cp["measurements_value"] - dso_tx_df_cp["measurements_value"].shift(1)
                dso_tx_df_cp["meter_diff"].fillna(0, inplace=True)
                dso_tx_df_cp["time_diff"] = OSCP_INTERVAL_M / 60
                dso_tx_df_cp["charge_speed"] = dso_tx_df_cp["meter_diff"] / dso_tx_df_cp["time_diff"] #wh in w
                dso_tx_df_cp["charge_speed"].fillna(0, inplace=True)
            if False:
                dso_tx_df_cp["meter_diff"] = dso_tx_df_cp["measurements_value"] - dso_tx_df_cp["measurements_value"].shift(1)
                dso_tx_df_cp["meter_diff"].fillna(0, inplace=True)
                dso_tx_df_cp["time_diff"] = (dso_tx_df_cp[ts_string] - dso_tx_df_cp[ts_string].shift(1)).apply(lambda x: x.total_seconds())
                dso_tx_df_cp["time_diff"] = dso_tx_df_cp["time_diff"] / (60*60) #s to h
                dso_tx_df_cp["charge_speed"] = dso_tx_df_cp["meter_diff"] / dso_tx_df_cp["time_diff"] #wh in w
                dso_tx_df_cp["charge_speed"].fillna(0, inplace=True)
            if False:
                dso_tx_df_cp["charge_speed"] = dso_tx_df_cp["measurements_value"]
                dso_tx_df_cp["time_diff"] = (dso_tx_df_cp[ts_string] - dso_tx_df_cp[ts_string].shift(1)).apply(lambda x: x.total_seconds())
                dso_tx_df_cp["time_diff"] = dso_tx_df_cp["time_diff"] / (60*60) #s to h
                dso_tx_df_cp["meter_diff"] = dso_tx_df_cp["charge_speed"] * dso_tx_df_cp["time_diff"]
                dso_tx_df_cp["meter_diff"].fillna(0, inplace=True)
                dso_tx_df_cp["measurements_value"] = 0
                dso_tx_df_cp["measurements_value"] = dso_tx_df_cp["meter_diff"].cumsum()
            # print(dso_tx_df_cp)
            # exit()
            # if cp_g == "CP_11":
            #     with pd.option_context('display.max_rows', 100, 'display.max_columns', 10):
            #         print(dso_tx_df_cp.head(100)[["sim_time", "meter_diff", "time_diff" , "charge_speed", "measurements_value"]])
            #     exit()
            
            
            dso_tx_df_cp = dso_tx_df_cp.set_index(ts_string)
            dso_tx_df_cp = dso_tx_df_cp.sort_index(ascending=True)
            # dso_tx_df_cp = dso_tx_df_cp.asfreq(freq='5Min', method='bfill')

            if False: #TODO: START_DATE=
                dso_tx_df_cp = dso_tx_df_cp[["group_id", meas_string]]
                # print(dso_tx_df_cp)
                dso_tx_df_cp.loc[dso_tx_df_cp.index[0] - pd.Timedelta(seconds=1)] = {"group_id":dso_tx_df_cp["group_id"].iloc[0],meas_string:0}
                dso_tx_df_cp.loc[pd.to_datetime(START_DATE)] = {"group_id":dso_tx_df_cp["group_id"].iloc[0],meas_string:0}
                dso_tx_df_cp = dso_tx_df_cp.sort_index(ascending=True)
                # print(dso_tx_df_cp)
                # exit()

            #dso_tx_df_cp = dso_tx_df_cp.resample(str(oscp_interval)+'Min', offset="0s").bfill()
            if False: #TODO: resample?
                dso_tx_df_cp = dso_tx_df_cp.resample('5Min', offset="0s").bfill()
            dso_tx_df_cp_dict[cp_g] = dso_tx_df_cp
            # dso_tx_df_cp.plot(ax=axes2, y=meas_string, label=cp_g)
        except Exception as e:
            logger.error(e)
            logger.error(dso_tx_df_cp.index)
            logger.error(len(dso_tx_df_cp.index))
            logger.error(len(dso_tx_df_cp.index.drop_duplicates()))
            logger.error(dso_tx_df_cp[dso_tx_df_cp.index.duplicated(keep=False)])
            raise e
    # plt.show()
    # plt.close()
    return dso_tx_df_cp_dict

def plot_dso_oscp_data(dso_tx_df_cp_dict, meas_string = "charge_speed"):
    fig2, axes2 = plt.subplots()
    for cp_g,dso_tx_df_cp in dso_tx_df_cp_dict.items():
        dso_tx_df_cp.plot(ax=axes2, y=meas_string, label=cp_g)
    # plt.show()
    # plt.close()

def get_measurements_expo(file_df, DIR) -> pd.DataFrame:
    pp=get_grid_pp(DIR)

    sgen = get_grid_measurements_from_export(DIR, pp, bus_ns=[], target_elem="sgen", target_var=".MEASUREMENT.active_power" )
    bus = get_grid_measurements_from_export(DIR, pp, bus_ns=[], target_elem="bus", target_var=".MEASUREMENT.active_power" )
    load = get_grid_measurements_from_export(DIR, pp, bus_ns=[], target_elem="load", target_var=".MEASUREMENT.active_power" )
    #storage = get_grid_measurements_from_export(DIR, pp, bus_ns=[], target_elem="storage", target_var=".MEASUREMENT.active_power" )
    elems=[sgen, bus, load]#, storage

    out_df = None
    for e in elems:
        for group,vals in e.items():
            for val in vals:
                if out_df is None:
                    out_df = val.to_frame()
                else:
                    out_df[val.name] = val
                    out_df[val.name] = val
                new_name = val.name.replace(".MEASUREMENT","").replace("grid_expo_","0.0.")
                out_df = out_df.rename(columns={val.name: new_name}).copy()
    return out_df


def get_measurements(file_df, DataPointMapsDIR) -> pd.DataFrame:
    regex_meas = re.compile(r"^measurements\-(\d+)\.jsonl$")
    measurements = file_df[(file_df["type"] == "controller-export") & (file_df["sub_type"] == "measurements")][["x_path", "file"]]
    logger.debug(measurements)
    measurement_dict = defaultdict(lambda: [])
    for index, row in tqdm(measurements.iterrows(), desc="Loading measurements"):
        result = regex_meas.search(row["file"])
        if result:
            with open(os.path.join(row["x_path"], row["file"]), "rb") as f:
                for line in f:
                    measurement_data = json.loads(line)
                    measurement_dict[result.group(1)].append(measurement_data)
        else:
            logger.error(f"unk file {row['file']}")

    for k,v in list(measurement_dict.items()):
        logger.debug(k, len(v))
    for v in measurement_dict["401"][:10]:
        #{'coa': 401, 'value_map': {'10010': 1.0299999713897705, '10020': 0.0, '10030': -8289678.0, '10040': -10461183.0, '10050': 8289678.0, '10060': 10461183.0}, 'sim-time': 1698972846.4926767, 'clock-time': 1703443759.360438}
        #{'coa': 401, 'value_map': {'10010': 1.0299999713897705, '10020': 0.0}, 'sim-time': 1698982975.2340364, 'clock-time': 1703443770.6145942}
        logger.debug(v)

    regex_dp = re.compile(r"^(\d+)\-data\-points\.yml$")
    # DataPointMapsDIR = configs[dataset]["DataPointMapsDIR"]
    onlyfiles = [f for f in os.listdir(DataPointMapsDIR) if os.path.isfile(os.path.join(DataPointMapsDIR, f))]
    logger.debug(onlyfiles)
    
    dp_dict = defaultdict(lambda: [])
    for f in tqdm(onlyfiles, desc="Loading Data Point Maps"):
        result = regex_dp.search(f)
        if result:
            with open(os.path.join(DataPointMapsDIR, f), "rb") as f:
                dp_data = yaml.load(f, Loader=yaml.FullLoader)
                dp_dict[result.group(1)].append(dp_data)
        else:
            logger.error(f"unk file {f}")
    
    #[{'401': [{'identifier': '401.10010', 'protocol': '60870-5-104', 'protocol_data': {'coa': 401, 'cot': 1, 'direction': 'monitoring', 'ioa': 10010, 'type_id': 13}, 'providers': {'sources': [{'domain': 'source', 'provider_data': {'attribute': 'voltage', 'context': 'MEASUREMENT', 'grid_element': 'bus.0', 'type': 'float'}, 'provider_type': 'POWER_GRID'}]}, 'value': None}, {'identifier': '401.10020', 'protocol': '60870-5-104', 'protocol_data': {'coa': 401, 'cot': 1, .... 
    logger.debug(dp_dict["401"])

    meas_data = []
    for coa,meas_s in tqdm(list(measurement_dict.items()), desc="Loading meas_data"):
        for meas in meas_s:
            if str(meas["coa"]) != coa:
                logger.error(f"unk meas coa {coa, meas}")
            dps = dp_dict[coa]
            for dp in dps:
                coa_dps = dp[coa]
                for coa_dp in coa_dps:
                    ioa = coa_dp['protocol_data']["ioa"] #10010
                    if str(ioa) not in meas["value_map"]:
                        continue #removes control IOAs...
                    
                    if str(coa_dp['protocol_data']["coa"]) != coa:
                        logger.error(f"unk dp coa {coa, coa_dp}")
                    coa_ioa = coa_dp["identifier"] #'401.10010'

                    direction = coa_dp['protocol_data']["direction"] #control or monitoring
                    direction_s = "sources" if direction == "monitoring" else "targets"

                    if direction_s not in coa_dp['providers']:
                        logger.error(f"no {direction_s=} in {coa, coa_dp}")
                    if len(coa_dp['providers'][direction_s]) > 1:
                        logger.error(f"unk {direction_s=} len>1 {coa, coa_dp}")
                    domain = coa_dp['providers'][direction_s][0]["domain"] #source or target
                    attribute = coa_dp['providers'][direction_s][0]["provider_data"]["attribute"] #voltage
                    grid_element = coa_dp['providers'][direction_s][0]["provider_data"]["grid_element"] #bus.0

                    meas_value = meas["value_map"][str(ioa)]
                    sim_time = meas["sim-time"]
                    #clock_time = meas["clock-time"]

                    meas_data.append({"coa_ioa":coa_ioa,"coa":coa,"ioa":ioa,"direction":direction,"domain":domain,"attribute":attribute,"grid_element":grid_element,"meas_value":meas_value,"sim_time":sim_time})#,"clock_time":clock_time})

    meas_df = pd.DataFrame(meas_data)
    meas_df["sim_time"] = pd.to_datetime(meas_df["sim_time"], unit="s")
    #meas_df["clock_time"] = pd.to_datetime(meas_df["clock_time"], unit="s")
    meas_df.sort_values(by="sim_time", ascending=True)
    logger.debug(meas_df)
    meas_df2 = meas_df[(meas_df["direction"] == "monitoring") & (meas_df["domain"] == "source")] #does nothing
    # logger.debug(meas_df)

    dups = defaultdict(lambda:[]) #dups for two sides of lines

    clean_meas_data = []
    for coa_ioa in tqdm(meas_df["coa_ioa"].drop_duplicates(), desc="Loading clean_meas_data"):
        meas_df_i = meas_df[meas_df["coa_ioa"] == coa_ioa].copy()
        # logger.debug(meas_df_i)
        if len(meas_df_i["attribute"].drop_duplicates()) != 1 or len(meas_df_i["grid_element"].drop_duplicates()) != 1:
            logger.error(f'error w dups in {meas_df_i["attribute"].drop_duplicates(), meas_df_i["grid_element"].drop_duplicates()}')

        att = meas_df_i["attribute"].drop_duplicates().iloc[0]
        ge = meas_df_i["grid_element"].drop_duplicates().iloc[0]

        meas_df_i = meas_df_i.set_index("sim_time")
        meas_df_i = meas_df_i.sort_index(ascending=True)

        def resample_meas(df, freq='15Min'):
            oidx = df.index
            nidx = pd.date_range(oidx.min(), oidx.max(), freq=freq, normalize=True)
            res = df.reindex(oidx.union(nidx)).interpolate('time', limit_direction="both").reindex(nidx)
            return res
    
        if True: #TODO: resample? -> yes in order to concat diff freqs...
            #meas_df_i = meas_df_i.resample('5Min', offset="0s").bfill()
            #meas_df_i = meas_df_i.resample('5Min', offset="0s").interpolate('time')
            meas_df_i = resample_meas(meas_df_i, freq="5Min")
            
        # logger.debug(meas_df_i)
        meas_df_i = meas_df_i[["meas_value"]]
        meas_df_i = meas_df_i.rename(columns={"meas_value": coa_ioa+"."+ge+"."+att}) #duplicates lines witout coa_ioa... (dups for two sides of lines)
        dups[ge+"."+att].append(coa_ioa)
        # logger.debug(meas_df_i)
        
        clean_meas_data.append(meas_df_i)
    
    clean_meas_df = pd.concat(clean_meas_data, axis=1)
    # logger.debug(clean_meas_df)
    clean_meas_df = clean_meas_df.fillna(clean_meas_df.mean(skipna=True))
    logger.info(clean_meas_df)

    # for k,v in dups.items():
    #     if len(v) > 1:
    #         logger.debug("dup",k,v)
    return clean_meas_df

            

def get_estimations(file_df, drop_dups=False) -> pd.DataFrame:
    est = file_df[(file_df["type"] == "controller-export") & (file_df["sub_type"] == "estimation")][["x_path", "file"]]
    logger.info(est)
    est_dict = defaultdict(lambda: list())
    for index, row, in tqdm(est.iterrows(), total = len(est), desc="Loading estimates"):

        regex_est = re.compile(r"^WALL\-[\-\dT\+]+__SIM\-([\-\dT\+]+)\.powerowl\.p\.gz$")
        #WALL-2023-12-27T15-50-38-049544+00-00__SIM-2023-11-03T02-28-11-789274+00-00.powerowl.p.gz
        result = regex_est.search(row["file"])
        if result:
            with gzip.open(os.path.join(row["x_path"], row["file"]), "rb") as f:
                est_data = pickle.load(f)
                pd.reset_option('^display.', silent=True)
            # logger.info(est_data)
            # logger.info(est_data["res_line_est"])
            # logger.info(est_data["line"])
            time = "".join(result.group(1).rsplit("-", 1))
            time = pd.to_datetime(time, format="%Y-%m-%dT%H-%M-%S-%f%z") #2023-11-23T19-05-29-585438+00-00
            
            for c in est_data.keys():
                if c.startswith("res_") and c.endswith("_est"):
                    c_name = c.replace("res_", "").replace("_est", "")
                    est_df = est_data[c]
                    if est_df.empty: continue
                    df_out = est_df.unstack().to_frame().T.sort_index(level=0)
                    df_out.columns = [f'{c_name}.{j}.{i}' for i, j in df_out.columns]
                    df_out.index = [time]
                    est_dict[c_name].append(df_out)
        else:
            logger.error(f"unk file {row['file']}")
    est_l=[]
    for k,v in est_dict.items():
        k_df = pd.concat(v)
        est_l.append(k_df)
    est_df = pd.concat(est_l, axis=1)

    if drop_dups:
        logger.warning("----dropping dups-----")
        logger.warning(est_df)
        # est_df = est_df.round(5)
        logger.warning(est_df.std())
        est_df = est_df.drop(est_df.std()[(est_df.std() == 0)].index, axis=1)
        logger.warning(est_df)
        logger.warning("----dropping dups-----")

    est_df = est_df.sort_index(ascending=True)
    if False: #TODO: resample?
        est_df = est_df.resample('5Min', offset="0s").bfill()
    logger.info(est_df)
    return est_df
                    
            