
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from copy import deepcopy
import gc
import hashlib
import itertools
import json
import multiprocessing
from multiprocessing.pool import ThreadPool
import os
import warnings
import pandas as pd
from sklearn.exceptions import ConvergenceWarning
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import KFold
from sklearn.neural_network import MLPRegressor
from sklearn.svm import LinearSVR
from sklearn.tree import DecisionTreeRegressor
from tqdm import tqdm

from sklearn.ensemble import GradientBoostingRegressor, HistGradientBoostingRegressor, RandomForestRegressor

import re

import logging

logger = logging.getLogger("WATTSON_EV_IDS.Regression")


m_dict = multiprocessing.Manager()
REG_CACHE=m_dict.dict()
#REG_CACHE=dict()
with open("ids.conf", 'r') as f:
    conf = json.load(f)
    NUM_THREADS = conf["NUM_THREADS"]
#NUM_THREADS=3
    
IGNORE_MISS_COLS=True

def get_regression_pred_conc_kfold(df_dict, TRAIN_START_DATE, ATK_START_DATE, reg_type, conf, actor, features, num_shifts=1, pbar=None, kfold_splits=[None], atk_df=None, recursive=False):
    y_pred_folds=[]
    for kfold in kfold_splits:
        if recursive:
            y_preds = get_regression_pred_rec(df_dict, TRAIN_START_DATE, ATK_START_DATE, reg_type, conf, num_shifts=num_shifts, pbar=pbar, kfold=kfold, atk_df=atk_df)
        else:
            y_preds = get_regression_pred_conc(df_dict, TRAIN_START_DATE, ATK_START_DATE, reg_type, conf, num_shifts=num_shifts, pbar=pbar, kfold=kfold, atk_df=atk_df)

        #pbar.update(len(y_preds))
        y_pred = pd.concat(y_preds, axis=1)
        # print(y_pred)
        # exit()
        y_pred["conf"] = [conf] * len(y_pred)
        y_pred["features"] = [features] * len(y_pred)
        y_pred["reg"] = reg_type
        y_pred["actor"] = actor
        # print(y_pred)
        # exit()
        y_pred_folds.append(y_pred)

    y_pred_all = pd.concat(y_pred_folds)
    y_pred_all.sort_index(ascending=True, inplace=True)
    return y_pred_all


def get_regression_pred_rec(df_dict, TRAIN_START_DATE, ATK_START_DATE, reg_type, conf, num_shifts=1, pbar=None, kfold=None, atk_df=None):
    if kfold is None:
        y_preds = get_regression_pred_rec_pred(df_dict, TRAIN_START_DATE, ATK_START_DATE, reg_type, conf, num_shifts=num_shifts, pbar=pbar, kfold=kfold, atk_df=atk_df)
    else:
        y_preds = get_regression_pred_rec_kfold(df_dict, TRAIN_START_DATE, ATK_START_DATE, reg_type, conf, num_shifts=num_shifts, pbar=pbar, kfold=kfold, atk_df=atk_df)
    return y_preds

def get_regression_pred_rec_pred(df_dict, TRAIN_START_DATE, ATK_START_DATE, reg_type, conf, num_shifts=1, pbar=None, kfold=None, atk_df=None):
    y_preds = []
    atk_df_do = atk_df.copy()
    atk_df_s = atk_df.copy()

    for x in range(num_shifts):
        # index, y_pred = get_regression_pred(df_dict, TRAIN_START_DATE, ATK_START_DATE, reg_type, conf, shift=x)
        index, y_pred, used_shift = get_regression_pred(df_dict, TRAIN_START_DATE, ATK_START_DATE, reg_type, conf, 0, atk_df=atk_df_do)
        y_pred = pd.DataFrame(y_pred, index=index)
        y_pred[0] = y_pred[0].shift(used_shift)
        y_pred.rename(columns={0: "prediction_"+str(x)}, inplace=True)
        y_preds.append(y_pred)

        # print(atk_df.loc[ATK_START_DATE:][[c for c in atk_df.columns if  c.startswith("charge_speed_lag_")]])
        # print(y_pred.shift(1))
        y_pred_s = y_pred.shift(x+1)
        atk_df_s = atk_df_do.copy()
        atk_df_s["charge_speed_lag_"+str(x+1)] = y_pred_s
        atk_df_s[atk_df_s["charge_speed_lag_"+str(x+1)].isna()] = atk_df[atk_df_s["charge_speed_lag_"+str(x+1)].isna()]
        atk_df_do=atk_df_s
        if pbar is not None:
            pbar.update(1)
        # print(atk_df_s.loc[ATK_START_DATE:][[c for c in atk_df.columns if  c.startswith("charge_speed_lag_")]])
        # exit()

    return y_preds


def get_regression_pred_rec_kfold(df_dict, TRAIN_START_DATE, ATK_START_DATE, reg_type, conf, num_shifts=1, pbar=None, kfold=None, atk_df=None):
    y_preds = []

    df_all = df_dict["cp_g_df"].loc[TRAIN_START_DATE:ATK_START_DATE] #.copy()

    feat_cols = df_dict["feat_cols"]
    pred_col = df_dict["pred_col"]

    y_preds = dict()
    y_preds_fold = defaultdict(lambda:list())
    kf = KFold(n_splits=kfold, random_state=None, shuffle=False)
    for i, (train_index, test_index) in enumerate(kf.split(df_all)):
        df = df_all.iloc[train_index].copy()
        tbd_df = df_all.iloc[test_index].copy()

        for x in range(num_shifts):
            try:
                index, y_pred, used_shift = _do_get_regression_pred(df, tbd_df, pred_col, feat_cols, reg_type, conf, shift=0)

                y_pred = pd.DataFrame(y_pred, index=index)
                y_pred.rename(columns={0: "prediction_"+str(x)}, inplace=True)
                y_preds_fold[x+1].append(y_pred)

                # print(tbd_df.loc[TRAIN_START_DATE:ATK_START_DATE][[c for c in tbd_df.columns if  c.startswith("charge_speed_lag_")]])
                # print(y_pred.shift(0))
                y_pred_s = y_pred.shift(x+1)
                tbd_df_s = tbd_df.copy()
                tbd_df_s["charge_speed_lag_"+str(x+1)] = y_pred_s
                tbd_df_s[tbd_df_s["charge_speed_lag_"+str(x+1)].isna()] = tbd_df[tbd_df_s["charge_speed_lag_"+str(x+1)].isna()]
                tbd_df=tbd_df_s
                # print(tbd_df_s.loc[TRAIN_START_DATE:ATK_START_DATE][[c for c in tbd_df_s.columns if  c.startswith("charge_speed_lag_")]])
                # exit()
                if pbar is not None:
                    pbar.update(1)

            except ValueError as e:
                logger.error(df)
                logger.error(tbd_df)
                logger.error(pred_col)
                logger.error(feat_cols)
                logger.error(e)
                raise e
        
    for k,v in y_preds_fold.items():
        y_pred = pd.concat(v)
        y_pred.sort_index(ascending=True, inplace=True)
        y_preds[k] = y_pred
    ret_y_pred = [v for v in y_preds.values()]
    return ret_y_pred

def get_regression_pred_conc(df_dict, TRAIN_START_DATE, ATK_START_DATE, reg_type, conf, num_shifts=1, pbar=None, kfold=None, atk_df=None):
    nt=NUM_THREADS
    if reg_type == "RandomForestRegressor" or reg_type == "MLPRegressor":
        nt=1
    y_preds = []
    results=[]

        
    with ThreadPoolExecutor(nt) as pool:
    #with ProcessPoolExecutor(nt) as pool:
        for x in range(num_shifts):
            # index, y_pred = get_regression_pred(df_dict, TRAIN_START_DATE, ATK_START_DATE, reg_type, conf, shift=x)
            if kfold is None: #no kfold for atk data sets
                if nt == 1:
                    results.append(get_regression_pred(df_dict, TRAIN_START_DATE, ATK_START_DATE, reg_type, conf, x, atk_df=atk_df))
                else:
                    results.append(pool.submit(get_regression_pred, df_dict, TRAIN_START_DATE, ATK_START_DATE, reg_type, conf, x, atk_df))
            else: # yes kfold for training data sets
                if nt == 1:
                    results.append(get_regression_kfold(df_dict, TRAIN_START_DATE, ATK_START_DATE, reg_type, conf, x, kfold, pbar))
                else:
                    results.append(pool.submit(get_regression_kfold, df_dict, TRAIN_START_DATE, ATK_START_DATE, reg_type, conf, x, kfold, pbar))

        for r in results:
            if nt == 1:
                index, y_pred, used_shift = r
            else:
                index, y_pred, used_shift = r.result()
            if pbar is not None and kfold is None:
                pbar.update(1)
            y_pred = pd.DataFrame(y_pred, index=index)
            y_pred[0] = y_pred[0].shift(used_shift)
            y_pred.rename(columns={0: "prediction_"+str(used_shift)}, inplace=True)
            y_preds.append(y_pred)

    return y_preds

def get_regression_kfold(df_dict, TRAIN_START_DATE, ATK_START_DATE, reg_type, conf, shift=0, nfolds=5, pbar=None):
    df_all = df_dict["cp_g_df"].loc[TRAIN_START_DATE:ATK_START_DATE] #.copy()

    feat_cols = df_dict["feat_cols"]
    pred_col = df_dict["pred_col"]

    y_preds = []
    kf = KFold(n_splits=nfolds, random_state=None, shuffle=False)
    for i, (train_index, test_index) in enumerate(kf.split(df_all)):
        df = df_all.iloc[train_index].copy()
        tbd_df = df_all.iloc[test_index].copy()

        try:
            index, y_pred, used_shift = _do_get_regression_pred(df, tbd_df, pred_col, feat_cols, reg_type, conf, shift=shift)
        except ValueError as e:
            logger.error(df)
            logger.error(tbd_df)
            logger.error(pred_col)
            logger.error(feat_cols)
            logger.error(e)
            raise e
        del df
        del tbd_df
        gc.collect()
        y_pred = pd.DataFrame(y_pred, index=index)
        y_preds.append(y_pred)
        if pbar is not None:
            pbar.update(1)
    y_pred = pd.concat(y_preds)
    y_pred.sort_index(ascending=True, inplace=True)
    return y_pred.index, y_pred[0], shift


def get_regression_pred(df_dict, TRAIN_START_DATE, ATK_START_DATE, reg_type, conf, shift=0, atk_df=None):
    df = df_dict["cp_g_df"].loc[TRAIN_START_DATE:ATK_START_DATE].copy()
    feat_cols = df_dict["feat_cols"]
    pred_col = df_dict["pred_col"]
    if atk_df is None:
        tbd_df = df_dict["cp_g_df"].loc[ATK_START_DATE:].copy()
    else:
        tbd_df = atk_df.loc[ATK_START_DATE:].copy()
        mis_col = [c for c in df.columns if c not in tbd_df.columns and "date_exog" in c]
        tbd_df[mis_col] = 0

    miss_cols = [c for c in feat_cols if c not in tbd_df.columns]
    if len(miss_cols) > 0:
        if IGNORE_MISS_COLS:
            cp_miss = set()
            other_miss = list()
            for c in miss_cols:
                if c.startswith("CP_"):
                    cp_miss.add("_".join(c.split("_")[0:2]))
                else:
                    other_miss.append(c)
            if len(cp_miss)>1:
                logger.warning(f"ignoring missing columns in atk datafrtame, cp_miss: {len(cp_miss)}")
                logger.debug(f"ignoring missing columns in atk datafrtame, cp_miss: {cp_miss}")
            if len(other_miss)>1:
                logger.warning(f"ignoring missing columns in atk datafrtame, other_miss: {len(other_miss),other_miss}")
            feat_cols = [c for c in feat_cols if c in tbd_df.columns]
    # print(feat_cols)
    # exit()
    # print([c for c in tbd_df.columns])
    #test = tbd_df[feat_cols]
    return _do_get_regression_pred(df, tbd_df, pred_col, feat_cols, reg_type, conf, shift=shift)

    # df[pred_col] = df[pred_col].shift(shift*-1)
    # df.dropna(subset=[pred_col], inplace=True)

    # tbd_df[pred_col] = tbd_df[pred_col].shift(shift*-1)
    # tbd_df.dropna(subset=[pred_col], inplace=True)

    # return tbd_df.index, _do_regression(df, feat_cols, pred_col, tbd_df, conf, reg_type=reg_type), shift

def _do_get_regression_pred(df, tbd_df, pred_col, feat_cols, reg_type, conf, shift=0):
    df[pred_col] = df[pred_col].shift(shift*-1)
    df.dropna(subset=[pred_col], inplace=True)

    tbd_df[pred_col] = tbd_df[pred_col].shift(shift*-1)
    tbd_df#.dropna(subset=[pred_col], inplace=True)

    return tbd_df.index, _do_regression(df, feat_cols, pred_col, tbd_df, conf, reg_type=reg_type), shift


def _do_regression(df, feat_cols, pred_col, tbd_df, conf, reg_type=""):
    # print(df)
    # print(feat_cols)
    # exit()
    if reg_type == "RandomForestRegressor":
        #_do_RandomForestRegressor(df, feat_cols, pred_col, tbd_df, conf)
        return _do_RandomForestRegressor(df, feat_cols, pred_col, tbd_df, conf)
    elif reg_type == "DecisionTreeRegressor" or reg_type == "DecisionTreeRegressor2":
        return _do_DecisionTreeRegressor(df, feat_cols, pred_col, tbd_df, conf)
    elif reg_type == "GradientBoostingRegressor":
        return _do_GradientBoostingRegressor(df, feat_cols, pred_col, tbd_df, conf)
    elif reg_type == "HistGradientBoostingRegressor":
        return _do_HistGradientBoostingRegressor(df, feat_cols, pred_col, tbd_df, conf)
    elif reg_type == "LinearSVR":
        return _do_LinearSVR(df, feat_cols, pred_col, tbd_df, conf)
    elif reg_type == "MLPRegressor" or reg_type == "MLPRegressor2": # or reg_type == "test":
        return _do_MLPRegressor(df, feat_cols, pred_col, tbd_df, conf)
    else:
        logger.error(f"unk reg_type {reg_type}")
        return []
    
def optimize_regression(df_dict, TRAIN_START_DATE, VALIDATION_START_DATE, ATK_START_DATE, reg_type, group=""):
    df = df_dict["cp_g_df"].loc[TRAIN_START_DATE:VALIDATION_START_DATE]
    feat_cols = df_dict["feat_cols"]
    # print([c for c  in df.columns])
    # print(feat_cols)
    # exit()
    pred_col = df_dict["pred_col"]
    tbd_df = df_dict["cp_g_df"].loc[VALIDATION_START_DATE:ATK_START_DATE]

    df = df[feat_cols+[pred_col]].copy()
    tbd_df = tbd_df[feat_cols+[pred_col]].copy()
    if len(tbd_df) > 0:
        ret_ds = _optimize_regression(df, feat_cols, pred_col, tbd_df, group=group, reg_type=reg_type)
    else:
        ret_ds=[]
        logger.warning(f"Empty ATK df. Skipping {group=}")
    ret_df = pd.DataFrame(ret_ds)
    if len(ret_df) > 0:
        ret_df = ret_df.sort_values(by="mse", ascending=True)
    return ret_df
        
def _optimize_regression(df, feat_cols, pred_col, tbd_df, group="", reg_type=""):
    if reg_type == "RandomForestRegressor":
        return optimize_RandomForestRegressor(df, feat_cols, pred_col, tbd_df, group)
    elif reg_type == "DecisionTreeRegressor" or reg_type == "DecisionTreeRegressor2":
        return optimize_DecisionTreeRegressor(df, feat_cols, pred_col, tbd_df, group)
    elif reg_type == "GradientBoostingRegressor":
        return optimize_GradientBoostingRegressor(df, feat_cols, pred_col, tbd_df, group)
    elif reg_type == "HistGradientBoostingRegressor":
        return optimize_HistGradientBoostingRegressor(df, feat_cols, pred_col, tbd_df, group)
    elif reg_type == "LinearSVR":
        return optimize_LinearSVR(df, feat_cols, pred_col, tbd_df, group)
    elif reg_type == "MLPRegressor" or reg_type == "MLPRegressor2" or reg_type == "test":
        return optimize_MLPRegressor(df, feat_cols, pred_col, tbd_df, group)
    else:
        logger.error(f"unk reg_type {reg_type}")
        return []


def _do_DecisionTreeRegressor(df, feat_cols, pred_col, tbd_df, conf):
    lags = conf[6]
    #forecaster_reg = DecisionTreeRegressor(**{'criterion': conf[0], 'max_depth': conf[1], 'min_samples_split': conf[2], 'min_samples_leaf': conf[3], 'max_leaf_nodes': conf[4], 'max_features': conf[5], 
    #                                            'ccp_alpha': conf[7], 'splitter': conf[8], 'min_weight_fraction_leaf': conf[9], 'min_impurity_decrease': conf[10], 'random_state': 1234}) #
    reg_id = get_reg_id("DecisionTreeRegressor", df, feat_cols, pred_col, conf)
    if reg_id in REG_CACHE:
        logger.debug(f"hit {len(REG_CACHE)}")
        forecaster_reg = deepcopy(REG_CACHE[reg_id])
    else:
        logger.debug(f"miss {len(REG_CACHE)}")
        forecaster_reg = DecisionTreeRegressor(**{'criterion': conf[0], 'max_depth': conf[1], 'min_samples_split': conf[2], 'min_samples_leaf': conf[3], 'max_leaf_nodes': conf[4], 'max_features': conf[5], 
                                                'ccp_alpha': conf[7], 'splitter': conf[8], 'min_weight_fraction_leaf': conf[9], 'min_impurity_decrease': conf[10], 'random_state': 1234}) #
        forecaster_reg.fit(X = df[feat_cols], y = df[pred_col])
        REG_CACHE[reg_id] = forecaster_reg

    # forecaster_reg.n_estimators += n_estimators
    #forecaster_reg.fit(X = df[feat_cols], y = df[pred_col])
    y_pred = forecaster_reg.predict(tbd_df[feat_cols])
    return y_pred

def _optimize_DecisionTreeRegressor(_df, feat_cols, pred_col, _tbd_df, conf):
    df = _df#.copy()
    tbd_df = _tbd_df#.copy()
    y_pred = _do_DecisionTreeRegressor(df, feat_cols, pred_col, tbd_df, conf)
    y_true = tbd_df[pred_col]
    mse = mean_squared_error(y_true, y_pred)
    rmse = mean_squared_error(y_true, y_pred, squared=False)
    #print(f"{group=} {mse=} {rmse=} {conf=}")
    return {"conf":conf,"mse":mse,"rmse":rmse}


def optimize_DecisionTreeRegressor(df, feat_cols, pred_col, tbd_df, group=""):
    #criterion = ["squared_error", "absolute_error", "friedman_mse", "poisson"]
    criterion = ["squared_error", "absolute_error", "friedman_mse"]
    max_depth = [14, None]
    min_samples_split = [2,3]
    min_samples_leaf = [1,2]
    max_leaf_nodes = [None, 50]
    max_features = ["sqrt", None]
    lags=[16,32]
    ccp_alpha = [0.0]
    splitter  = ["best", "random"]
    min_weight_fraction_leaf = [0.0]
    min_impurity_decrease  = [0.0]

    criterion = ["absolute_error", "friedman_mse"]
    # max_depth = [None]
    min_samples_split = [2,5]
    min_samples_leaf = [1,3]
    max_leaf_nodes = [None]
    max_features = [None]
    #ccp_alpha  = [0.0, 0.5]
    splitter  = ["best"]
    # min_weight_fraction_leaf = [0.0, 0.1, 0.2]
    #min_impurity_decrease  = [0.0, 1.0]
    lags=[None]


    c = list(itertools.product(criterion, max_depth, min_samples_split, min_samples_leaf, max_leaf_nodes, max_features, lags,  ccp_alpha, splitter, min_weight_fraction_leaf, min_impurity_decrease))
    #c = [('absolute_error', 14, 2, 1, None, None, None, 0.0, 'random', 0.0, 0.0), ]
    #c = [('friedman_mse', None, 5, 3, None, None, None, 0.5, 'random', 0.0, 1.0), ]

    # ret_ds = []
    # for conf in tqdm(c, desc="conf "+group, disable=False):
    #     ret_ds.append(_optimize_DecisionTreeRegressor(df, feat_cols, pred_col, tbd_df, conf))
    # return ret_ds
        
    ret_ds = []
    results=[]
    print()
    with ThreadPoolExecutor(NUM_THREADS) as pool:
    # with ProcessPoolExecutor(NUM_THREADS) as pool:
        for conf in tqdm(c, desc="conf "+group, disable=True):
            results.append(pool.submit(_optimize_DecisionTreeRegressor, df, feat_cols, pred_col, tbd_df, conf))
            #_optimize_DecisionTreeRegressor(df, feat_cols, pred_col, tbd_df,conf, ret_ds)
        for r in tqdm(results, desc="conf "+group):
            ret_ds.append(r.result())   
    return ret_ds


def _do_RandomForestRegressor(df, feat_cols, pred_col, tbd_df, conf):
    lags = conf[6]
    n_estimators=conf[10]
    # forecaster_reg = RandomForestRegressor(**{'criterion': conf[0], 'max_depth': conf[1], 'min_samples_split': conf[2], 'min_samples_leaf': conf[3], 'max_leaf_nodes': conf[4], 'max_features': conf[5], 
    #                                             'oob_score': conf[7], 'ccp_alpha': conf[8], 'max_samples': conf[9], 'n_estimators': n_estimators, 'n_jobs': min(NUM_THREADS, n_estimators), 'random_state': 1234,
    #                                             'warm_start': False}) #
    reg_id = get_reg_id("RandomForestRegressor", df, feat_cols, pred_col, conf)
    if reg_id in REG_CACHE:
        logger.debug(f"hit {len(REG_CACHE)}")
        forecaster_reg = deepcopy(REG_CACHE[reg_id])
    else:
        logger.debug(f"miss {len(REG_CACHE)}")
        forecaster_reg = RandomForestRegressor(**{'criterion': conf[0], 'max_depth': conf[1], 'min_samples_split': conf[2], 'min_samples_leaf': conf[3], 'max_leaf_nodes': conf[4], 'max_features': conf[5], 
                                                'oob_score': conf[7], 'ccp_alpha': conf[8], 'max_samples': conf[9], 'n_estimators': n_estimators, 'n_jobs': min(NUM_THREADS, n_estimators), 'random_state': 1234,
                                                'warm_start': False}) #
        forecaster_reg.fit(X = df[feat_cols], y = df[pred_col])
        REG_CACHE[reg_id] = forecaster_reg

    # forecaster_reg.n_estimators += n_estimators
    #forecaster_reg.fit(X = df[feat_cols], y = df[pred_col])

    y_pred = forecaster_reg.predict(tbd_df[feat_cols])
    return y_pred

def _optimize_RandomForestRegressor(_df, feat_cols, pred_col, _tbd_df, conf):
    df = _df#.copy()
    tbd_df = _tbd_df#.copy()
    y_pred = _do_RandomForestRegressor(df, feat_cols, pred_col, tbd_df, conf)
    y_true = tbd_df[pred_col]
    mse = mean_squared_error(y_true, y_pred)
    rmse = mean_squared_error(y_true, y_pred, squared=False)
    #print(f"{group=} {mse=} {rmse=} {conf=}")
    return {"conf":conf,"mse":mse,"rmse":rmse}

def optimize_RandomForestRegressor(df, feat_cols, pred_col, tbd_df, group=""):
    #criterion = ["squared_error", "absolute_error", "friedman_mse", "poisson"]
    criterion = ["squared_error", "absolute_error"]
    #max_depth = [None, 14]
    max_depth = [None, 12, 14]
    min_samples_split = [2,3]
    min_samples_leaf = [1,2]
    max_leaf_nodes = [None, 5]
    max_features = ["sqrt", None]
    lags=[16,32]
    oob_score = [False]
    ccp_alpha = [0.0]
    max_samples = [None, 0.8]
    _n_estimators=[20,50,100,150]

    criterion = ["squared_error"]
    max_depth = [None, 12]
    # min_samples_split = [3]
    min_samples_split = [2]
    min_samples_leaf = [1]
    max_leaf_nodes = [None]
    max_features = [None]
    lags=[None]
    # max_samples = [None]
    # _n_estimators=[40,50,60]
    # _n_estimators=[5,10,15,20,30]
    _n_estimators=[50,100,150]


    c = list(itertools.product(criterion, max_depth, min_samples_split, min_samples_leaf, max_leaf_nodes, max_features, lags, oob_score, ccp_alpha, max_samples, _n_estimators))
    #c = [('squared_error', 12, 3, 2, None, None, None, False, 0.0, None, 20), ]

    ret_ds = []
    print()
    for conf in tqdm(c, desc="conf "+group, disable=(len(c) == 1)):
        res = _optimize_RandomForestRegressor(df, feat_cols, pred_col, tbd_df, conf)
        #print(f"{group=} {mse=} {rmse=} {conf=}")
        ret_ds.append(res)
    return ret_ds


def _do_HistGradientBoostingRegressor(df, feat_cols, pred_col, tbd_df, conf):
    #c = list(itertools.product(loss, learning_rate, max_iter, min_samples_leaf, max_depth, max_leaf_nodes))
    
    reg_id = get_reg_id("HistGradientBoostingRegressor", df, feat_cols, pred_col, conf)
    if reg_id in REG_CACHE:
        logger.debug(f"hit {len(REG_CACHE)}")
        forecaster_reg = deepcopy(REG_CACHE[reg_id])
    else:
        logger.debug(f"miss {len(REG_CACHE)}")
        forecaster_reg = HistGradientBoostingRegressor(**{'loss': conf[0], 'learning_rate': conf[1], 'max_iter': conf[2], 'min_samples_leaf': conf[3], 'max_depth': conf[4], 'max_leaf_nodes': conf[5], 
                                                         'random_state': 1234}) #
        forecaster_reg.fit(X = df[feat_cols], y = df[pred_col])
        REG_CACHE[reg_id] = forecaster_reg

    # forecaster_reg.n_estimators += n_estimators
    #forecaster_reg.fit(X = df[feat_cols], y = df[pred_col])

    y_pred = forecaster_reg.predict(tbd_df[feat_cols])
    return y_pred

def _optimize_HistGradientBoostingRegressor(_df, feat_cols, pred_col, _tbd_df, conf):
    df = _df#.copy()
    tbd_df = _tbd_df#.copy()
    y_pred = _do_HistGradientBoostingRegressor(df, feat_cols, pred_col, tbd_df, conf)
    y_true = tbd_df[pred_col]
    mse = mean_squared_error(y_true, y_pred)
    rmse = mean_squared_error(y_true, y_pred, squared=False)
    #print(f"{group=} {mse=} {rmse=} {conf=}")
    return{"conf":conf,"mse":mse,"rmse":rmse}


def optimize_HistGradientBoostingRegressor(df, feat_cols, pred_col, tbd_df, group=""):
    loss=['absolute_error']
    learning_rate=[0.1, 0.5]
    max_iter = [100]
    max_leaf_nodes = [50, 100, 150]
    max_depth = [None, 15]
    min_samples_leaf = [10,10,30]


    c = list(itertools.product(loss, learning_rate, max_iter, min_samples_leaf, max_depth, max_leaf_nodes))

    ret_ds = []
    results=[]
    print()
    with ThreadPoolExecutor(NUM_THREADS) as pool:
    # with ProcessPoolExecutor(NUM_THREADS) as pool:
        for conf in tqdm(c, desc="conf "+group, disable=True):
            results.append(pool.submit(_optimize_HistGradientBoostingRegressor, df, feat_cols, pred_col, tbd_df, conf))
            #_optimize_DecisionTreeRegressor(df, feat_cols, pred_col, tbd_df,conf, ret_ds)
        for r in tqdm(results, desc="conf "+group):
            ret_ds.append(r.result())   
    return ret_ds


def _do_GradientBoostingRegressor(df, feat_cols, pred_col, tbd_df, conf):
    n_estimators=conf[2]

    reg_id = get_reg_id("GradientBoostingRegressor", df, feat_cols, pred_col, conf)
    if reg_id in REG_CACHE:
        logger.debug(f"hit {len(REG_CACHE)}")
        forecaster_reg = deepcopy(REG_CACHE[reg_id])
    else:
        logger.debug(f"miss {len(REG_CACHE)}")
        if len(conf) >14:
            forecaster_reg = GradientBoostingRegressor(**{'loss': conf[0], 'learning_rate': conf[1], 'subsample': conf[3], 'criterion': conf[4], 'min_samples_split': conf[5], 'min_samples_leaf': conf[6], 
                                                        'min_weight_fraction_leaf': conf[7], 'max_depth': conf[8], 'min_impurity_decrease': conf[9], 'max_features': conf[10], 'alpha': conf[11], 
                                                        'max_leaf_nodes': conf[12], 'ccp_alpha': conf[13], 'n_iter_no_change': conf[14],
                                                        'n_estimators': n_estimators, 'random_state': 1234}) #
        else:
            forecaster_reg = GradientBoostingRegressor(**{'loss': conf[0], 'learning_rate': conf[1], 'subsample': conf[3], 'criterion': conf[4], 'min_samples_split': conf[5], 'min_samples_leaf': conf[6], 
                                                        'min_weight_fraction_leaf': conf[7], 'max_depth': conf[8], 'min_impurity_decrease': conf[9], 'max_features': conf[10], 'alpha': conf[11], 
                                                        'max_leaf_nodes': conf[12], 'ccp_alpha': conf[13],
                                                        'n_estimators': n_estimators, 'random_state': 1234}) #
        forecaster_reg.fit(X = df[feat_cols], y = df[pred_col])
        REG_CACHE[reg_id] = forecaster_reg

    # forecaster_reg.n_estimators += n_estimators
    #forecaster_reg.fit(X = df[feat_cols], y = df[pred_col])

    y_pred = forecaster_reg.predict(tbd_df[feat_cols])
    return y_pred

def _optimize_GradientBoostingRegressor(_df, feat_cols, pred_col, _tbd_df, conf):
    df = _df#.copy()
    tbd_df = _tbd_df#.copy()
    y_pred = _do_GradientBoostingRegressor(df, feat_cols, pred_col, tbd_df, conf)
    y_true = tbd_df[pred_col]
    mse = mean_squared_error(y_true, y_pred)
    rmse = mean_squared_error(y_true, y_pred, squared=False)
    #print(f"{group=} {mse=} {rmse=} {conf=}")
    return{"conf":conf,"mse":mse,"rmse":rmse}


def optimize_GradientBoostingRegressor(df, feat_cols, pred_col, tbd_df, group=""):
    loss=['squared_error', 'absolute_error', 'huber', 'quantile']
    learning_rate=[0.1, 0.2]
    n_estimators=[25, 50, 75, 100]
    subsample=[0.9, 1.0]
    criterion = ["squared_error", "friedman_mse"]
    min_samples_split = [2,8]
    min_samples_leaf = [1,4]
    min_weight_fraction_leaf = [0.0]
    max_depth = [None, 14]
    min_impurity_decrease  = [0.0]
    max_features = ["sqrt", None]
    alpha=[0.9, 0.95]
    max_leaf_nodes = [None, 10, 15]
    ccp_alpha = [0.0]

    loss=['absolute_error']
    learning_rate=[0.1]
    n_estimators=[50, 100, 150]
    subsample=[1.0]
    criterion = ["friedman_mse"]
    max_depth = [None]
    max_features = [None]
    max_leaf_nodes = [None, 10]
    alpha=[0.9]


    c = list(itertools.product(loss, learning_rate, n_estimators, subsample, criterion, min_samples_split, min_samples_leaf, min_weight_fraction_leaf, max_depth, min_impurity_decrease, max_features, alpha, max_leaf_nodes, ccp_alpha))
    #c = [('absolute_error', 0.1, 100, 1.0, 'friedman_mse', 8, 4, 0.0, None, 0.0, None, 0.9, 10, 0.0), ]

    ret_ds = []
    results=[]
    print()
    with ThreadPoolExecutor(NUM_THREADS) as pool:
    # with ProcessPoolExecutor(NUM_THREADS) as pool:
        for conf in tqdm(c, desc="conf "+group, disable=True):
            results.append(pool.submit(_optimize_GradientBoostingRegressor, df, feat_cols, pred_col, tbd_df, conf))
            #_optimize_DecisionTreeRegressor(df, feat_cols, pred_col, tbd_df,conf, ret_ds)
        for r in tqdm(results, desc="conf "+group):
            ret_ds.append(r.result())   
    return ret_ds


def _do_LinearSVR(df, feat_cols, pred_col, tbd_df, conf):
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=ConvergenceWarning)
        #forecaster_reg = LinearSVR(**{'epsilon': conf[0], 'tol': conf[1], 'C': conf[2], 'loss': conf[3], 'fit_intercept': conf[4], 'intercept_scaling': conf[5], 'dual': conf[6], 'max_iter': conf[7], 'random_state': 1234}) #
        
        # print(feat_cols, pred_col)
        # exit()
        reg_id = get_reg_id("LinearSVR", df, feat_cols, pred_col, conf)
        if reg_id in REG_CACHE:
            logger.debug(f"hit {len(REG_CACHE)}")
            forecaster_reg = deepcopy(REG_CACHE[reg_id])
        else:
            logger.debug(f"miss {len(REG_CACHE)}")
            forecaster_reg = LinearSVR(**{'epsilon': conf[0], 'tol': conf[1], 'C': conf[2], 'loss': conf[3], 'fit_intercept': conf[4], 'intercept_scaling': conf[5], 'dual': conf[6], 'max_iter': conf[7], 'random_state': 1234}) #
            forecaster_reg.fit(X = df[feat_cols], y = df[pred_col])
            REG_CACHE[reg_id] = forecaster_reg

        # forecaster_reg.n_estimators += n_estimators
        #forecaster_reg.fit(X = df[feat_cols], y = df[pred_col])

        y_pred = forecaster_reg.predict(tbd_df[feat_cols])
        return y_pred

def _optimize_LinearSVR(_df, feat_cols, pred_col, _tbd_df, conf):
    df = _df#.copy()
    tbd_df = _tbd_df#.copy()
    y_pred = _do_LinearSVR(df, feat_cols, pred_col, tbd_df, conf)
    y_true = tbd_df[pred_col]
    mse = mean_squared_error(y_true, y_pred)
    rmse = mean_squared_error(y_true, y_pred, squared=False)
    #print(f"{group=} {mse=} {rmse=} {conf=}")
    return {"conf":conf,"mse":mse,"rmse":rmse}


def optimize_LinearSVR(df, feat_cols, pred_col, tbd_df, group=""):
    epsilon=[0.0, 100, 1000]
    tol=[0.0001, 0.01, 0.3]
    C=[1.0, 0.1, 2.0]
    loss=["epsilon_insensitive", "squared_epsilon_insensitive"]
    fit_intercept=[True, False]
    intercept_scaling=[1.0, 2.0]
    dual=[True, False]
    max_iter=[1000, 1200]

    C=[1.0, 2.0]
    loss=["squared_epsilon_insensitive"]

    c = list(itertools.product(epsilon, tol, C, loss, fit_intercept, intercept_scaling, dual, max_iter))
    #c = [(1000, 0.0001, 2.0, 'squared_epsilon_insensitive', False, 1.0, False, 1200), ]

    ret_ds = []
    results=[]
    print()
    # for conf in tqdm(c, desc="conf "+group, disable=False):
    #     if conf[4] == False and conf[5] != 1.0: #not fit_intercept and yes scaling
    #         continue
    #     if conf[3] == 'epsilon_insensitive' and conf[6] == False: #Parameters: penalty='l2', loss='epsilon_insensitive', dual=False
    #         continue
    #     try:
    #         ret_v = _optimize_LinearSVR(df, feat_cols, pred_col, tbd_df, conf)
    #         ret_ds.append(ret_v)
    #     except Exception as e:
    #         print(e)
        
    with ThreadPoolExecutor(NUM_THREADS) as pool:
    # with ProcessPoolExecutor(NUM_THREADS) as pool:
        for conf in tqdm(c, desc="conf "+group, disable=True):
            if conf[4] == False and conf[5] != 1.0: #not fit_intercept and yes scaling
                continue
            results.append(pool.submit(_optimize_LinearSVR, df, feat_cols, pred_col, tbd_df, conf))
            #_optimize_DecisionTreeRegressor(df, feat_cols, pred_col, tbd_df,conf, ret_ds)
        for r in tqdm(results, desc="conf "+group):
            ret_ds.append(r.result())   
    return ret_ds


def _do_MLPRegressor(df, feat_cols, pred_col, tbd_df, conf):
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=ConvergenceWarning)

        reg_id = get_reg_id("MLPRegressor", df, feat_cols, pred_col, conf)
        if reg_id in REG_CACHE:
            logger.debug(f"hit {len(REG_CACHE)}")
            forecaster_reg = deepcopy(REG_CACHE[reg_id])
        else:
            logger.debug(f"miss {len(REG_CACHE)}")
            forecaster_reg = MLPRegressor(**{'hidden_layer_sizes': conf[0], 'solver': conf[1], 'alpha': conf[2], 'max_iter': conf[3], 'activation': conf[4], 'random_state': 1234}) #
            forecaster_reg.fit(X = df[feat_cols], y = df[pred_col])
            REG_CACHE[reg_id] = forecaster_reg

        y_pred = forecaster_reg.predict(tbd_df[feat_cols])#
    return y_pred

def _optimize_MLPRegressor(_df, feat_cols, pred_col, _tbd_df, conf):
    df = _df#.copy()
    tbd_df = _tbd_df#.copy()
    y_pred = _do_MLPRegressor(df, feat_cols, pred_col, tbd_df, conf)
    y_true = tbd_df[pred_col]
    mse = mean_squared_error(y_true, y_pred)
    rmse = mean_squared_error(y_true, y_pred, squared=False)
    #print(f"{group=} {mse=} {rmse=} {conf=}")
    return {"conf":conf,"mse":mse,"rmse":rmse}

def optimize_MLPRegressor(df, feat_cols, pred_col, tbd_df, group=""):
    hidden_layer_sizes = [(10, 10),(10, 10, 10),(100, ),(10, ),] #,(10, 10, 10, 10), (100, 100, 100),  (140, ), (100, 100),
    solver = ["lbfgs", 'adam']
    alpha = [0.0001, 0.00005, 0.0005, 0.00001] #
    max_iter = [20000, 200000]
    activation=["relu", 'logistic']

    # hidden_layer_sizes = [(100, ),(90, ),(80, ),(110, ),(120, ),] #,(10, 10, 10, 10)
    solver = ["lbfgs"]
    # alpha = [0.00001,0.000015,0.000005] #
    max_iter = [20000]
    activation=["relu"]

    # hidden_layer_sizes = [(120, ),(140, ),(160, ),(180, ),(200, ),] #,(10, 10, 10, 10)
    # alpha = [0.00001,] #

    # hidden_layer_sizes = [(140, ),] #,(10, 10, 10, 10)
    # alpha = [0.00001,0.000011, 0.000009,] #

    c = list(itertools.product(hidden_layer_sizes, solver, alpha, max_iter, activation))
    #c = [((100,), 'lbfgs', 1e-05, 20000, 'relu'), ] 4.363318e+04
    #c = [((140,), 'lbfgs', 1e-05, 20000, 'relu'), ] 4.255844e+04

    results = []
    ret_ds = []
    print()

#
    if True:
        for conf in tqdm(c, desc="conf "+group, disable=(len(c) == 1)):
            res = _optimize_MLPRegressor(df, feat_cols, pred_col, tbd_df, conf)
            ret_ds.append(res)
    else: #slower
        logger.warning(f"mlp w/ {NUM_THREADS}")
        #with ThreadPoolExecutor(NUM_THREADS) as pool:
        with ProcessPoolExecutor(NUM_THREADS) as pool:
            for conf in tqdm(c, desc="conf "+group, disable=True):
                results.append(pool.submit(_optimize_MLPRegressor, df, feat_cols, pred_col, tbd_df, conf))
                #_optimize_DecisionTreeRegressor(df, feat_cols, pred_col, tbd_df,conf, ret_ds)
            for r in tqdm(results, desc="conf "+group):
                ret_ds.append(r.result())   


    return ret_ds


def eval_tuning(x_path, full=False):
    onlyfiles = [f for f in os.listdir(x_path) if os.path.isfile(os.path.join(x_path, f))]

    regex = re.compile(r"^(CP_\d+)_(\w+)\.csv\.gz$")
    eval_tuning_files=defaultdict(lambda: dict())
    for f in tqdm(onlyfiles, desc="eval_tuning", disable=True):
        #cpo_ocpp_data_CP_1_0bc01518df38656accbde55bfb0e38591.csv.gz
        result = regex.search(f)
        if result:
            f_df = pd.read_csv(os.path.join(x_path, f), index_col=0)
            eval_tuning_files[result.group(2)][result.group(1)] = f_df
        else:
            logger.info(f"unk file {os.path.join(x_path, f)}")
            continue

    if full:
        return eval_tuning_files

    ret_dict=dict()
    for reg,group_d in eval_tuning_files.items():
        for group, df in group_d.items():
            df["rmse_n"] = (df["rmse"] - df["rmse"].min()) /  (df["rmse"].max() - df["rmse"].min())
            #print(reg,group,df[df["rmse"] == df["rmse"].min()])
        sum_df = pd.concat([df for group, df in group_d.items()])
        sum_df = sum_df.groupby(by="conf").sum() / len(group_d)
        best_df = sum_df[sum_df["rmse_n"] == sum_df["rmse_n"].min()]
        best_df = sum_df[sum_df["rmse"] == sum_df["rmse"].min()]
        # print(reg,best_df)
        for i,r in best_df.iterrows():
            ret_dict[reg] = {"conf":i, "eval":r}

    return ret_dict


def get_eval_dicts(OutDataDIR):
    ret_d=dict()
    DIR=OutDataDIR+"/results/"
    sub_folders = [name for name in os.listdir(DIR) if os.path.isdir(os.path.join(DIR, name)) if name.startswith("CPO_")]
    for cpo in sub_folders:
        ret_dict=eval_tuning(OutDataDIR+"/results/"+cpo)
        ret_d[cpo] = ret_dict

    sub_folders = [name for name in os.listdir(DIR) if os.path.isdir(os.path.join(DIR, name)) if name.startswith("DSO")]
    for dso in sub_folders:
        ret_dict=eval_tuning(OutDataDIR+"/results/"+dso)
        ret_d[dso] = ret_dict

    return ret_d

def get_cp_group_eval_dicts(OutDataDIR):
    ret_d=dict()
    DIR=OutDataDIR+"/results/"
    sub_folders = [name for name in os.listdir(DIR) if os.path.isdir(os.path.join(DIR, name)) if name.startswith("CPO_")]
    for cpo in sub_folders:
        ret_dict=eval_tuning(OutDataDIR+"/results/"+cpo, full=True)
        ret_d[cpo] = ret_dict

    sub_folders = [name for name in os.listdir(DIR) if os.path.isdir(os.path.join(DIR, name)) if name.startswith("DSO")]
    for dso in sub_folders:
        ret_dict=eval_tuning(OutDataDIR+"/results/"+dso, full=True)
        ret_d[dso] = ret_dict

    return ret_d


def get_prediction_dicts(OutDataDIR):
    regex = re.compile(r"^(CP_\d+)_(\w+)_(\d+).csv.gz$")

    ret_d=defaultdict(lambda:defaultdict(lambda:list()))
    DIR=OutDataDIR+"/predictions/"
    sub_folders = [name for name in os.listdir(DIR) if os.path.isdir(os.path.join(DIR, name))]
    for actor in sub_folders:
        if ".bak" in actor: # or ".all" in actor:
            logger.warning(f'skiping {OutDataDIR+"/predictions/"+actor}')
            continue
        if len(actor.split(".")) > 1:
            features = actor.split(".")[1:]
        else:
            features=["all"]
            logger.warning(f"assuming all features for {actor}")
        x_path=OutDataDIR+"/predictions/"+actor
        onlyfiles = [f for f in os.listdir(x_path) if os.path.isfile(os.path.join(x_path, f))]
        for f in onlyfiles:
            if ".bak" in f or ".test" in f: # or ".all" in actor:
                logger.warning(f'skiping {OutDataDIR+"/predictions/"+actor} {f}')
                continue
            result = regex.search(f)
            if result:
                result.group(1)
                ret_d[actor][result.group(1)].append({"reg":result.group(2),"shifts":result.group(3),"file":os.path.join(x_path, f),"features":features})
            else:
                logger.error(f"unk file {f}")
                continue
    return ret_d


def get_reg_id(reg_n, df, feat_cols, pred_col, conf):
    h = hashlib.new('sha256')
    h.update(repr(conf).encode("utf-8"))
    h.update(pd.util.hash_pandas_object(df, index=True).values)
    h.update(repr(sorted(feat_cols)).encode("utf-8"))
    h.update(repr(pred_col).encode("utf-8"))
    ha = h.hexdigest()
    reg_id = (reg_n, ha)
    return reg_id
