
import numpy as np
from tqdm import tqdm
import ujson as json

from wattson.util import get_logger
import os
import pandas as pd
from wattson.util.progress_printer import ProgressPrinter

from wattson.util.random import Random

class CPDataInterface:
    def __init__(
        self,
        dataset,
        start_time,
    ):
        self.logger = get_logger("CPDataInterface", "CPDataInterface")
        self._dataset_orig = dataset
        self.dataset = dataset
        self.start_time = start_time
        self.supported_ds = ["elaadnl", "acn_office", "acn_caltech", "acn_jpl"]
        self.add_cutoff_to_end = True

        if self.dataset not in self.supported_ds and not self.dataset.endswith("_pre"):
            raise ValueError(f"{self.dataset} not in {self.supported_ds}")
        self.baseDir = os.path.dirname(__file__)
        self._load_data()
        self._shift_year()

    def _shift_year(self):
        self.logger.info(f'start&end times pre _shift_year {self.data["UTCTime"].min(), self.data["UTCTime"].max()}')
        y_diff = self.start_time.year - self.data["UTCTime"].min().year
        pd_y_diff = pd.offsets.DateOffset(years=y_diff)
        #t_diff = pd.Timedelta(days=y_diff*365)
        self.data["UTCTransactionStart"] += pd_y_diff
        self.data["UTCTransactionStop"] += pd_y_diff
        self.data["UTCTime"] += pd_y_diff
        data_from_start = self.data[self.data["UTCTransactionStart"] >= self.start_time]

        data_prev_start = self.data[self.data["UTCTransactionStart"] < self.start_time].copy() #pre cutoff
        self.logger.info(f'{len(data_prev_start)} data points pre cutoff')
        if self.add_cutoff_to_end and len(data_prev_start) > 0:
            pd_y_diff = pd.offsets.DateOffset(years=1)
            data_prev_start["UTCTransactionStart"] += pd_y_diff
            data_prev_start["UTCTransactionStop"] += pd_y_diff
            data_prev_start["UTCTime"] += pd_y_diff
            data_prev_start = data_prev_start[data_prev_start["UTCTransactionStart"] > data_from_start["UTCTransactionStart"].max()] #after end of normal data
            self.logger.warning(f'adding {len(data_prev_start)} data points from start (pre cutoff) to end')
            self.logger.warning(f'prev size {len(data_from_start)}')
            data_from_start = pd.concat([data_from_start, data_prev_start])
            self.logger.warning(f'post size {len(data_from_start)}')

        self.data = data_from_start
        self.data = self.data[self.data["UTCTransactionStart"] >= self.start_time]
        self.data = self.data.sort_values(by='UTCTime').reset_index(drop=True).copy()
        self.logger.info(f'start&end times post _shift_year {self.data["UTCTime"].min(), self.data["UTCTime"].max()}')

    def _load_data(self):
        if self.dataset == "elaadnl":
            #TransactionId	ChargePoint	Connector	UTCTime	Collectedvalue	EnergyInterval	AveragePower	cp_c	
            # id	UTCTransactionStart	UTCTransactionStop	StartCard	ConnectedTime	ChargeTime	TotalEnergy	MaxPower
            file = self.baseDir + "/cp_data/elaadnl_open_ev_datasets.xlsx"
            df_tx = pd.read_excel(file, sheet_name="open_transactions", parse_dates=["UTCTransactionStart", "UTCTransactionStop"])#, infer_datetime_format=True
            df_mv = pd.read_excel(file, sheet_name="open_metervalues", parse_dates=["UTCTime"])#, infer_datetime_format=True

            df_tx["cp_c"] = df_tx["ChargePoint"].astype(str) + df_tx["Connector"].astype(str) 
            df_mv["cp_c"] = df_mv["ChargePoint"].astype(str) + df_mv["Connector"].astype(str) 
            df_tx["id"] = df_tx["cp_c"].astype(str) + df_tx["TransactionId"].astype(str) 
            df_mv["id"] = df_mv["cp_c"].astype(str) + df_mv["TransactionId"].astype(str) 

            for c in ['UTCTransactionStart','UTCTransactionStop','StartCard','ConnectedTime','ChargeTime','TotalEnergy','MaxPower']:
                df_mv[c] = df_mv["id"].map(df_tx.set_index("id")[c])

            df_mv_txs = None
            for txid in tqdm(df_mv["TransactionId"].drop_duplicates(), desc=f"adding IntervalFactor {self.dataset}"):
                df_mv_tx = df_mv[df_mv["TransactionId"] == txid].copy()
                df_mv_tx = df_mv_tx.sort_values(by='UTCTime')
                time_diff = df_mv_tx["UTCTime"] - df_mv_tx["UTCTime"].shift(1)
                df_mv_tx["IntervalFactor"] = (time_diff.dt.total_seconds() / (60*60)).shift(-1).fillna(0)
                if df_mv_txs is None:
                    df_mv_txs = df_mv_tx
                else:
                    df_mv_txs = pd.concat([df_mv_txs, df_mv_tx])
            df_mv = df_mv_txs

            #df_mv["IntervalFactor"] = df_mv["EnergyInterval"] / df_mv["AveragePower"] #nan handled in get_data
            #df_mv["IntervalFactor"] = df_mv["EnergyInterval"].div(df_mv["AveragePower"].where(df_mv["AveragePower"] != 0, np.nan)) #nan handled in get_data

            # y_diff = self.start_time.year - df_mv["UTCTransactionStart"].min().year
            # t_diff = pd.Timedelta(days=y_diff*365)
            # df_mv["UTCTransactionStart"] += t_diff
            # df_mv["UTCTransactionStop"] += t_diff
            # df_mv["UTCTime"] += t_diff
            # df_mv = df_mv[df_mv["UTCTransactionStart"] >= self.start_time]
            self.data = df_mv.sort_values(by='UTCTime').reset_index(drop=True).copy()
        
        elif self.dataset == "acn_office" or self.dataset == "acn_caltech" or self.dataset == "acn_jpl":
            #Index(['userInputs', 'userID', 'sessionID', 'stationID', 'spaceID', 'siteID',
            #    'clusterID', 'connectionTime', 'disconnectTime', 'kWhDelivered',
            #    'pilotSignal', 'chargingCurrent', 'doneChargingTime', 'timezone'],

            if self.dataset == "acn_office":
                acn_file = "office001_cp_charges_full.gz"
            elif self.dataset == "acn_caltech":
                acn_file = "caltech_cp_charges_full.gz"
            elif self.dataset == "acn_jpl":
                acn_file = "jpl_cp_charges_full.gz"
            
            file = self.baseDir + "/cp_data/" + acn_file
            df = pd.read_csv(file, parse_dates=["connectionTime", "disconnectTime", "doneChargingTime"], infer_datetime_format=True) #, index_col="_id"
            #df = pd.read_csv("/home/dk/Downloads/"+acn_file, parse_dates=["connectionTime", "disconnectTime", "doneChargingTime"], infer_datetime_format=True) #, index_col="_id"
            
            df = df.drop('userInputs', axis=1)
            df = df.drop('pilotSignal', axis=1)
            df = df.drop('spaceID', axis=1)
            df = df.drop('siteID', axis=1)
            df = df.drop('clusterID', axis=1)
            df = df.drop('timezone', axis=1)

            df = df.dropna(subset=["chargingCurrent"])
            df = df.sort_values("connectionTime")
        
            # df["cp_c"] = df["stationID"].map(str) + df["spaceID"].map(str) + df["siteID"].map(str) + df["clusterID"].map(str)
            df.rename(columns={"_id": "TransactionId", "stationID": "cp_c", "connectionTime": "UTCTransactionStart", "disconnectTime": "UTCTransactionStop"}, inplace=True)
            #print(df)
            df["UTCTransactionStart"] = df["UTCTransactionStart"].dt.tz_localize(None)
            df["UTCTransactionStop"] = df["UTCTransactionStop"].dt.tz_localize(None)
            df["doneChargingTime"] = df["doneChargingTime"].dt.tz_localize(None)
            start_time = pd.to_datetime("2019-12-01 00:00:00+00:00").tz_localize(None)
            end_time = pd.to_datetime("2021-01-31 00:00:00+00:00").tz_localize(None)
            df = df[(df["UTCTransactionStart"] >= start_time) & (df["UTCTransactionStart"] <= end_time)]
        
            df_mv_l = []
            #df_dict = dict((k,v) for k,v in df.groupby("cp_c"))
            #for k,v in tqdm(df_dict.items(), desc=f"loading dataset {self.dataset}"):
            for cp_c in tqdm(df["cp_c"].drop_duplicates(), desc=f"loading dataset {self.dataset}"):
                v = df[df["cp_c"] == cp_c]
                start_time = pd.to_datetime("2019-12-01 00:00:00+00:00").tz_localize(None)
                end_time = pd.to_datetime("2021-01-31 00:00:00+00:00").tz_localize(None)
                v = v[(v["UTCTransactionStart"] >= start_time) & (v["UTCTransactionStart"] <= end_time)].copy()
                #print(v)
        
                v["chargingCurrent"] = v["chargingCurrent"].str.replace("'",'"').apply(lambda x: json.loads(x))
                v_s = v['chargingCurrent'].apply(pd.Series)
                v["UTCTime"] = v_s["timestamps"]
                v["current"] = v_s["current"]
                v = v.drop('chargingCurrent', axis=1)
                v = v.explode(["UTCTime", "current"], ignore_index=True)

                # print(v)
                v['_id2'] = v['TransactionId'].shift(periods=1)
                v['_id3'] = v['TransactionId'].shift(periods=-1)
                v['TX_Status'] = "cont"
                v.loc[(v['_id2'] != v['TransactionId']),'TX_Status'] = "start"
                v.loc[(v['_id3'] != v['TransactionId']),'TX_Status'] = "stop"
                v = v.drop('_id2', axis=1)
                v = v.drop('_id3', axis=1)
                v = v.loc[(v["current"].shift(1) != v["current"]) | (v["TX_Status"].shift(1) != v["TX_Status"])] #remove non-changing...

                v["UTCTime"] = pd.to_datetime(v["UTCTime"], infer_datetime_format=True).dt.tz_localize(None)
        
                #208V Default value in:
                #https://github.com/zach401/acnportal/blob/1ef0d9e37cd89c5f7bd4754615328a6a157152b4/acnportal/acnsim/network/sites/office001_acn.py#L7
                #https://github.com/zach401/acnportal/blob/1ef0d9e37cd89c5f7bd4754615328a6a157152b4/acnportal/acnsim/network/sites/caltech_acn.py#L7
                #https://github.com/zach401/acnportal/blob/1ef0d9e37cd89c5f7bd4754615328a6a157152b4/acnportal/acnsim/network/sites/jpl_acn.py#L33
                v["AveragePower"] = v["current"] * 208 / 1000 # A * V = W # in kW
                #
                v["MaxPower"] = v["AveragePower"].max()
                v["charging_cap"] = np.NaN
                v.loc[(v["MaxPower"] < 22.0), "charging_cap"] = 22.0
                v.loc[(v["MaxPower"] < 11.0), "charging_cap"] = 11.0
                v.loc[(v["MaxPower"] < 3.8), "charging_cap"] = 3.8
        
                v["timestamp_diff"] = (v["UTCTime"] - v["UTCTime"].shift(periods=1)).dt.total_seconds() / (60 * 60)
        
                v["EnergyInterval"] = v['AveragePower'] * v["timestamp_diff"] # in kWh
                v["EnergyInterval"] = v["EnergyInterval"].fillna(0.0)

                df_mv_txs = None
                for txid in tqdm(v["TransactionId"].drop_duplicates(), desc=f"adding IntervalFactor {self.dataset}", disable=True):
                    df_mv_tx = v[v["TransactionId"] == txid].copy()
                    df_mv_tx = df_mv_tx.sort_values(by='UTCTime')
                    time_diff = df_mv_tx["UTCTime"] - df_mv_tx["UTCTime"].shift(1)
                    df_mv_tx["IntervalFactor"] = (time_diff.dt.total_seconds() / (60*60)).shift(-1).fillna(0)
                    if df_mv_txs is None:
                        df_mv_txs = df_mv_tx
                    else:
                        df_mv_txs = pd.concat([df_mv_txs, df_mv_tx])
                v = df_mv_txs

                # v["IntervalFactor"] = v["EnergyInterval"] / v["AveragePower"] #nan handled in get_data
                #v["IntervalFactor"] = v["EnergyInterval"].div(v["AveragePower"].where(v["AveragePower"] != 0, np.nan)) #nan handled in get_data

                v = v.drop('timestamp_diff', axis=1)
                df_mv_l.append(v)
            df_mv = pd.concat(df_mv_l)
            start_time2 = pd.to_datetime("2020-01-01 00:00:00+00:00").tz_localize(None)
            end_time2 = pd.to_datetime("2020-12-31 00:00:00+00:00").tz_localize(None)
            df_mv = df_mv[(df_mv["UTCTime"] >= start_time2) & (df_mv["UTCTime"] <= end_time2)]
            self.data = df_mv.sort_values(by='UTCTime').reset_index(drop=True)#.copy()


        elif self.dataset.endswith("_pre"):
            self.data = pd.read_csv(self.baseDir+"/cp_data/"+self.dataset+"loaded.csv.gz", index_col=0, parse_dates=["UTCTime", "UTCTransactionStart", "UTCTransactionStop"])
            self.dataset = self.dataset.removesuffix("_pre")

            if "MaxPowerMax" in self.data:
                self.data.rename(columns={"MaxPowerMax": "MaxPower",}, inplace=True)
        print(self.data)
        
        if not self._dataset_orig.endswith("_pre"):
            self.data.to_csv(self.baseDir+"/cp_data/"+self.dataset+"_preloaded.csv.gz")

    def get_data_cp_num(self):
        return len(self.data["cp_c"].drop_duplicates())
    
    def get_data(self, cp_num):
        progress_printer = ProgressPrinter(max_progress=cp_num, on_stop_margin=True)
        progress_printer.start()
        _prog=0
        
        cp_dict=None

        random = Random.get_instance("cp_data")
        cps = self.data["cp_c"].drop_duplicates().sample(cp_num, replace=False, random_state=random)
        cp_dict = dict()
        for cp in cps:
            cp_dict[cp] = self.data[self.data["cp_c"] == cp].set_index("UTCTime", drop=True).sort_index()
            cp_dict[cp].index = cp_dict[cp].index.tz_localize(None)
            #cp_dict[cp]["IntervalFactor"].fillna(cp_dict[cp]["IntervalFactor"].mean(), inplace=True) # for simplicity set to average reporting speed
            cp_dict[cp]["OriginalAveragePower"] = cp_dict[cp]["AveragePower"]
            cp_dict[cp]["OriginalEnergyInterval"] = cp_dict[cp]["EnergyInterval"]
            cp_dict[cp]["atk_type"] = "None"
            _prog+=1
            progress_printer.set_progress(_prog)
        
        progress_printer.stop()

        #{cpid : DataFrame(index,            #=Timestamp ; first per session =conn time
        #                  "AveragePower",   #in kW; avg consumption speed until next timestamp; (last value per session should be 0)
        #                  "MaxPower",       #max capacity of cp in kW
        #                  "cp_c",           #id of CP + connector == cpid
        #                //"Collectedvalue", #meter reading in Wh -> just simulated (no longer needed)
        #                  "EnergyInterval", #energy consumption until next timestamp in kWh; (last value per session should be 0)
        #                  "IntervalFactor", #EnergyInterval / AveragePower
        #                  "TransactionId",  #session ID string),...}
        #
        #          "OriginalEnergyInterval", # keep track of "EnergyInterval" before attacks; set during get_data
        #            "OriginalAveragePower", # keep track of "AveragePower" before attacks; set during get_data
        #                  "atk_type",       # keep track of attacks; set during get_data
        return cp_dict
