
import datetime
from pathlib import Path
from typing import Optional, Union
import typing
import pandas as pd
from powerowl.layers.powergrid.values.units.unit import Unit
from powerowl.layers.powergrid.values.grid_value_context import GridValueContext

from powerowl.layers.powergrid import PowerGridModel
from wattson.powergrid.profiles.profile_provider_interface import PowerProfileProviderInterface
from powerowl.layers.powergrid.elements import GridElement
from powerowl.layers.powergrid.elements.storage import Storage

from wattson.util.random import Random
from wattson.util.progress_printer import ProgressPrinter

from tqdm import tqdm

class CPPowerProfileProviderInterface(PowerProfileProviderInterface):
    def __init__(
        self,
        grid_model: PowerGridModel,
        profiles: dict,
        seed: int = 0,
        noise: str = "0",
        interpolate: Union[bool, str] = False,
        logger=None,
        step_size_sec: int = 300, # ?= wattson_time_speed (no)
        step_interpolation_type: str = "linear",
        base_dir: Optional[Path] = None,
    ):
        super().__init__(grid_model, profiles, seed, noise, interpolate, logger, step_size_sec, step_interpolation_type, base_dir)
        self.logger.info(f"CPPowerProfileProviderInterface {profiles, seed, noise, interpolate, step_size_sec, step_interpolation_type, base_dir}")
        self.last_values = dict()
        self.last_log = None
        #self.log_every_x_sec = 60 * 60 * 24
        self.total_sec = 60 * 60 * 24 * 365 #1year; only for visuals
        self.total_steps = self.total_sec / step_size_sec
        self.progress_printer = None
        self.progress_printer_fac = 1 / (60 * 60 * 24)
        self.progress_printer_fac_s = "Day"
        # self.progress_printer = ProgressPrinter(max_progress=self.total_steps, on_stop_margin=True)
        # self.progress_printer.start()
        self.disable_progress_printer=False
        self.print_cp = sorted(self.grid_model._kwargs["cp_data_dicts"].keys())[0]


    """PPowerProfileProviderInterface ({'type': 'storage', 'index': 9, 'attributes': 
    {<GridValueContext.PROPERTY: 'PROPERTY'>: {'minimum_reactive_power': None Var, 'maximum_reactive_power': None Var, 'minimum_active_power': 0.0 W, 
        'maximum_active_power': 400000.0 W, 'minimum_charge': 200.0 Wh, 'maximum_charge': 45000.0 Wh, 'nominal_power': nan VA, 'controllable': False, 
        'type': 'CP', 'bus': <powerowl.layers.powergrid.elements.bus.Bus object at 0x7fc3905bac50>},
    <GridValueContext.CONFIGURATION: 'CONFIGURATION'>: {'state_of_charge': 0 %, 'target_reactive_power': 0.0 Var, 'target_active_power': 300000.0 W, 
        'cos_phi': None, 'in_service': True, 'scaling': 0.9}, 
    <GridValueContext.MEASUREMENT: 'MEASUREMENT'>: {'state_of_charge': 0.0 %, 'reactive_power': 0.0 Var, 'active_power': 270000.0 W}, 
    <GridValueContext.ESTIMATION: 'ESTIMATION'>: {'state_of_charge': 0 %, 'reactive_power': nan Var, 'active_power': nan W},
    <GridValueContext.GENERIC: 'GENERIC'>: {'observable': True, 'name': 'CP 7'}}}, 
    'CP 7', 
    datetime.datetime(2023, 3, 24, 10, 6, 55, 144293, tzinfo=datetime.timezone(datetime.timedelta(seconds=3600), 'CET')), 
    'active_power')
    """
    def get_value(self, element: GridElement, date_time, dimension: str = "active_power") -> Optional[float]:
        if dimension != "active_power":
            raise Exception(f"unsupported dimension {dimension}")
        if element.get_measurement(dimension).unit != Unit.WATT:
            raise Exception(f"unsupported unit {element.get_measurement(dimension).unit}")
        
        
        if type(element) == Storage and element.get_property("type").get_value() == "CP":
            if element.get_generic_value('name') == self.print_cp:
                # if self.last_log == None or (date_time - self.last_log).total_seconds() > (self.log_every_x_sec):
                #     self.logger.info(f"CPPowerProfileProviderInterface currently at {date_time}")
                #     self.last_log = date_time
                if self.progress_printer is None:
                    self.last_time = date_time
                    self.progress_printer = tqdm(desc=f"Current power profile seconds (step_size_sec {self.step_size_sec})", total=self.total_sec * self.progress_printer_fac, initial=0,dynamic_ncols=True,unit_scale=True,smoothing=0.1,unit=self.progress_printer_fac_s,disable=self.disable_progress_printer)
                    #self.progress_printer = tqdm(desc=f"Current power step count (step_size {self.step_size_sec})", total=self.total_steps, initial=0,dynamic_ncols=True,unit_scale=True,smoothing=0.1,unit="steps",disable=self.disable_progress_printer) #doesn't work... seems like same CP only every 2 seconds??
                if self.progress_printer.n >= self.progress_printer.total:
                    self.progress_printer.close() # only closes once
                elif self.progress_printer is not None and not self.progress_printer.disable:
                    progress = (date_time - self.last_time).total_seconds()
                    self.progress_printer.update(progress * self.progress_printer_fac)
                    self.last_time = date_time
                    #self.progress_printer.update()

            if self.grid_model._kwargs["cp_data_dicts"] == None:
                #self.logger.debug("grid_model cp_data_dicts == None")
                #raise Exception(f"grid_model cp_data_dicts == None")
                return None
            # if (self.grid_model._kwargs['cp_data_dicts']['CP_1']['ba3d45d4bf5714a2d6bc4360466731a31']["AveragePower"] >= 100).any():
            #     raise Exception(f"AveragePower >= 100")
            cp_data = self.grid_model._kwargs["cp_data_dicts"][element.get_generic_value('name')]
            all_cps_data=[]
            tz_date_time = pd.to_datetime(date_time).tz_localize(None)
            if cp_data != None:
                cp_power = 0
                for k,v in cp_data.items():
                    cl_interp = CPPowerProfileProviderInterface._interp(v, tz_date_time)
                    cl_interp_W = cl_interp * 1000 #kw to W
                    cp_power+=cl_interp_W
                    if cl_interp_W > 0:
                        all_cps_data.append((k,cl_interp_W))
                if cp_power > 0 and (element.get_generic_value('name') not in self.last_values or self.last_values[element.get_generic_value('name')] != cp_power):
                    # self.logger.debug(f"CPPowerProfileProviderInterface {element.get_generic_value('name'), date_time, dimension, [k for k in element.to_dict()['attributes'][GridValueContext.GENERIC]]}")
                    # self.logger.debug(f"CPPowerProfileProviderInterface cp_data {element.get_measurement(dimension).get_value(), cp_power}")
                    # self.logger.debug(f"CPPowerProfileProviderInterface time {tz_date_time.timestamp(), tz_date_time.isoformat()}")
                    # self.logger.debug(f"CPPowerProfileProviderInterface all_cps_data {all_cps_data}")
                    self.last_values[element.get_generic_value('name')] = cp_power
                return cp_power
            return None
        return super().get_value(element, date_time, dimension)
    
    @staticmethod
    def _interp2(ts, target):
        if target in ts.index:
            return ts[target]
        ts1 = ts.sort_index()
        b = (ts1.index.tz_localize(None) > target).argmax() # index of first entry after target
        s = ts1.iloc[b-1:b+1]
        # Insert empty value at target time.
        # s = s.reindex(pd.to_datetime(list(s.index.values) + [pd.to_datetime(target)]))
        s = s.reindex(pd.to_datetime(list(s.index.values) + [target]))
        return s.interpolate(method='time').loc[target]
    
    @staticmethod
    def _interp(ts, target):
        if target in ts.index:
            return ts["AveragePower"].loc[target]
        if target < ts.index.min() or target > ts.index.max():  #before or after table
            return 0.0
        ts1 = ts.sort_index()
        b = (ts1.index > target).argmax() # index of first entry after target
        s = ts1.iloc[b-1:b+1]
        if len(s)==0: #before or after table
            return 0.0
        if len(s)==2 and s.iloc[0]["TransactionId"] != s.iloc[1]["TransactionId"]: #inbetweeen sessions
            return 0.0
        # Insert empty value at target time.
        s = s.reindex(pd.to_datetime(list(s.index.values) + [target]))
        return s["AveragePower"].interpolate(method='time').loc[target]
