
from collections import defaultdict
from multiprocessing.pool import ThreadPool
from subprocess import PIPE, Popen
import threading
import time
from pathlib import Path
from typing import Any, Optional
import uuid
import zlib
import numpy as np

from powerowl.simulators.pandapower import PandaPowerGridModel
from scenarios.powerowl_example.extensions.ev_charging_extension import MAX_MSG_SIZE, EVChargingExtension

from wattson.apps.interface.clients import CombiClient
from wattson.cosimulation.control.interface.wattson_client import WattsonClient
from wattson.iec104.interface.types import COT
from wattson.powergrid.remote.remote_grid_value import RemoteGridValue
from wattson.powergrid.wrapper.grid_wrapper import GridWrapper
from wattson.powergrid.remote.remote_power_grid_model import RemotePowerGridModel
from wattson.time import WattsonTime
from wattson.time.wattson_time_type import WattsonTimeType
from wattson.util import get_logger
from wattson.apps.interface.util import messages
from wattson.util.random import Random
import pandas as pd
import hashlib
import base64
#
CONTROLLER_WAIT_FOR_MTU_PER_RTU_S = 2

import asyncio

from ocpp.v201.enums import RegistrationStatusType
import logging
import websockets
from datetime import datetime

from ocpp.v201 import call
from ocpp.v201 import ChargePoint as cp

from pandas import Series
from powerowl.layers.powergrid.elements.storage import Storage

from contextlib import contextmanager
import pickle


@contextmanager
def non_blocking_lock(lock=threading.Lock()):
    if not lock.acquire(blocking=False):
        raise PermissionError("no lock")
    try:
        yield lock
    finally:
        lock.release()

class CPService:
    def __init__(self,
                 host_ip: str,
                 cp_id: str,
                 cpo_ip: str,
                 cp_data: Series,
                 mtu_ip: str,
                 scaling: float,
                 cp_node_id: str,
                 min_sec_between_meter: float,
                 wattson_client_query_socket: str,
                 wattson_client_publish_socket: str):

        self.host_ip = host_ip
        self.cp_id = cp_id
        self.cpo_ip = cpo_ip
        self.cp_data = cp_data
        self.mtu_ip = mtu_ip
        self.scaling = scaling
        self.cp_node_id=cp_node_id
        self.wattson_client_query_socket = wattson_client_query_socket
        self.wattson_client_publish_socket = wattson_client_publish_socket
        self.min_sec_between_meter = min_sec_between_meter

        self.atkThreadLock = threading.Lock()
        self.cp_dict = dict()

        self.wattson_time: Optional[WattsonTime] = None
        self.power_grid_model: Optional[PandaPowerGridModel] = None
        self.remote_grid_model: Optional[RemotePowerGridModel] = None
        self.grid_wrapper: Optional[GridWrapper] = None

        self.logger = get_logger("CPService")

        self.logger.info(f"CPService is {self.cp_id, self.host_ip, self.cpo_ip}")
        self.logger.info(f"Export wattson_client_query_socket is {self.wattson_client_query_socket}")
        self.logger.info(f"Export wattson_client_publish_socket is {self.wattson_client_publish_socket}")
        self.logger.info(f"Export cp_data cps: {self.cp_data.keys()}")

        self.wattson_client = WattsonClient(query_server_socket_string=self.wattson_client_query_socket,
                                            publish_server_socket_string=self.wattson_client_publish_socket,
                                            client_name="CPService_"+cp_id)

        self.mtu_client = CombiClient(
            "CP_Service",
            mtu_ip=mtu_ip,
            on_cmd_reply=self._on_cmd_reply,
            on_cmd_update=self._on_cmd_update,
            on_dp_update=self._on_dp_update,
            on_update=self._on_update,
            store_dp_update=True
        )

        self._terminate = threading.Event()
        self._terminateAs = []
        self._run=True


    def _on_cmd_reply(self, update: messages.Confirmation, orig_msg: messages.IECMsg):
        pass

    def _on_cmd_update(self, update: messages.Confirmation, orig_msg: messages.IECMsg = None):
        pass

    def _on_update(self, update: messages.IECMsg, orig_msg: messages.IECMsg = None):
        pass

    def _on_dp_update(self, update: messages.ProcessInfoMonitoring, ref_arg=False):
        pass
        # return self.grid_wrapper.handle_measurement(update)
    
    def add_on_grid_value_change_callback(self, grid_value: RemoteGridValue, old_value: Any, new_value: Any):
        if self.cp_id == "CP_1":
            grid_elem = grid_value.get_grid_element()
            if type(grid_elem) == Storage:
                sim_clock_time = self.wattson_time.sim_clock_time()
                self.logger.info(f"add_on_grid_value_change_callback {grid_elem.get_name(),grid_value.name,grid_value.value, datetime.utcfromtimestamp(grid_value.last_changed).isoformat(), datetime.utcfromtimestamp(sim_clock_time).isoformat()})")

        pass

    def sync(self):
        self.logger.info("re-Loading RemotePowerGridModel")
        self.remote_grid_model.synchronize(force=True, block=True)
        self.logger.info("Creating local PowerGridModel")
        self.power_grid_model = PandaPowerGridModel()
        self.power_grid_model.from_primitive_dict(self.remote_grid_model.to_primitive_dict())
        # self.grid_wrapper = GridWrapper(self.power_grid_model, self.datapoints, self.logger.getChild("GridModel"), wattson_time=self.wattson_time)


    def start(self):
        # Wattson Client
        self.logger.info("CPService Connecting Wattson Client...")
        self.wattson_client.start()
        self.wattson_client.require_connection()
        self.wattson_client.register()
        self.logger.info("CPService Connected Wattson Client")

        # Wattson Time
        self.wattson_time = self.wattson_client.get_wattson_time(enable_synchronization=True)

        # Power Grid Models: Remote and local fork
        self.logger.info("Loading RemotePowerGridModel")
        self.remote_grid_model = RemotePowerGridModel(wattson_client=self.wattson_client)
        #self.remote_grid_model.add_on_grid_value_change_callback(self.add_on_grid_value_change_callback)
        self.logger.info("Creating local PowerGridModel")
        self.power_grid_model = PandaPowerGridModel()
        self.power_grid_model.from_primitive_dict(self.remote_grid_model.to_primitive_dict())

        # self.grid_wrapper = GridWrapper(self.power_grid_model, self.datapoints, self.logger.getChild("GridModel"), wattson_time=self.wattson_time)

        # MTU connection
        self.logger.info(f"Connecting to MTU")
        self.mtu_client.start()


        # Wait for MTU
        # rtu_count = len(self.datapoints.keys())
        rtu_count=34
        wait_time = CONTROLLER_WAIT_FOR_MTU_PER_RTU_S * rtu_count
        wait_time = 10
        self.logger.info(f"Waiting for MTU (at most {int(wait_time)}s)")
        self.wattson_client.event_wait(event_name="MTU_READY", timeout=wait_time)

        self.cp_node = self.wattson_client.get_remote_network_node(self.cp_node_id)
        cp_ip = [iface for iface in self.cp_node.get_interfaces() if not iface.is_management][0]
        # self.cp_node.start_pcap(cp_ip)
        self.logger.info(f"CPService ip {cp_ip.get_ip_address()}")

        self.thread_atk = threading.Thread(target = self.atk_sim_server, daemon=True)#, args = (10, )
        self.thread_atk.start()
        #thread.join()

        wait_time = Random.float(lowest=1, highest=5)
        while self._run:
            self.logger.info(f"Waiting {wait_time}")
            if self._terminate.wait(wait_time):
                break
            try:
                #asyncio.run(self.main())
                self.logger.info(f"ThreadPool {len(self.cp_data.keys())}")
                with ThreadPool(len(self.cp_data.keys())) as pool:
                    # call a function on each item in a list and handle results
                    for result in pool.map(self.start_async, self.cp_data.keys()):
                        result.wait()
            except Exception as e:
                self.logger.error(f"self._run {e}")
            break
        self.logger.info("CPService end")

    def start_async(self, cp_name):
        try:
            asyncio.run(self.main(cp_name))
        except Exception as e:
            self.logger.error(f"start_async {e}")
            raise e


    def run_tracert(self):
        p = Popen(['tracepath', self.cpo_ip], stdin=PIPE, stdout=PIPE, stderr=PIPE)
        output, err = p.communicate()
        output_l = output.decode("utf-8").splitlines()
        self.logger.info(f"{output_l=}")
        return output

    def atk_sim_server(self):
        import socket
        try:
            with non_blocking_lock(self.atkThreadLock):

                UDP_IP = [i for i in self.cp_node.get_interfaces() if i.is_management][0].get_ip_address()
                UDP_PORT = 5005
                self.logger.info(f"sock bind: {self.cp_id, self.cp_node_id, str(UDP_IP), UDP_PORT}")

                sock = socket.socket(socket.AF_INET, # Internet
                                    socket.SOCK_DGRAM) # UDP
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                sock.bind((str(UDP_IP), UDP_PORT))
                #sock.setblocking(False)

                while self._run:
                    # try:
                    #     data = await asyncio.wait_for(loop.sock_recv(sock, 1024), timeout=60)
                    # except asyncio.TimeoutError:
                    #     pass
                    data, atk_ip = sock.recvfrom(MAX_MSG_SIZE)
                    # data = data.decode("utf-8")
                    rec_data = pickle.loads(zlib.decompress(data))
                    self.logger.info(f"received message: {rec_data, atk_ip}")
                    if rec_data["msg"] == 'getCPs':
                        out_msg=zlib.compress(pickle.dumps(list(self.cp_data.keys())))
                        EVChargingExtension.send_to_and_check(sock, out_msg, atk_ip, self.logger)
                    elif rec_data["msg"] == 'test':
                        self.logger.info(f"test: {rec_data=}")
                        EVChargingExtension.send_to_and_check(sock, b"test ok", atk_ip, self.logger)
                    elif rec_data["msg"] == 'tracert':
                        self.logger.info(f"tracert: {rec_data=}")
                        ret = self.run_tracert()
                        EVChargingExtension.send_to_and_check(sock, ret, atk_ip, self.logger)
                    elif rec_data["msg"] == 'pwr_atk':
                        self.logger.info(f"pwr_atk: {list(_atk_cp for _atk_cp in rec_data['atk_cps'] if _atk_cp in self.cp_data.keys())}")
                        EVChargingExtension.send_to_and_check(sock, b"ok", atk_ip, self.logger) #early ok to avoid delays/timeouts

                        s = rec_data["start"]
                        e = rec_data["end"]
                        for _atk_cp in rec_data['atk_cps']:
                            if _atk_cp not in self.cp_data:
                                self.logger.debug(f"pwr_atk cp not found: {_atk_cp}")
                                continue

                            self.logger.debug(f"pwr_atk for cp: {_atk_cp}")
                            cp_df = self.cp_data[_atk_cp]
                            EVChargingExtension.do_pwr_atk(cp_df, s, e, set_to=rec_data["set_to"], add=rec_data["add"], fac=rec_data["fac"], ignore_limits=rec_data["ignore_limits"], atk_type=rec_data["atk_type"])
                            # if _atk_cp == "ba3d45d4bf5714a2d6bc4360466731a31":
                            #     self.logger.info(cp_df.loc[s:e, "AveragePower"])

                        # EVChargingExtension.send_to_and_check(sock, b"ok", atk_ip, self.logger)

                    # data, addr = sock.recvfrom(1024) # buffer size is 1024 bytes
                self.logger.info("atk_sim_server done.")
        except PermissionError as e:
            self.logger.error(f"atk_sim_server {e}")
        except Exception as e:
            self.logger.error(f"atk_sim_server {e}")
            raise e

    def stop(self):
        self._run=False
        self._terminate.set()
        for _terminateA in self._terminateAs:
            _terminateA.set()



    class ChargePoint(cp):
        def __init__(self, id, connection, response_timeout=30, logger=None, end_event=None, cp_data=None, cp_service=None, cp_group=""):
            super().__init__(id, connection, response_timeout=response_timeout)
            if not logger:
                self.logger = get_logger("CPService")
            else:
                self.logger = logger
            self.tx_lock = asyncio.Lock()
            self.end_event = end_event
            self.cp_data = cp_data[self.id]
            self.cp_service = cp_service
            self.wattson_time = self.cp_service.wattson_time
            self.scaling = self.cp_service.scaling
            self.cp_group = cp_group
            self.meter = int(round(Random.float(lowest=1, highest=10000000))) #in W
            self.meter_no_atk = self.meter
            self.background_tasks = set()
            self.last_meter_val_ts = None
            self.started_tx_seq = {}
            self.awaiting_tx = []

        async def send_boot_notification(self):
            request = call.BootNotificationPayload(
                charging_station={
                    'model': 'Wallbox ' +self.cp_group,
                    'vendor_name': 'anewone'
                },
                reason="PowerUp"
            )
            response = await self.call(request)

            if response.status == RegistrationStatusType.accepted:
                self.logger.info(f"Connected to central system. {self.id}")
            return response.interval
        
        async def send_status_notification(self, status="Available"):
            request = call.StatusNotificationPayload(
                    timestamp=datetime.utcnow().isoformat(),
                    connector_status=status,
                    evse_id=1,
                    connector_id=1
            )
            self.logger.info(f"send_status_notification. {self.id}")
            response = await self.call(request)

        async def send_heartbeat(self, interval):
            self.logger.info(f"send_heartbeat init. {self.id}")
            while True:
                # await asyncio.sleep(interval)
                # if self.end_event.wait(interval):
                #     break

                # interval= 5
                # if self.id == "ba3d45d4bf5714a2d6bc4360466731a31":
                #     if (self.cp_data["AveragePower"] >= 100).any():
                #         self.logger.info(f"send_heartbeat. {self.id, self.cp_data[self.cp_data['AveragePower'] >= 100]['AveragePower']}")

                try:
                    await asyncio.wait_for(self.end_event.wait(), timeout=interval)
                    break
                except asyncio.exceptions.TimeoutError:
                    pass

                request = call.HeartbeatPayload()
                self.logger.info(f"send_heartbeat. {self.id}")
                response = await self.call(request)

        async def send_txs(self):
            self.tx_lock = asyncio.Lock()
            self.logger.info("send_txs.")

            tx_changes = [self.cp_data[self.cp_data["TransactionId"] == txid].sort_index(ascending=True) for txid in self.cp_data["TransactionId"].drop_duplicates()]
            self.logger.info(f"send_txs {self.cp_data['cp_c'].iloc[0], len(tx_changes), [i.index[0] for i in tx_changes]}.")
            # for i in tx_changes.index.tz_localize(None):
            for tx in tx_changes:
                self.awaiting_tx.append(tx.index[0])

            for tx in tx_changes:
                event_time = tx.index[0].timestamp()
                if await self.wait_time(event_time): #true if end_event
                    break
                loop = asyncio.get_event_loop()
                task_tx = loop.create_task(self.send_tx(event="Started", tx=tx.iloc[0]))
                self.background_tasks.add(task_tx)
                task_tx.add_done_callback(self.background_tasks.discard)

        #@DeprecationWarning
        async def send_txs2(self):
            self.tx_lock = asyncio.Lock()
            self.logger.info("send_txs.")

            tx_changes = self.cp_data[self.cp_data["TransactionId"].shift() != self.cp_data["TransactionId"]]            
            self.logger.info(f"send_txs {self.cp_data['cp_c'].iloc[0], len(tx_changes), [i for i in tx_changes.index]}.")
            # for i in tx_changes.index.tz_localize(None):
            for index, row in tx_changes.iterrows():
                self.awaiting_tx.append(index)

            for index, row in tx_changes.iterrows():
                event_time = index.timestamp()
                if await self.wait_time(event_time): #true if end_event
                    break
                loop = asyncio.get_event_loop()
                task_tx = loop.create_task(self.send_tx(event="Started", tx=row))
                self.background_tasks.add(task_tx)
                task_tx.add_done_callback(self.background_tasks.discard)
                
        def get_offset_in_sim(self, event_time):
            # sim_clock_time = self.wattson_time.sim_clock_time()
            # sim_clock_time = pd.to_datetime(datetime.fromtimestamp(sim_clock_time)).tz_localize(None).timestamp()
            sim_clock_time = self.wattson_time.to_local_datetime(WattsonTimeType.SIM) #same as ProfileLoader
            sim_clock_time = pd.to_datetime(sim_clock_time).tz_localize(None).timestamp()
            offset = event_time - sim_clock_time
            offset_in_sim = offset / self.wattson_time.speed
            if offset_in_sim < 0:
                offset_in_sim = 0
            return sim_clock_time, offset_in_sim

        async def wait_time(self, event_time, recursive=0, sim_clock_time=None, offset_in_sim=None): #returns true if end_event set
            self.logger.info(f"wait_time (recursion #{recursive}).")
            if sim_clock_time is None or offset_in_sim is None:
                sim_clock_time, offset_in_sim = self.get_offset_in_sim(event_time)
            if recursive == 0:
                self.logger.debug(f"next tx at {event_time} (current = {sim_clock_time}, offset_in_sim = {offset_in_sim}).")
                self.logger.info(f"next tx at {datetime.fromtimestamp(event_time).isoformat()} (current = {datetime.fromtimestamp(sim_clock_time).isoformat()}, offset_in_sim = {offset_in_sim}).")

            if offset_in_sim == 0:
                return False
            try:
                await asyncio.wait_for(self.end_event.wait(), timeout=offset_in_sim)
                return True
            except asyncio.exceptions.TimeoutError:
                if recursive > 5:
                    return False
                return await self.wait_time(event_time, recursive=recursive+1)
            # return self.end_event.wait(offset_in_sim)

        async def just_wait(self, timeout=0.05):
            try:
                await asyncio.wait_for(self.end_event.wait(), timeout=timeout)
                return True
            except asyncio.exceptions.TimeoutError:
                return False

        async def send_tx(self, event="Updated", tx=None): #triggerReason
            # self.logger.info("send_tx.")
            tx_data = self.cp_data[self.cp_data["TransactionId"] == tx["TransactionId"]].copy()

            tx_data["event"] = "Updated"
            tx_data.iloc[0, tx_data.columns.get_loc('event')] = "Started"
            tx_data.iloc[-1,tx_data.columns.get_loc('event')] = "Ended"
            tx_data["triggerReason"] = "MeterValuePeriodic"
            tx_data.iloc[0, tx_data.columns.get_loc('triggerReason')] = "Authorized"
            tx_data.iloc[-1,tx_data.columns.get_loc('triggerReason')] = "StopAuthorized"

            # self.meter = tx_data['Collectedvalue'].iloc[0]
            prev_index = None

            while tx_data.index[0] != min(self.awaiting_tx):
                if await self.just_wait():
                    break

            async with self.tx_lock: #prevent simultanious meter access...
                for index, row in tx_data.iterrows():

                    event_time = index.timestamp()
                    if await self.wait_time(event_time):
                        break

                    if tx_data.index[0] == index:
                        self.awaiting_tx.remove(index)

                    # update CP meter register
                    if prev_index != None:
                        new_meter = self.meter + (1000 * tx_data['EnergyInterval'].loc[prev_index] * self.scaling) #kW to W
                        if np.isnan(new_meter) or np.isinf(new_meter):
                            self.logger.error(f"nan value for {index, prev_index} {self.scaling}")
                            self.logger.error(f"nan value prev_index {tx_data.loc[prev_index]}")
                            self.logger.error(f"nan value index      {tx_data.loc[index]}")
                        else:
                            self.meter = round(new_meter)

                        new_meter_no_atk = self.meter_no_atk + (1000 * tx_data['OriginalEnergyInterval'].loc[prev_index] * self.scaling) #kW to W
                        if np.isnan(new_meter_no_atk) or np.isinf(new_meter_no_atk):
                            pass
                        else:
                            self.meter_no_atk = round(new_meter_no_atk)
                    # print(curr_cv)
                    
                    #skip sending if too fast
                    if self.last_meter_val_ts is None:
                        diff_to_last = self.cp_service.min_sec_between_meter +1
                    else:
                        diff_to_last = (index - self.last_meter_val_ts).total_seconds()
                    
                    if diff_to_last < self.cp_service.min_sec_between_meter:
                        if row["event"] == "Started" or row["event"] == "Ended": #dont skip start/end
                            pass
                        else:
                            continue
                    self.last_meter_val_ts = index


                    sim_clock_time = self.wattson_time.to_local_datetime(WattsonTimeType.SIM) #same as ProfileLoader
                    sim_clock_time = pd.to_datetime(sim_clock_time).tz_localize(None) #.timestamp()
                    # sim_time_iso = datetime.fromtimestamp(sim_clock_time).isoformat()
                    # event_time_iso = datetime.fromtimestamp(event_time).isoformat()
                    sim_time_iso = sim_clock_time.isoformat()
                    event_time_iso = index.isoformat()
                    row_event = row["event"]
                    if row['TransactionId'] not in self.started_tx_seq:
                        self.started_tx_seq[row['TransactionId']] = 0
                        row_event= "Started"
                    self.started_tx_seq[row['TransactionId']] += 1
                    
                    if "StartCard" in row:
                        idToken = base64.b64encode(hashlib.sha1(row["StartCard"].encode("utf-8")).digest()).decode("utf-8")
                    else:
                        idToken = uuid.uuid4().hex
                    if "Connector" in row:
                        connectorId = row["Connector"]
                    else:
                        connectorId = 0

                    self.logger.info(f"send_tx. {row['event'], 'TransactionId', row['TransactionId'], 'event_time_iso', event_time_iso, 'sim_time_iso', sim_time_iso} {self.meter=} {'EnergyInterval', tx_data['EnergyInterval'].loc[prev_index] if prev_index != None else '0 (no previous index)'}")
                    request = call.TransactionEventPayload(
                            event_type=row_event, #"Started", "Ended"
                            timestamp=sim_time_iso,
                            trigger_reason=row["triggerReason"],
                            seq_no=self.started_tx_seq[row['TransactionId']],
                            transaction_info={
                                "transactionId": str(row['TransactionId']),
                                "chargingState": "Charging",
                            },
                            id_token={ #optional
                                'idToken': idToken,
                                'type': 'eMAID'
                            } if row_event == "Started" else None,
                            evse={
                                "id":int(hashlib.sha1(row["cp_c"].encode("utf-8")).hexdigest(), 16) % (10 ** 8),
                                "connectorId":connectorId,
                            },
                            meter_value=[{
                                "timestamp": event_time_iso,
                                "sampledValue":[{
                                    "value":self.meter, #row["Collectedvalue"],
                                    # "signedMeterValue":{
                                    #     "signedMeterData":"",#base64(sig)
                                    #     "signingMethod":"sha256ecdsa",
                                    #     "encodingMethod":"exi",
                                    #     "publicKey":"",#base64
                                    # },
                                },]
                            },],
                            custom_data={"vendorId": "wattson_v2g", "atk_type": "None" if prev_index == None else tx_data['atk_type'].loc[prev_index]}
                    )
                    request.custom_data.update({
                                "meter_no_atk": self.meter_no_atk,
                                "scaling": self.scaling,
                                        })
                    if prev_index != None: # and tx_data['atk_type'].loc[prev_index] != "None":
                        request.custom_data.update({
                                    "EnergyInterval": 1000 * tx_data['EnergyInterval'].loc[prev_index] * self.scaling,
                                    "OriginalEnergyInterval": 1000 * tx_data['OriginalEnergyInterval'].loc[prev_index] * self.scaling,
                                    "AveragePower": tx_data['AveragePower'].loc[prev_index],
                                    "OriginalAveragePower": tx_data['OriginalAveragePower'].loc[prev_index],
                                            })
                    prev_index = index
                    response = await self.call(request)


    async def main(self, cp_name="None"):
        try:
            _terminateA = asyncio.Event()
            self._terminateAs.append(_terminateA)
            wait_time = Random.float(lowest=0, highest=1)
            #await asyncio.sleep(wait_time)
            try:
                await asyncio.wait_for(_terminateA.wait(), timeout=wait_time)
                return
            except asyncio.exceptions.TimeoutError:
                pass
            
            cp_try=0
            async for ws in websockets.connect(
                    'ws://'+self.cpo_ip+':9000/'+cp_name,
                    subprotocols=['ocpp2.0.1']
            ):         
                self.logger.debug(f"main for {cp_name, self.cp_id, cp_try}")
                try:   
                    cp_try+=1
                    cp = self.ChargePoint(cp_name, ws, response_timeout=600, logger=self.logger, end_event=_terminateA, cp_data=self.cp_data, cp_service=self, cp_group=self.cp_id)
                    self.cp_dict[cp_name] = cp

                    loop = asyncio.get_event_loop()
                    cp_ts = loop.create_task(cp.start())
                    #await asyncio.gather(cp.start(), cp.send_boot_notification())
                    interval = await cp.send_boot_notification()
                    await cp.send_status_notification()
                    cp_th = loop.create_task(cp.send_heartbeat(interval))
                    # await cp.send_heartbeat(interval)
                    await cp.send_txs()
                    #cp_tx = loop.create_task(cp.send_txs())
                    await asyncio.gather(cp_ts, cp_th)

                    # await asyncio.gather(cp.start(), cp.send_boot_notification(), cp.send_status_notification(), cp.send_heartbeat(600), cp.send_txs())
                except websockets.ConnectionClosed as e:
                    self.logger.error(f"main for {cp_try,e}")
                if cp_try > 5:
                    break

            self.logger.error(f"main ending after {cp_try, cp_name}")
            if cp_name in self.cp_data:
                del self.cp_data[cp_name]
        except Exception as e:
            self.logger.error(f"main {e=}")
            if cp_name in self.cp_data:
                del self.cp_data[cp_name]
            raise e
