
from collections import defaultdict
import threading
import time
from pathlib import Path
import traceback
from typing import Optional
import zlib
import numpy as np

from powerowl.simulators.pandapower import PandaPowerGridModel
from scenarios.powerowl_example.extensions.ev_charging_extension import MAX_MSG_SIZE, OSCP_INTERVAL_M, EVChargingExtension

from wattson.apps.interface.clients import CombiClient
from wattson.cosimulation.control.interface.wattson_client import WattsonClient
from wattson.hosts.cp.cpo_oscp import CPO_OSCP
from wattson.iec104.interface.types import COT
from wattson.powergrid.wrapper.grid_wrapper import GridWrapper
from wattson.powergrid.remote.remote_power_grid_model import RemotePowerGridModel
from wattson.time import WattsonTime
from wattson.util import get_logger
from wattson.apps.interface.util import messages
from wattson.util.random import Random

CONTROLLER_WAIT_FOR_MTU_PER_RTU_S = 2

import asyncio
import logging
import websockets
from datetime import datetime

from ocpp.routing import on
from ocpp.v201 import ChargePoint as cp
from ocpp.v201 import call_result
from ocpp.v201.enums import RegistrationStatusType


import pandas as pd
from wattson.time.wattson_time_type import WattsonTimeType

import resource
from wattson.util.random import Random
import pickle


class CPOService:
    def __init__(self,
                 mgmt_ip: str, #"!management_ips.ctrl"
                 mtu_ip: str,
                 dso_ip: str,
                 all_cp_ids: list,
                 cpo_node_id: str,
                 oscp_interval_m: int,
                 wattson_client_query_socket: str,
                 wattson_client_publish_socket: str):

        self.mgmt_ip = mgmt_ip
        self.mtu_ip = mtu_ip
        self.dso_ip = dso_ip
        self.all_cp_ids=all_cp_ids
        self.cpo_node_id=cpo_node_id
        self.wattson_client_query_socket = wattson_client_query_socket
        self.wattson_client_publish_socket = wattson_client_publish_socket
        self.oscp_interval_m = oscp_interval_m

        self.last_group_measurements = defaultdict(lambda: [])
        self.group_measurements_total = defaultdict(lambda: 0)
        self.cp_groups = set()
        self.cp_dict=dict()
        self.cp_min_meter=dict()
        self.cp_tx_df = pd.DataFrame(columns=('time', 'meas', 'cp_group', 'cp_c'))
        self.meter_df_lock = threading.Lock()
        self.last_oscp_msg_time = None
        self.last_oscp_msg_time_CP = None

        self.wattson_time: Optional[WattsonTime] = None
        self.power_grid_model: Optional[PandaPowerGridModel] = None
        self.remote_grid_model: Optional[RemotePowerGridModel] = None
        self.grid_wrapper: Optional[GridWrapper] = None


        self.logger = get_logger("CPOService")

        self.logger.info(f"Export wattson_client_query_socket is {self.wattson_client_query_socket}")
        self.logger.info(f"Export wattson_client_publish_socket is {self.wattson_client_publish_socket}")

        self.wattson_client = WattsonClient(query_server_socket_string=self.wattson_client_query_socket,
                                            publish_server_socket_string=self.wattson_client_publish_socket,
                                            client_name="CPOService")
        
        self.cpo_node = self.wattson_client.get_remote_network_node(self.cpo_node_id)
        self.cpo_ip = str([i for i in self.cpo_node.get_interfaces() if not i.is_management][0].get_ip_address())

        self.mtu_client = CombiClient(
            "CPO_Service",
            mtu_ip=mtu_ip,
            on_cmd_reply=self._on_cmd_reply,
            on_cmd_update=self._on_cmd_update,
            on_dp_update=self._on_dp_update,
            on_update=self._on_update,
            store_dp_update=True
        )

        self._rand = Random.get_instance("getSeed")
        self._terminate = threading.Event()
        self._run=True
        self.charge_point_ids=[]


    def _on_cmd_reply(self, update: messages.Confirmation, orig_msg: messages.IECMsg):
        pass

    def _on_cmd_update(self, update: messages.Confirmation, orig_msg: messages.IECMsg = None):
        pass

    def _on_update(self, update: messages.IECMsg, orig_msg: messages.IECMsg = None):
        pass

    def _on_dp_update(self, update: messages.ProcessInfoMonitoring, ref_arg=False):
        pass
        # return self.grid_wrapper.handle_measurement(update)

    def setlimits(self):
        resource.setrlimit(resource.RLIMIT_NOFILE, (262144, 1048576))

    def sync(self):
        self.logger.info("re-Loading RemotePowerGridModel")
        self.remote_grid_model.synchronize(force=True, block=True)
        self.logger.info("Creating local PowerGridModel")
        self.power_grid_model = PandaPowerGridModel()
        self.power_grid_model.from_primitive_dict(self.remote_grid_model.to_primitive_dict())
        # self.grid_wrapper = GridWrapper(self.power_grid_model, self.datapoints, self.logger.getChild("GridModel"), wattson_time=self.wattson_time)


    def start(self):
        # Wattson Client
        self.logger.info("CPOService Connecting Wattson Client...")
        self.wattson_client.start()
        self.wattson_client.require_connection()
        self.wattson_client.register()
        self.logger.info("CPOService Connected Wattson Client")

        # Wattson Time
        self.wattson_time = self.wattson_client.get_wattson_time(enable_synchronization=True)

        # Power Grid Models: Remote and local fork
        self.logger.info("Loading RemotePowerGridModel")
        self.remote_grid_model = RemotePowerGridModel(wattson_client=self.wattson_client)
        self.logger.info("Creating local PowerGridModel")
        self.power_grid_model = PandaPowerGridModel()
        self.power_grid_model.from_primitive_dict(self.remote_grid_model.to_primitive_dict())

        # self.grid_wrapper = GridWrapper(self.power_grid_model, self.datapoints, self.logger.getChild("GridModel"), wattson_time=self.wattson_time)

        # MTU connection
        self.logger.info(f"Connecting to MTU")
        self.mtu_client.start()


        # Wait for MTU
        # rtu_count = len(self.datapoints.keys())
        rtu_count=20
        wait_time = CONTROLLER_WAIT_FOR_MTU_PER_RTU_S * rtu_count
        wait_time = 6
        self.logger.info(f"Waiting for MTU (at most {int(wait_time)}s)")
        self.wattson_client.event_wait(event_name="MTU_READY", timeout=wait_time)

        self.cpo_node.start_pcap([i for i in self.cpo_node.get_interfaces() if not i.is_management][0])

        self.thread_atk = threading.Thread(target = self.atk_sim_server, daemon=True)#, args = (10, )
        self.thread_atk.start()
        #thread.join()

        self.thread_oscp = threading.Thread(target = self.oscp_server, daemon=True)#, args = (10, )
        self.thread_oscp.start()

        while self._run:
            asyncio.run(self.main())
            if self._terminate.wait(1):
                break
        self.logger.info("CPOService end")


    def oscp_server(self):
        self.cpo_server = CPO_OSCP(self, DST_IP=self.dso_ip, HOST_IP=self.cpo_ip)
        self.cpo_server.init_threaded(dso_oob_ip=self.mtu_ip, dso_oob_port = 5005) #get token
        self.cpo_server.init_thread.join()
        
        oscp_interval = (self.oscp_interval_m * 60) / self.wattson_time.speed
        self.cpo_server.start_server_threaded(init_delay=20, init_meas_delay=oscp_interval, meas_interval=oscp_interval)

    def atk_sim_server(self):
        import socket

        UDP_IP = [i for i in self.cpo_node.get_interfaces() if i.is_management][0].get_ip_address()
        UDP_PORT = 5005
        self.logger.info(f"sock bind: {self.cpo_node_id, str(UDP_IP), UDP_PORT}")

        sock = socket.socket(socket.AF_INET, # Internet
                            socket.SOCK_DGRAM) # UDP
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((str(UDP_IP), UDP_PORT))
        #sock.setblocking(False)

        while self._run:
            # try:
            #     data = await asyncio.wait_for(loop.sock_recv(sock, 1024), timeout=60)
            # except asyncio.TimeoutError:
            #     pass
            data, atk_ip = sock.recvfrom(MAX_MSG_SIZE)
            #data = data.decode("utf-8")
            rec_data = pickle.loads(zlib.decompress(data))
            self.logger.info(f"received message: {rec_data, atk_ip}")
            if rec_data["msg"] == 'getSeed':
                getSeed_val = self._rand.randint(low=1,high=10000000)
                out_msg = zlib.compress(pickle.dumps(getSeed_val))
                EVChargingExtension.send_to_and_check(sock, out_msg, atk_ip, self.logger)

            # data, addr = sock.recvfrom(1024) # buffer size is 1024 bytes
        self.logger.info("atk_sim_server done.")

    def stop(self):
        self.cpo_server.RUN=False
        self._run=False
        self.server.close()
        self._terminate.set()
        # self.thread_atk.join()

    @staticmethod
    def _interp(ts, target, col="charge_speed"):
        if target in ts.index:
            return ts[col].loc[target]
        if target < ts.index.min() or target > ts.index.max():  #before or after table
            return 0.0
        ts1 = ts.sort_index()
        b = (ts1.index > target).argmax() # index of first entry after target
        s = ts1.iloc[b-1:b+1]
        if len(s)==0: #before or after table
            return 0.0
        if len(s)==2 and s.iloc[0]["TransactionId"] != s.iloc[1]["TransactionId"]: #inbetweeen sessions
            return 0.0
        # Insert empty value at target time.
        s = s.reindex(pd.to_datetime(list(s.index.values) + [target]))
        return s[col].interpolate(method='time').loc[target]

    @staticmethod
    def _interp2(ts, target, col="charge_speed"):
        if target in ts.index:
            return ts[col].loc[target]
        if target < ts.index.min():  #before  table
            return ts[col].loc[ts.index.min()]
        if target > ts.index.max():
            return ts[col].loc[ts.index.max()]
        ts1 = ts.sort_index()
        b = (ts1.index > target).argmax() # index of first entry after target
        s = ts1.iloc[b-1:b+1]
        # if len(s)==0: #before or after table
        #     return 0.0
        # if len(s)==2 and s.iloc[0]["TransactionId"] != s.iloc[1]["TransactionId"]: #inbetweeen sessions
        #     return 0.0
        # Insert empty value at target time.
        s = s.reindex(pd.to_datetime(list(s.index.values) + [target]))
        return s[col].interpolate(method='time').loc[target]
    
    def get_group_measurements(self):
        sim_clock_time = self.wattson_time.to_local_datetime(WattsonTimeType.SIM) #same as ProfileLoader
        sim_time = pd.to_datetime(sim_clock_time).tz_localize(None) #.isoformat()
        sim_time_max = sim_time + pd.Timedelta(minutes=4*OSCP_INTERVAL_M)
        sim_time_do = sim_time - pd.Timedelta(minutes=1*OSCP_INTERVAL_M)
        start_sim_time = sim_time - pd.Timedelta(minutes=24*4*OSCP_INTERVAL_M)

        group_measurements = defaultdict(lambda: 0)
        group_times = defaultdict(lambda: list())

        # for cp_g in self.cp_groups:
        #     group_measurements[cp_g] = 0
        #     group_times[cp_g].append(sim_time)

        if self.last_oscp_msg_time is None:
            self.last_oscp_msg_time = defaultdict(lambda: sim_time)
            self.last_oscp_msg_time_CP = defaultdict(lambda: sim_time)
            self.cp_tx_df["done"]=True
        else:
            #sim_time_start = self.last_oscp_msg_time
            with self.meter_df_lock:
                # self.logger.info(f"{len(self.cp_tx_df)=}")
                # df = self.cp_tx_df.dropna().copy()
                # df = df[df["time"] >= start_sim_time]
                df = self.cp_tx_df[self.cp_tx_df["time"] >= start_sim_time].dropna().copy()

            for cp_group in df["cp_group"].drop_duplicates():
                df_g = df[df["cp_group"] == cp_group]                    
                df_interval_g = df_g[(df_g["time"] <= sim_time_max)]

                for cp_id in df_interval_g["cp_c"].drop_duplicates():
                    df_interval_id = df_interval_g[df_interval_g["cp_c"] == cp_id]
                    if len(df_interval_id) == 0:
                        continue

                    if True:
                        df_interval_id["meter_diff"] = df_interval_id["meas"] - df_interval_id["meas"].shift(1)
                        df_interval_id["meter_diff"].fillna(0, inplace=True)
                        df_interval_id["time_diff"] = (df_interval_id["time"] - df_interval_id["time"].shift(1)).apply(lambda x: x.total_seconds())
                        df_interval_id["time_diff"] = df_interval_id["time_diff"] / (60*60) #s to h
                        df_interval_id["charge_speed"] = df_interval_id["meter_diff"] / df_interval_id["time_diff"] #wh in w
                        df_interval_id["charge_speed"].fillna(0, inplace=True)
                    df_interval_id = df_interval_id.set_index("time",drop=False).sort_index()
                    # charge_speed = self.interpolate_sim_time(df_interval_id["charge_speed"], start_sim_time)

                    charge_speed = CPOService._interp(df_interval_id, sim_time_do)
                    t_diff = OSCP_INTERVAL_M / 60
                    #t_diff = (sim_time_do - self.last_oscp_msg_time_CP[cp_id]).total_seconds() / (60*60)
                    meas = charge_speed * t_diff
                    #meas = CPOService._interp2(df_interval_id, sim_time_do, col="meas")


                    # if cp_id not in self.cp_min_meter:
                    #     self.cp_min_meter[cp_id] = meas
                    #meas = meas - self.cp_min_meter[cp_id]
                    group_times[cp_group].append(sim_time_do)
                    self.group_measurements_total[cp_group] += meas
                    group_measurements[cp_group] = self.group_measurements_total[cp_group]

                    # tx_df_cp = tx_df_cp.resample('5Min', offset="0s").bfill()

                    self.last_oscp_msg_time[cp_group] = max(self.last_oscp_msg_time[cp_group], sim_time_do)
                    self.last_oscp_msg_time_CP[cp_id] = sim_time_do

        # self.last_oscp_msg_time = sim_time
        return group_measurements, group_times, sim_time_do
    
    def get_group_measurements22(self):
        sim_clock_time = self.wattson_time.to_local_datetime(WattsonTimeType.SIM) #same as ProfileLoader
        sim_time = pd.to_datetime(sim_clock_time).tz_localize(None) #.isoformat()
        sim_time = sim_time - pd.Timedelta(minutes=OSCP_INTERVAL_M)

        group_measurements = defaultdict(lambda: 0)
        group_times = defaultdict(lambda: list())

        # for cp_g in self.cp_groups:
        #     group_measurements[cp_g] = 0
        #     group_times[cp_g].append(sim_time)

        if self.last_oscp_msg_time is None:
            self.last_oscp_msg_time = defaultdict(lambda: sim_time)
            self.cp_tx_df["done"]=True
        else:
            #sim_time_start = self.last_oscp_msg_time
            with self.meter_df_lock:
                # self.logger.info(f"{len(self.cp_tx_df)=}")
                df = self.cp_tx_df.dropna().copy()

                for cp_group in df["cp_group"].drop_duplicates():
                    df_g = df[df["cp_group"] == cp_group]                    
                    df_interval_g = df_g[(df_g["time"] <= sim_time)]

                    for cp_id in df_interval_g["cp_c"].drop_duplicates():
                        df_interval_id = df_interval_g[df_interval_g["cp_c"] == cp_id]
                        if len(df_interval_id) == 0:
                            continue

                        t_max = df_interval_id["time"].max()
                        meas_max = df_interval_id[df_interval_id["time"] == t_max]["meas"].max()
                        if cp_id not in self.cp_min_meter:
                            self.cp_min_meter[cp_id] = meas_max
                        meas = meas_max - self.cp_min_meter[cp_id]
                        group_times[cp_group].append(t_max)
                        group_measurements[cp_group] += meas

                        # tx_df_cp = tx_df_cp.resample('5Min', offset="0s").bfill()

                        self.last_oscp_msg_time[cp_group] = max(self.last_oscp_msg_time[cp_group], t_max)

        # self.last_oscp_msg_time = sim_time
        return group_measurements, group_times, sim_time
    

    def get_group_measurements2(self):
        sim_clock_time = self.wattson_time.to_local_datetime(WattsonTimeType.SIM) #same as ProfileLoader
        sim_time_o = pd.to_datetime(sim_clock_time).tz_localize(None) #.isoformat()
        # sim_time_start = sim_time - pd.Timedelta(minutes=OSCP_INTERVAL_M)
        sim_time = sim_time_o - pd.Timedelta(minutes=2)

        group_measurements = defaultdict(lambda: 0)
        group_times = defaultdict(lambda: list())

        if self.last_oscp_msg_time is None:
            self.last_oscp_msg_time = defaultdict(lambda: sim_time)
            self.cp_tx_df["done"]=True
        else:
            #sim_time_start = self.last_oscp_msg_time
            with self.meter_df_lock:
                # self.logger.info(f"{len(self.cp_tx_df)=}")
                df = self.cp_tx_df.dropna().copy()
                # df_interval = df[(df["time"] > sim_time_start) & (df["time"] <= sim_time)]
                for cp_group in df["cp_group"].drop_duplicates():
                    df_g = df[df["cp_group"] == cp_group]
                    if cp_group not in self.last_oscp_msg_time:
                        self.last_oscp_msg_time[cp_group] = sim_time
                        continue
                    sim_time_start = self.last_oscp_msg_time[cp_group] - pd.Timedelta(minutes=5) #OSCP_INTERVAL_M
                    # df_interval_g = df_g[(df_g["time"] > sim_time_start) & (df_g["time"] <= sim_time)]
                    # df_interval_g = df_interval_g[df_interval_g["done"] == False]
                    # self.cp_tx_df.loc[df_interval_g.index, "done"]=True

                    for cp_id in df_g["cp_c"].drop_duplicates():
                        df_interval_id = df_g[df_g["cp_c"] == cp_id]
                        df_interval_id = df_interval_id[df_interval_id["done"] == False]
                        df_interval_id_index = df_interval_id.index
                        try:
                            # self.logger.error(df_interval_id)
                            df_interval_id = df_interval_id.set_index("time")
                            df_interval_id = df_interval_id.resample('5Min', offset="0s").bfill()
                            df_interval_id_time = df_interval_id.index
                            df_interval_id = df_interval_id.reset_index(drop=True)
                            df_interval_id["time"] = df_interval_id_time
                            # self.logger.error(df_interval_id)

                        except Exception as e:
                            self.logger.error(e)
                            self.logger.error(df_interval_id)
                            raise e

                        df_interval_id = df_interval_id[(df_interval_id["time"] > sim_time_start) & (df_interval_id["time"] <= sim_time)]
                        if len(df_interval_id) == 0:
                            continue
                        self.cp_tx_df.loc[df_interval_id_index, "done"]=True


                        t_min = df_interval_id["time"].min()
                        t_max = df_interval_id["time"].max()
                        meas_min = df_interval_id[df_interval_id["time"] == t_min]["meas"].min()
                        meas_max = df_interval_id[df_interval_id["time"] == t_max]["meas"].max()
                        meter_diff = meas_max - meas_min #in wh
                        # time_diff = (t_max - t_min).total_seconds() / (60*60) #in h
                        # if time_diff == 0:
                        #     meas = 0
                        # else:
                        #     meas = meter_diff / time_diff # in w
                        # if np.isnan(meas):
                        #     self.logger.error(f"np.isnan({meas=}): {df_interval_id}") #empty df
                        #     continue

                        group_times[cp_group].append(t_min)
                        group_times[cp_group].append(t_max)
                        group_measurements[cp_group] += meter_diff

                        # tx_df_cp = tx_df_cp.resample('5Min', offset="0s").bfill()

                        self.last_oscp_msg_time[cp_group] = max(self.last_oscp_msg_time[cp_group], t_max)
                        self.last_oscp_msg_time[cp_group] = max(self.last_oscp_msg_time[cp_group], sim_time - pd.Timedelta(minutes=5))

                    if group_measurements[cp_group] < 0:
                        self.logger.warning(f"group_measurements < 0 {cp_group=}")
                        self.logger.warning(df_g)
         
        last_none_len=3
        for cp_g in self.cp_groups:
            if cp_g not in group_measurements:
                self.last_group_measurements[cp_g].append(None)
                if len(self.last_group_measurements[cp_g]) > last_none_len:
                    if self.last_group_measurements[cp_g][last_none_len:] == [None]*last_none_len:
                        group_measurements[cp_g] = 0
                        group_times[cp_g].append(sim_time)

        for cp_g in group_measurements.keys():
            if len(group_times[cp_g]) == 0:
                group_times[cp_g].append(sim_time)

        # self.last_oscp_msg_time = sim_time
        return group_measurements, group_times, sim_time

        # for cp_id, cp in self.cp_dict.items():
        #     with cp.meter_reg_lock:
        #         if cp.last_meter_reg is not None:
        #             if cp.last_reported_meter_reg is None:
        #                 cp.last_reported_meter_reg = cp.last_meter_reg
        #             else:
        #                 cp_last_meter_reg = cp.last_meter_reg[1]
        #                 cp_last_reported_meter_reg = cp.last_reported_meter_reg[1]
        #                 meter_reg_diff = cp_last_meter_reg - cp_last_reported_meter_reg
        #                 group_measurements[cp.cp_group] += meter_reg_diff
        #                 group_times[cp.cp_group].append(pd.to_datetime(cp.last_meter_reg[0]))
        #                 cp.last_reported_meter_reg = cp.last_meter_reg
        # return group_measurements, group_times



    class ChargePoint(cp):
        def __init__(self, id, connection, response_timeout=600, logger=None, cpo_service=None):
            super().__init__(id, connection, response_timeout=response_timeout)
            if not logger:
                self.logger = get_logger("CPService")
            else:
                self.logger = logger

            self.cpo_service = cpo_service
            self.cp_group = ""
            self.wattson_time = self.cpo_service.wattson_time

            self.meter_reg_lock = threading.Lock()
            self.last_meter_reg = None
            self.last_reported_meter_reg = None
            self._rand_b = Random.get_instance("boot_"+id)

            

        @on('BootNotification')
        async def on_boot_notification(self, charging_station, reason, **kwargs):
            self.logger.info(f"on_boot_notification {self.id, charging_station['model']}")
            self.cp_group = charging_station['model'].split(" ")[-1]
            if self.cp_group not in self.cpo_service.cp_groups:
                self.cpo_service.cp_groups.add(self.cp_group)
            base_time=20*self.wattson_time.speed
            r_wait_time = int(self._rand_b.randint(low=0,high=50)*self.wattson_time.speed)
            #r_wait_time = int(Random.float(lowest=0, highest=50)*self.wattson_time.speed)
            return call_result.BootNotificationPayload(
                current_time=datetime.utcnow().isoformat(),
                interval=int(base_time+r_wait_time),
                # interval=600,
                status=RegistrationStatusType.accepted
            )
        
        @on('StatusNotification')
        def on_status_notitication(self, timestamp, connector_status, evse_id, connector_id, **kwargs):
            # self.logger.info(f"on_status_notitication {self.id}")
            return call_result.StatusNotificationPayload()

        @on('Heartbeat')
        def on_heartbeat(self, **kwargs):
            # q.put(('authReq','MTIzNDU2'))
            self.logger.info(f"on_heartbeat {self.cp_group, self.id}")
            return call_result.HeartbeatPayload(
                current_time=datetime.utcnow().isoformat()
            )
        
        def save_meter_reg(self, meter_value, txID=None): #updates meter register representation
            #meter_value = [{'timestamp': '2023-03-24T18:59:59', 'sampled_value': [{'value': 3612236}]}])
            meter_time = pd.to_datetime(meter_value[0]["timestamp"]).tz_localize(None)
            meter_reg = meter_value[0]["sampled_value"][0]["value"] # in Wh

            if np.isnan(meter_reg) or np.isinf(meter_reg):
                self.logger.warning(f"save_meter_reg ignoring: {meter_value=}")
                meter_reg=None
                with self.cpo_service.meter_df_lock:
                    self.cpo_service.cp_tx_df = self.cpo_service.cp_tx_df.append({"time":meter_time,"meas":None, "cp_group":self.cp_group, "cp_c":self.id, "done":False, "TransactionId":txID}, ignore_index=True)
                # if self.last_meter_reg is not None:
                #     meter_reg = self.last_meter_reg[1]
                #     self.last_meter_reg = (meter_time, meter_reg)
                return meter_reg
            with self.meter_reg_lock:
                #if self.last_meter_reg is not None:
                with self.cpo_service.meter_df_lock:
                    # meas = meter_reg - self.last_meter_reg[1] 
                    meas = meter_reg
                    self.cpo_service.cp_tx_df = self.cpo_service.cp_tx_df.append({"time":meter_time,"meas":meas, "cp_group":self.cp_group, "cp_c":self.id, "done":False, "TransactionId":txID}, ignore_index=True)
                self.last_meter_reg = (meter_time, meter_reg)
                # if self.last_reported_meter_reg is None:
                #     self.last_reported_meter_reg = (meter_time, meter_reg)
            return meter_reg

        @on('TransactionEvent')
        def on_tx(self, **kwargs):
            sim_clock_time = self.wattson_time.to_local_datetime(WattsonTimeType.SIM) #same as ProfileLoader
            sim_time = pd.to_datetime(sim_clock_time).tz_localize(None).isoformat()
            self.logger.debug(f"22on_tx {sim_time, [k for k in kwargs.keys()]}") #['event_type', 'timestamp', 'trigger_reason', 'seq_no', 'transaction_info', 'meter_value', 'evse', 'id_token']
            self.logger.info(f"on_tx {self.id, self.cp_group , sim_time, kwargs['timestamp'], kwargs['event_type'], kwargs['meter_value'], kwargs['custom_data']}")


            meter_reg = self.save_meter_reg(kwargs['meter_value'], txID=kwargs['transaction_info']["transaction_id"])
            
            if kwargs["event_type"] == "Started":
                return call_result.TransactionEventPayload(
                    id_token_info={ #optional if in req
                        'status': "Accepted",
                    },
                )
            else:
                return call_result.TransactionEventPayload(
                    # total_cost=0, #optional at end
                    # charging_priority=0, #optional
                    # id_token_info={ #optional if in req
                    #     'status': "Accepted",
                    #     # 'cacheExpiryDateTime': "",
                    #     # 'evseId': "",
                    # },
                    # updated_personal_message={ #optional
                    #     'format': "UTF8",
                    #     # 'language': "",
                    #     'content': "",
                    # },
                )
        
    async def on_connect(self, websocket, path):
        """ For every new charge point that connects, create a ChargePoint
        instance and start listening for messages.
        """
        try:
            requested_protocols = websocket.request_headers[
                'Sec-WebSocket-Protocol']
        except KeyError:
            self.logger.info("Client hasn't requested any Subprotocol. "
                    "Closing Connection")
            return await websocket.close()

        if websocket.subprotocol:
            self.logger.info("Protocols Matched: %s", websocket.subprotocol)
        else:
            # In the websockets lib if no subprotocols are supported by the
            # client and the server, it proceeds without a subprotocol,
            # so we have to manually close the connection.
            self.logger.warning('Protocols Mismatched | Expected Subprotocols: %s,'
                            ' but client supports  %s | Closing connection',
                            websocket.available_subprotocols,
                            requested_protocols)
            return await websocket.close()

        charge_point_id = path.strip('/')
        self.logger.info(f"charge_point_id: {charge_point_id}; {len(self.charge_point_ids)+1} out of {len(self.all_cp_ids)}")

        if charge_point_id in self.charge_point_ids:
            self.logger.warning(f"charge_point_id aleady in list: {charge_point_id, self.charge_point_ids}")
        if charge_point_id not in self.all_cp_ids:
            self.logger.warning(f"charge_point_id aleady not in list: {charge_point_id, self.all_cp_ids}")

        self.charge_point_ids.append(charge_point_id)
        cp = self.ChargePoint(charge_point_id, websocket, logger=self.logger, cpo_service=self)

        self.cp_dict[charge_point_id] = cp

        try:
            await cp.start()
        except Exception as e:
            self.logger.error(f"error in cp.start() of {charge_point_id}: {e}")
            self.logger.error(traceback.format_tb(e.__traceback__))
        self.charge_point_ids.remove(charge_point_id)



    async def main(self):
        l1=resource.getrlimit(resource.RLIMIT_NOFILE)
        self.setlimits()
        l2=resource.getrlimit(resource.RLIMIT_NOFILE)
        self.logger.info(f"WebSocket Server RLIMIT_NOFILE {l1} -> {l2}")
        self.server = await websockets.serve(
            self.on_connect,
            '0.0.0.0',
            9000,
            ping_interval=600,
            ping_timeout=600,
            #max_queue=2**8,
            max_queue=2**11,
            subprotocols=['ocpp2.0.1']
        )
        self.logger.info("WebSocket Server Started")
        await self.server.wait_closed()