
import threading
from typing import Optional

from powerowl.simulators.pandapower import PandaPowerGridModel

from wattson.apps.interface.clients import CombiClient
from wattson.cosimulation.control.interface.wattson_client import WattsonClient
from wattson.hosts.cp.dso_oscp import DSO_OSCP
from wattson.powergrid.wrapper.grid_wrapper import GridWrapper
from wattson.powergrid.remote.remote_power_grid_model import RemotePowerGridModel
from wattson.time import WattsonTime
from wattson.util import get_logger
from wattson.apps.interface.util import messages

CONTROLLER_WAIT_FOR_MTU_PER_RTU_S = 2



class DsoOscpService:
    def __init__(self,
                 management_ip: str,
                 mtu_ip: str,
                 cpo_ip: str,
                 dso_ip: str,
                 all_cp_ids: list,
                 cpo_node_id: str,
                 wattson_client_query_socket: str,
                 wattson_client_publish_socket: str):

        self.management_ip = management_ip
        self.mtu_ip = mtu_ip
        self.cpo_ip = cpo_ip
        self.dso_ip = dso_ip
        self.all_cp_ids=all_cp_ids
        self.cpo_node_id=cpo_node_id
        self.wattson_client_query_socket = wattson_client_query_socket
        self.wattson_client_publish_socket = wattson_client_publish_socket

        self.wattson_time: Optional[WattsonTime] = None
        self.power_grid_model: Optional[PandaPowerGridModel] = None
        self.remote_grid_model: Optional[RemotePowerGridModel] = None
        self.grid_wrapper: Optional[GridWrapper] = None

        self.logger = get_logger("DsoOscpService")

        self.logger.info(f"Export wattson_client_query_socket is {self.wattson_client_query_socket}")
        self.logger.info(f"Export wattson_client_publish_socket is {self.wattson_client_publish_socket}")

        self.wattson_client = WattsonClient(query_server_socket_string=self.wattson_client_query_socket,
                                            publish_server_socket_string=self.wattson_client_publish_socket,
                                            client_name="DsoOscpService")

        self.mtu_client = CombiClient(
            "DsoOscpService",
            mtu_ip=mtu_ip,
            on_cmd_reply=self._on_cmd_reply,
            on_cmd_update=self._on_cmd_update,
            on_dp_update=self._on_dp_update,
            on_update=self._on_update,
            store_dp_update=False
        )

        self._terminate = threading.Event()
        self._run=True
        self.charge_point_ids=[]


    def _on_cmd_reply(self, update: messages.Confirmation, orig_msg: messages.IECMsg):
        pass

    def _on_cmd_update(self, update: messages.Confirmation, orig_msg: messages.IECMsg = None):
        pass

    def _on_update(self, update: messages.IECMsg, orig_msg: messages.IECMsg = None):
        pass

    def _on_dp_update(self, update: messages.ProcessInfoMonitoring, ref_arg=False):
        pass
        # return self.grid_wrapper.handle_measurement(update)


    def sync(self):
        self.logger.info("re-Loading RemotePowerGridModel")
        self.remote_grid_model.synchronize(force=True, block=True)
        self.logger.info("Creating local PowerGridModel")
        self.power_grid_model = PandaPowerGridModel()
        self.power_grid_model.from_primitive_dict(self.remote_grid_model.to_primitive_dict())
        # self.grid_wrapper = GridWrapper(self.power_grid_model, self.datapoints, self.logger.getChild("GridModel"), wattson_time=self.wattson_time)


    def start(self):
        # Wattson Client
        self.logger.info("DsoOscpService Connecting Wattson Client...")
        self.wattson_client.start()
        self.wattson_client.require_connection()
        self.wattson_client.register()
        self.logger.info("DsoOscpService Connected Wattson Client")

        # Wattson Time
        self.wattson_time = self.wattson_client.get_wattson_time(enable_synchronization=True)

        # Power Grid Models: Remote and local fork
        self.logger.info("Loading RemotePowerGridModel")
        self.remote_grid_model = RemotePowerGridModel(wattson_client=self.wattson_client)
        self.logger.info("Creating local PowerGridModel")
        self.power_grid_model = PandaPowerGridModel()
        self.power_grid_model.from_primitive_dict(self.remote_grid_model.to_primitive_dict())

        # self.grid_wrapper = GridWrapper(self.power_grid_model, self.datapoints, self.logger.getChild("GridModel"), wattson_time=self.wattson_time)

        # MTU connection
        self.logger.info(f"Connecting to MTU")
        self.mtu_client.start()


        # Wait for MTU
        # rtu_count = len(self.datapoints.keys())
        rtu_count=20
        wait_time = CONTROLLER_WAIT_FOR_MTU_PER_RTU_S * rtu_count
        wait_time = 5
        self.logger.info(f"Waiting for MTU (at most {int(wait_time)}s)")
        self.wattson_client.event_wait(event_name="MTU_READY", timeout=wait_time)

        self.cpo_node = self.wattson_client.get_remote_network_node(self.cpo_node_id)
        self.cpo_node.start_pcap([i for i in self.cpo_node.get_interfaces() if not i.is_management][0])


        while self._run:
            self.thread_atk = threading.Thread(target = self.oscp_server, daemon=True)#, args = (10, )
            self.thread_atk.start()
            self.thread_atk.join()
            if self._terminate.wait(1):
                break
        self.logger.info("DsoOscpService end")


    def oscp_server(self):
        self.dso_server = DSO_OSCP(HOST_IP=self.dso_ip, mgmt_ip=self.management_ip, wattson_time=self.wattson_time)
        self.dso_server.init() #get token
        self.dso_server.oob_threaded(dso_oob_ip=self.mtu_ip, dso_oob_port=5005) #mgmt iface ip
        # self.dso_server.start_server_threaded()
        # self.dso_server.server_thread.join()
        self.dso_server.start_server()

    def stop(self):
        self.dso_server.RUN=False
        self._run=False
        # self.server.close()
        self._terminate.set()
        # self.thread_atk.join()

